/****************************/
/* 	ENVIRONMENTMAP.C   		*/
/* (c)1997 Pangea Software  */
/* By Brian Greenstone      */
/****************************/


/****************************/
/*    EXTERNALS             */
/****************************/

#include <QD3D.h>
#include <QD3DCamera.h>
#include <QD3DErrors.h>
#include <QD3DGeometry.h>
#include <QD3DGroup.h>
#include <QD3DMath.h>
#include	<QD3DTransform.h>

#include "myglobals.h"
#include "qd3d_support.h"
#include "mywindows.h"
#include "misc.h"
#include "environmentmap.h"



extern	TQ3Matrix4x4 		gWorkMatrix;
extern	Boolean				gCanDo_frsqrte;

/****************************/
/*    PROTOTYPES            */
/****************************/

inline void ReflectVector(float viewX, float viewY, float viewZ, TQ3Vector3D *N, TQ3Vector3D *out);
static void CalcEnvMap_Recurse(TQ3Object obj);
static void EnvironmentMapTriMesh(TQ3Object theTriMesh);
inline void NormalizeVector(float vx, float vy, float vz, TQ3Vector3D *outV);



/****************************/
/*    CONSTANTS             */
/****************************/

#define	MAX_REFLECTIONMAP_QUEUESIZE	50


/*********************/
/*    VARIABLES      */
/*********************/

TQ3Point3D		gCamCoord = {0,0,0};

Boolean			gShowAsReflectionMap = false;

short			gReflectionMapQueueSize = 0;
TQ3TriMeshData	gReflectionMapQueue[MAX_REFLECTIONMAP_QUEUESIZE];


/******************* INIT REFLECTION MAP QUEUE ************************/

void InitReflectionMapQueue(void)
{
short	i;

	for (i=0; i < gReflectionMapQueueSize; i++)				// free the data
	{
		if ((gReflectionMapQueue[i].numTriangleAttributeTypes == 0) &&		// see if I had zapped the #types earlier
			(gReflectionMapQueue[i].triangleAttributeTypes != nil))
			gReflectionMapQueue[i].numTriangleAttributeTypes = 1;			// reset (this is a hack, so be careful!)
			
		Q3TriMesh_EmptyData(&gReflectionMapQueue[i]);		
	}

	gReflectionMapQueueSize = 0;
}


/******************* SUBMIT REFLECTION MAP QUEUE ************************/

void SubmitReflectionMapQueue(QD3DSetupOutputType *viewInfo)
{
short	i;

	for (i=0; i < gReflectionMapQueueSize; i++)
	{
		Q3TriMesh_Submit(&gReflectionMapQueue[i], viewInfo->viewObject);	
	}
}




/************ CALC ENVIRONMENT MAPPING COORDS ************/

void CalcEnvironmentMappingCoords(TQ3Object theObj)
{
	if (gShowAsReflectionMap)
	{
		Q3Matrix4x4_SetIdentity(&gWorkMatrix);					// init to identity matrix
		CalcEnvMap_Recurse(theObj);
	}
}



/****************** CALC ENV MAP_RECURSE *********************/

static void CalcEnvMap_Recurse(TQ3Object obj)
{
TQ3Matrix4x4		transform;
TQ3GroupPosition	position;
TQ3Object   		object,baseGroup;
TQ3Matrix4x4  		stashMatrix;
TQ3ObjectType		oType;


				/*******************************/
				/* SEE IF ACCUMULATE TRANSFORM */
				/*******************************/
				
	if (Q3Object_IsType(obj,kQ3ShapeTypeTransform))
	{
  		Q3Transform_GetMatrix(obj,&transform);
  		Q3Matrix4x4_Multiply(&transform,&gWorkMatrix,&gWorkMatrix);
 	}
				/*************************/
				/* SEE IF FOUND GEOMETRY */
				/*************************/

	else
	if (Q3Object_IsType(obj,kQ3ShapeTypeGeometry))
	{
		oType = Q3Geometry_GetType(obj);						// get geometry type
		if (oType == kQ3GeometryTypeTriMesh)
			EnvironmentMapTriMesh(obj);							// map this trimesh
	}
	
			/****************************/
			/* SEE IF RECURSE SUB-GROUP */
			/****************************/

	else
	if (Q3Object_IsType(obj,kQ3ShapeTypeGroup))
 	{
 		baseGroup = obj;
  		stashMatrix = gWorkMatrix;														// push matrix
  		
  		
  		Q3Group_GetFirstPosition(obj, &position);										// get 1st object in group
  		while(position != nil)															// scan all objects in group
 		{
   			Q3Group_GetPositionObject (obj, position, &object);							// get object from group
			if (object != nil)
   			{
					/* RECURSE THIS OBJ */
					
				CalcEnvMap_Recurse(object);												// recurse this object    			
    			Q3Object_Dispose(object);												// dispose local ref
   			}
   			Q3Group_GetNextPosition(obj, &position);									// get next object in group
  		}
  		gWorkMatrix = stashMatrix;														// pop matrix
	}
}



/****************************** ENVIRONMENT MAP TRI MESH *****************************************/

static void EnvironmentMapTriMesh(TQ3Object theTriMesh)
{
TQ3Status			status;
unsigned long		numVertecies,vertNum;
TQ3Point3D			*vertexList;
TQ3TriMeshData		*triMeshDataPtr;
TQ3TriMeshAttributeData	*attribList;
short				a;
TQ3Vector3D			*normals,reflectedVector;
TQ3Param2D			*uvList;
Boolean				hasUVs = true;
TQ3Matrix4x4	invTranspose,tempm;
register	float	M00, M01, M02;	
register	float	M10, M11, M12;	
register	float	M20, M21, M22;	

			/* GET POINTER TO TRIMESH DATA STRUCTURE */
			
	triMeshDataPtr = &gReflectionMapQueue[gReflectionMapQueueSize++];


				/* GET TRIMESH INFO */
				
	status = Q3TriMesh_GetData(theTriMesh,triMeshDataPtr);						// get trimesh data
	if (status != kQ3Success) 
		DoFatalAlert("\pEnvironmentMapTriMesh: Q3TriMesh_GetData failed!");
		
	numVertecies = triMeshDataPtr->numPoints;									// get # verts
	vertexList = triMeshDataPtr->points;										// point to vert list
	attribList = triMeshDataPtr->vertexAttributeTypes;							// point to vert attib list
	
	triMeshDataPtr->numTriangleAttributeTypes = 0;								// don't need face normals (needs cleanup when disposing of data since this will hoze things)

				/* FIND UV ATTRIBUTE LIST */
				
	for (a = 0; a < triMeshDataPtr->numVertexAttributeTypes; a++)				// scan for uv
	{
		if ((attribList[a].attributeType == kQ3AttributeTypeSurfaceUV) || 
			(attribList[a].attributeType == kQ3AttributeTypeShadingUV))
		{
			uvList = attribList[a].data;										// point to list of normals
			goto got_uv;
		}
	}
	hasUVs = false;																// no uv's

got_uv:
			/* TRANSFORM BOUNDING BOX */
			
	Q3Point3D_To3DTransformArray(&triMeshDataPtr->bBox.min, &gWorkMatrix, &triMeshDataPtr->bBox.min, 2,
							 sizeof(TQ3Point3D), sizeof(TQ3Point3D));				


				/* TRANSFORM VERTEX COORDS */
				
	Q3Point3D_To3DTransformArray(vertexList, &gWorkMatrix, vertexList, numVertecies,
								 sizeof(TQ3Point3D), sizeof(TQ3Point3D));				
				
	
				/* TRANSFORM VERTEX NORMALS */
				//
				// This could be done easily with Q3Vector3D_Transform, but it's slow,
				// so this ugly code does the same thing, but faster
				//

	Q3Matrix4x4_Invert(&gWorkMatrix, &tempm);									// calc inverse-transpose matrix to nuke scale factors
	Q3Matrix4x4_Transpose(&tempm,&invTranspose);			

	M00 = invTranspose.value[0][0];												// Load matrix values into registers
	M01 = invTranspose.value[0][1];
	M02 = invTranspose.value[0][2];
	M10 = invTranspose.value[1][0];
	M11 = invTranspose.value[1][1];
	M12 = invTranspose.value[1][2];
	M20 = invTranspose.value[2][0];
	M21 = invTranspose.value[2][1];
	M22 = invTranspose.value[2][2];
				
	for (a = 0; a < triMeshDataPtr->numVertexAttributeTypes; a++)				// scan for normals
	{
		if (attribList[a].attributeType == kQ3AttributeTypeNormal)
		{
			register 	float	x,y,z;

			normals = attribList[a].data;										// point to list of normals


					/* TRANSFORM VERTEX NORMALS */
					
			for (vertNum = 0; vertNum < numVertecies; vertNum++)				// transform all normals
			{				
				x = normals[vertNum].x;											// load normal values into regs
				y = normals[vertNum].y;
				z = normals[vertNum].z;
							
				NormalizeVector(x * M00 + y * M10 + z * M20,				// do the vector-matrix mult
						x * M01 + y * M11 + z * M21,
						x * M02 + y * M12 + z * M22,
						&normals[vertNum]);										// normalize the result
			}			
			break;
		}
	}


		/****************************/
		/* CALC UVS FOR EACH VERTEX */
		/****************************/	
		
	if (hasUVs)																	// only if has UV's
	{
		register float	camCoordX,camCoordY,camCoordZ;
		
		camCoordX = gCamCoord.x;												// get cam coords into registers
		camCoordY = gCamCoord.y;
		camCoordZ = gCamCoord.z;
		
		for (vertNum = 0; vertNum < numVertecies; vertNum++)
		{
			float	eyeVectorX,eyeVectorY,eyeVectorZ;
			
					/* CALC VECTOR TO VERTEX */
					
			eyeVectorX = vertexList[vertNum].x - camCoordX;
			eyeVectorY = vertexList[vertNum].y - camCoordY;
			eyeVectorZ = vertexList[vertNum].z - camCoordZ;

		
					/* REFLECT VECTOR AROUND VERTEX NORMAL */
		
			ReflectVector(eyeVectorX, eyeVectorY, eyeVectorZ,
						&normals[vertNum],&reflectedVector);
		
		
						/* CALC UV */
				
			uvList[vertNum].u = (reflectedVector.x * .5f) + .5f;
			uvList[vertNum].v = (reflectedVector.z * .5f) + .5f;	
		}
	}	
}



/*********************** REFLECT VECTOR *************************/
//
// compute reflection vector 
// which is N(2(NV)) - V 
// N - Surface Normal 
// V - View Direction 
//
//

inline void ReflectVector(float viewX, float viewY, float viewZ, TQ3Vector3D *N, TQ3Vector3D *out)
{
float	normalX,normalY,normalZ;
float	dotProduct,reflectedX,reflectedZ,reflectedY;

	normalX = N->x;
	normalY = N->y;
	normalZ = N->z;
							
	/* compute NxV */
	dotProduct = normalX * viewX;
	dotProduct += normalY * viewY;
	dotProduct += normalZ * viewZ;
	
	/* compute 2(NxV) */
	dotProduct += dotProduct;
	
	/* compute final vector */
	reflectedX = normalX * dotProduct - viewX;
	reflectedY = normalY * dotProduct - viewY;
	reflectedZ = normalZ * dotProduct - viewZ;
	
	/* Normalize the result */
		
	NormalizeVector(reflectedX,reflectedY,reflectedZ,out);	
}



/******************** NORMALIZE VECTOR ***********************/

inline void NormalizeVector(float vx, float vy, float vz, TQ3Vector3D *outV)
{
float	temp;
	
	temp = vx * vx;
	temp += vy * vy;
	temp += vz * vz;
	
	if (gCanDo_frsqrte)							// frsqrte only available on some processors (603,604,etc).
	{
		float	isqrt, temp1, temp2;		
		static const	float	minusOneOverTwo	= -1.0 / 2.0;	
		static const	float	threeOverTwo	=  3.0 / 2.0;

		isqrt = __frsqrte (temp);					// isqrt = first approximation of 1/sqrt() 
		temp1 = temp * minusOneOverTwo;				// temp1 = -a / 2 
			
		temp2 = isqrt * isqrt;						// temp2 = sqrt^2
		temp1 *= isqrt;								// temp1 = -a * sqrt / 2 
		isqrt *= threeOverTwo;						// isqrt = 3 * sqrt / 2 
		temp = isqrt + temp1 * temp2;				// isqrt = (3 * sqrt - a * sqrt^3) / 2 
	}	
	
	else
	{
		temp = 1.0f/sqrt(temp);							// using LibMoto, this should be ~50-100 cycles (vs. 3 for frsqrte)
	}

	outV->x = vx * temp;							// return results
	outV->y = vy * temp;
	outV->z = vz * temp;
}










