/****************************/
/*        EVENTS            */
/* (c)1995 Pangea Software  */
/* By Brian Greenstone      */
/****************************/


/***************/
/* EXTERNALS   */
/***************/

#include 	<ToolUtils.h>
#include 	<QD3D.h>
#include	"myglobals.h"
#include	"myevents.h"
#include	"mymenus.h"
#include	"main.h"
#include	"misc.h"
#include "qd3d_support.h"
#include "process.h"

extern Rect			gSizeRect;
extern	WindowPtr	gModelWindow;
extern	QD3DSetupOutputType		gModelViewInfo;
extern	FSSpec		gDragFSSpec;

/****************************/
/*    PROTOTYPES            */
/****************************/

static void DoHighLevelEvent(void);
static Boolean IsDAWindow(WindowPtr whichWindow);
static OSErr  hasGotRequiredParams(AppleEvent *appEvent);


/****************************/
/*    CONSTANTS             */
/****************************/

#define MIN_SLEEP			0L
#define	NIL_MOUSE_REGION	0L


/**********************/
/*     VARIABLES      */
/**********************/

EventRecord	gTheEvent;
Boolean		gDropped3DMF = false;

/******************/
/* HANDLE EVENTS  */
/******************/

void HandleEvents()
{
char	theChar;
GrafPtr	oldPort;
		
	WaitNextEvent(everyEvent,&gTheEvent, MIN_SLEEP, NIL_MOUSE_REGION);
		
	if (gDropped3DMF)
	{
		gDropped3DMF = false;
		OpenNewModel(false, &gDragFSSpec,false);
	}
		
	switch (gTheEvent.what)
	{
		case	nullEvent:
				HandleNullEvent();
				break;
		case	mouseDown:
				HandleMouseDown();
				break;
		case	keyDown:
		case	autoKey:
				theChar	= gTheEvent.message & charCodeMask;
				if	((gTheEvent.modifiers & cmdKey) != 0)
					HandleMenuChoice(MenuKey(theChar));
				else
					HandleKeyDown(theChar);
				break;
		case	updateEvt:
				if (!IsDAWindow((WindowPtr)gTheEvent.message))
				{
					GetPort (&oldPort);
					SetPort ((WindowPtr)gTheEvent.message);
					BeginUpdate((WindowPtr)gTheEvent.message);
					DrawControls((WindowPtr)gTheEvent.message);
					
					if ((WindowPtr)gTheEvent.message == gModelWindow)
						UpdateModelWindow();
											
					EndUpdate((WindowPtr)gTheEvent.message);
					SetPort(oldPort); 
				}
				else 
				{
					BeginUpdate((WindowPtr)gTheEvent.message);
					EndUpdate((WindowPtr)gTheEvent.message);
				}
				break;
		case	kHighLevelEvent:
				DoHighLevelEvent();
				break;
	}
}


/**************** DO HIGH LEVEL EVENT *****************/

static void DoHighLevelEvent(void)
{
OSErr	myErr;

	myErr = AEProcessAppleEvent(&gTheEvent);
}


/**********************/
/* HANDLE MOUSE DOWN  */
/**********************/

void HandleMouseDown(void)
{
	WindowPtr		whichWindow;
	short 			thePart;
	long			menuChoice, windSize;
	
	thePart	=	FindWindow(gTheEvent.where, &whichWindow);
	switch(thePart)
	{
		case	inMenuBar:
				menuChoice = MenuSelect(gTheEvent.where);
				HandleMenuChoice(menuChoice);
				break;
		case	inSysWindow:
				SystemClick(&gTheEvent, whichWindow);
				break;
		case	inDrag:
				DragWindow(whichWindow,gTheEvent.where, &qd.screenBits.bounds);
				break;
		case	inGoAway:
					DisposeWindow (whichWindow);
				break;
				
		case	inContent:
				SelectWindow(whichWindow);
				InteractiveRotate();
				break;
				
		case	inGrow:
				windSize = GrowWindow(whichWindow, gTheEvent.where,
									&gSizeRect);
				if (windSize != 0)
				{
					SetPort(whichWindow);
					EraseRect(&whichWindow->portRect);
					SizeWindow(whichWindow,LoWord(windSize),
								HiWord(windSize), NORMAL_UPDATES);
					InvalRect(&whichWindow->portRect);
					
					if (whichWindow == gModelWindow)			// see if change Skeleton window
					{
						QD3D_ChangeDrawSize(&gModelViewInfo);
					}
				} 
				break;
	}
}

/*****************/
/* Is DA Window? */
/*****************/
// DA WINDOWS HAVE NEGATEIVE windowKinds //

static Boolean IsDAWindow(WindowPtr whichWindow)
{
		if	(whichWindow == nil)
				return (false);
		else
				return (((WindowPeek)whichWindow)->windowKind < 0);
}


/**********************/
/* HANDLE NULL EVENT  */
/**********************/

void HandleNullEvent()
{
	DoModelWindowNullEvent();

}


/******************** HANDLE KEY DOWN *******************/

void	HandleKeyDown(char theKey)
{
	
}



/*********************** MY APPLE EVENT: OPEN DOCUMENT *************************/

pascal OSErr MyAE_OpenDocument(AppleEvent *theAppleEvent, AppleEvent *reply,
							SInt32 handlerRefcon)
{
FSSpec      fileSpec;
AEDescList  docList;
OSErr       osError, ignoreErr;
SInt32      index, numberOfItems;
Size        actualSize;
AEKeyword   keyWord;
DescType    returnedType;
FInfo       fileInfo;

	osError = AEGetParamDesc(theAppleEvent,keyDirectObject,typeAEList,&docList);

	if(osError == noErr)
	{
    	osError = hasGotRequiredParams(theAppleEvent);
        if(osError == noErr)
        {
			AECountItems(&docList,&numberOfItems);
            if(osError == noErr)
            {
            	for(index=1;index<=numberOfItems;index++)
                {
                	osError = AEGetNthPtr(&docList,index,typeFSS,&keyWord,&returnedType,
                                      (Ptr) &fileSpec,sizeof(fileSpec),&actualSize);
                  	if(osError == noErr)
                  	{
                    	osError = FSpGetFInfo(&fileSpec,&fileInfo);
                    	if(osError == noErr)
                    	{
                    		OpenNewModel(false, &fileSpec, false);
                    	}
                  	}
                  	else
                    	ShowSystemErr(osError);
                }
			}
		}
        else
        	ShowSystemErr(osError);

		ignoreErr = AEDisposeDesc(&docList);
	}
    else
    	ShowSystemErr(osError);

	return(osError);
}

							

// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx hasGotRequiredParams

static OSErr  hasGotRequiredParams(AppleEvent *appEvent)
{
DescType  returnedType;
Size      actualSize;
OSErr     osError;

	osError = AEGetAttributePtr(appEvent,keyMissedKeywordAttr,typeWildCard,&returnedType,
                                      NULL,0,&actualSize);
    if(osError == errAEDescNotFound)
    	return(noErr);
 	else
 	if(osError == noErr)
		return(errAEParamMissed);
		
	return(osError);
}



/*********************** MY APPLE EVENT: QUIT APPLICATION *************************/

pascal OSErr MyAE_QuitApplication(AppleEvent *theAppleEvent, AppleEvent *reply,
							SInt32 handlerRefcon)
{
	CleanQuit();
	return(noErr);
}
