/*
	File:		ATIRave.h

	Contains:	ATI specific RAVE settings

	Written by:	gclement

	Copyright:	Copyright 

	Change History (most recent first):

        <2+>     12/9/98    ???     
         <2>     12/4/98    ???     Chris's v function fix
*/


#ifndef ATI_RAVE_H
#define ATI_RAVE_H
/****************************************************************/
/* headers														*/
/****************************************************************/
#include "RAVE.h"
#include "gl.h"
#include "gldVertex.h"

/****************************************************************/
/* defines														*/
/****************************************************************/
#define kQAGL_Nearest		0
#define	kQAGL_Linear		1

#define QATIClearDrawBuffer(atiExtens,drawContext,rect) \
		(*((atiExtens)->clearDrawBuffer))( drawContext, rect )

#define QATIClearZBuffer(atiExtens,drawContext,rect) \
		(*((atiExtens)->clearZBuffer))( drawContext, rect )

#define QATITextureUpdate(atiExtens,flags,pixelType,images,texture) \
		(*((atiExtens)->textureUpdate))( flags, pixelType, images, texture )
		
#define QATITextureBindCodeBook(atiExtens,texture,pCodebook) \
		(*((atiExtens)->bindCodeBook))(texture, pCodebook)				

#define QATIAccessBuffer(atiExtens,drawContext,device) \
		(*((atiExtens)->accessBuffer))(drawContext, device)

#define QATISwapBuffers(atiExtens,drawContext,dirtyRect) \
		(*((atiExtens)->swapBuffers))(drawContext, dirtyRect)
					
#define QATIDrawTriGouraud(atiExtens,drawContext,v0,v1,v2,x_off,y_off,z_off,flags) \
		(*((atiExtens)->drawTriGouraud))(drawContext, v0, v1, v2, x_off, y_off, z_off, flags)
					
#define QATIDrawTriTexture(atiExtens,drawContext,v0,v1,v2,x_off,y_off,z_off,flags) \
		(*((atiExtens)->drawTriTexture))(drawContext, v0, v1, v2, x_off, y_off, z_off, flags)
					
#define QATIDrawVGouraud(atiExtens,drawContext,nVertices,vertexMode,v_pivot,vertices,x_off,y_off,z_off,flags) \
		(*((atiExtens)->drawVGouraud))(drawContext,nVertices,vertexMode,v_pivot,vertices,x_off,y_off,z_off,flags)

#define QATIDrawVTexture(atiExtens,drawContext,nVertices,vertexMode,v_pivot,vertices,x_off,y_off,z_off,flags) \
		(*((atiExtens)->drawVTexture))(drawContext,nVertices,vertexMode,v_pivot,vertices,x_off,y_off,z_off,flags)

#define QATIAllocGouraudVertexBuffer(atiExtens,drawContext,VerticesRequested,firstVertex,pVerticesAllocated) \
		(*((atiExtens)->allocGouraudVertexBuffer))(drawContext,VerticesRequested,firstVertex,pVerticesAllocated)

#define QATIAllocTextureVertexBuffer(atiExtens,drawContext,VerticesRequested,firstVertex,pVerticesAllocated) \
		(*((atiExtens)->allocTextureVertexBuffer))(drawContext,VerticesRequested,firstVertex,pVerticesAllocated)

#define QATIFreeVertexBuffer(atiExtens,drawContext,pVertexBuffer) \
		(*((atiExtens)->freeVertexBuffer))(drawContext,pVertexBuffer)

#define QATISubmitGouraudIndexList(atiExtens,drawContext,pIndexList,indexType,numVertices,vertexMode) \
		(*((atiExtens)->submitGouraudIndexList))(drawContext,pIndexList,indexType,numVertices,vertexMode)

#define QATISubmitTextureIndexList(atiExtens,drawContext,pIndexList,indexType,numVertices,vertexMode) \
		(*((atiExtens)->submitTextureIndexList))(drawContext,pIndexList,indexType,numVertices,vertexMode)

#define VertexBufLoadGouraudVtx( buf, _x, _y, _z, _invW, _argb )								\
{																								\
	__stwbrx( *((unsigned long *)&(_x)), buf++, 0 );											\
	__stwbrx( *((unsigned long *)&(_y)), buf++, 0 );											\
	__stwbrx( *((unsigned long *)&(_z)), buf++, 0 );											\
	__stwbrx( *((unsigned long *)&(_invW)), buf++, 0 );											\
	__stwbrx( (_argb), buf++, 0 );																\
}

#define VertexBufLoadTextureVtx( buf, _x, _y, _z, _invW, _argb, _u1, _v1, _u2, _v2, _invW2 )	\
{																								\
	__stwbrx( *((unsigned long *)&(_x)), buf++, 0 );											\
	__stwbrx( *((unsigned long *)&(_y)), buf++, 0 );											\
	__stwbrx( *((unsigned long *)&(_z)), buf++, 0 );											\
	__stwbrx( *((unsigned long *)&(_invW)), buf++, 0 );											\
	__stwbrx( (_argb), buf++, 0 );																\
	__stwbrx( *((unsigned long *)&(_u1)), buf++, 0 );											\
	__stwbrx( *((unsigned long *)&(_v1)), buf++, 0 );											\
	__stwbrx( *((unsigned long *)&(_u2)), buf++, 0 );											\
	__stwbrx( *((unsigned long *)&(_v2)), buf++, 0 );											\
	__stwbrx( *((unsigned long *)&(_invW2)), buf++, 0 );										\
}

/****************************************************************/
/* typedefs														*/
/****************************************************************/
typedef enum TQATIGestaltSelector
{
	kQATIGestalt_CurrentContext 	= 1000,
	kQATIGestalt_FrameBufferMemory	= 1001,
	kQATIGestalt_NumATISelectors	= 2
} TQATIGestaltSelector;

typedef enum TQATIImagePixelType {
	kQATIPixel_RGB4444				= 1000,
	kQATIPixel_ARGB4444				= 1001,
	kQATIPixel_YUV422				= 1002,
	kQATIPixel_ARGB8				= 1003,
	kQATIPixel_NumPixelTypes		= 4
} TQATIImagePixelType;

typedef enum {
	kQATIFogDisable					= 0,
	kQATIFogExp						= 1,
	kQATIFogExp2					= 2,
	kQATIFogAlpha					= 3,
	kQATIFogLinear					= 4
} TQATIFogMode;

typedef enum {
	kQATIComposeFunctionBlend		= 0,
	kQATIComposeFunctionModulate	= 1,
	kQATIComposeFunctionAddSpec		= 2
} TQATIComposeFunction;

typedef enum {
	kQATIDitherDisable				= 0,
	kQATIDitherErrDiff				= 1,
	kQATIDitherLUT					= 2
} TQATIDitherMode;

typedef struct {
    unsigned long       flags;              /* Mask of kQATexture_xxx flags */
    unsigned long		pixelType;          /* Depth, color space, etc. */
    TQAColorTable		*colorTable;		/* Previously allocated by QAColorTableNew() */
    TQAImage      		image;           	/* Image(s) for texture */
} TQATITexImage;

typedef enum {
	kQATIRaveVertexKind				= 0,
	kQATIConixVertexKind			= 1,
	kQATIBeOSVertexKind				= 2
} TQATIVertexKind;

typedef enum {
	kQATISolidShade					= 0,
	kQATIFlatShade					= 1,
	kQATIGouraudShade				= 2
} TQATIShadeMode;

typedef enum {
	kQATITriCacheOff				= 0,
	kQATITriCacheFast				= 1,
	kQATITriCacheMid				= 2,
	kQATITriCacheBest				= 3
} TQATITriCacheMode;

typedef enum
{
	kATITriCache		=	kQATag_EngineSpecific_Minimum,
	kATIClearPreCallback=	kQATag_EngineSpecific_Minimum + 1,
 	kATIFogMode			=	kQATag_EngineSpecific_Minimum + 2,
	kATIFogColor_r		=	kQATag_EngineSpecific_Minimum + 3,
 	kATIFogColor_g		=	kQATag_EngineSpecific_Minimum + 4,
 	kATIFogColor_b		=	kQATag_EngineSpecific_Minimum + 5,
 	kATIFogColor_a		=	kQATag_EngineSpecific_Minimum + 6,
 	kATIFogDensity      =   kQATag_EngineSpecific_Minimum + 7,
 	kATIFogStart        =   kQATag_EngineSpecific_Minimum + 8,
 	kATIFogEnd          =   kQATag_EngineSpecific_Minimum + 9,
 	kATILodBias			=	kQATag_EngineSpecific_Minimum + 10,
 	kATIChipID			=	kQATag_EngineSpecific_Minimum + 11,
 	kATICompositing     =   kQATag_EngineSpecific_Minimum + 12,
	kATICompositingFunc =   kQATag_EngineSpecific_Minimum + 13,
 	kATICompositingFactor=  kQATag_EngineSpecific_Minimum + 14,
 	kATISecondTexMin    =   kQATag_EngineSpecific_Minimum + 15,
 	kATISecondTexMag    =   kQATag_EngineSpecific_Minimum + 16,
	kATIVQTexCompresion =   kQATag_EngineSpecific_Minimum + 17,
 	kATISecond_Texture  =   kQATag_EngineSpecific_Minimum + 18,
 	kATICompositingAlpha=	kQATag_EngineSpecific_Minimum + 19,
 	kATINoDefaultClear  =	kQATag_EngineSpecific_Minimum + 20,
	kATIRaveExtFuncs	=	kQATag_EngineSpecific_Minimum + 21,
	kATIEnableZWrite	=	kQATag_EngineSpecific_Minimum + 22,
	kATIRop2			=	kQATag_EngineSpecific_Minimum + 23,
	kATIAlphaTestEnable =	kQATag_EngineSpecific_Minimum + 24,
	kATIAlphaTestRef	=	kQATag_EngineSpecific_Minimum + 25,
	kATIAlphaTestFunc	=	kQATag_EngineSpecific_Minimum + 26,
	kATITexCompress		=	kQATag_EngineSpecific_Minimum + 27,
	kATI_TextureContext =	kQATag_EngineSpecific_Minimum + 28,
	kATISecond_TextureContext = kQATag_EngineSpecific_Minimum + 29,
	kATISecondTextureWrapU 	  = kQATag_EngineSpecific_Minimum + 30,
	kATISecondTextureWrapV 	  = kQATag_EngineSpecific_Minimum + 31,
	kATISafeCallbacks	= 	kQATag_EngineSpecific_Minimum + 32,
	kATITextureAddr		=	kQATag_EngineSpecific_Minimum + 33,
	kATIAlphaEnable		=	kQATag_EngineSpecific_Minimum + 34,
	kATIRedEnable		=	kQATag_EngineSpecific_Minimum + 35,
	kATIGreenEnable		=	kQATag_EngineSpecific_Minimum + 36,
	kATIBlueEnable		=	kQATag_EngineSpecific_Minimum + 37,
	kATINotifyZBuffer	=	kQATag_EngineSpecific_Minimum + 38,
	kATIModulate2x		=   kQATag_EngineSpecific_Minimum + 39,
	kATIDitherMode		=	kQATag_EngineSpecific_Minimum + 40,
	kATIChipBusy		=	kQATag_EngineSpecific_Minimum + 41,
	kATIMeshAsStrip		=	kQATag_EngineSpecific_Minimum + 42,
	kATIScaleBlit		=	kQATag_EngineSpecific_Minimum + 43,
	kATIScaleLeft		=	kQATag_EngineSpecific_Minimum + 44,
	kATIScaleRight		=	kQATag_EngineSpecific_Minimum + 45,
	kATIScaleTop		=	kQATag_EngineSpecific_Minimum + 46,
	kATIScaleBottom		=	kQATag_EngineSpecific_Minimum + 47,
	kATIAlphaOverflow	=	kQATag_EngineSpecific_Minimum + 48,
	kATIMipMap			=	kQATag_EngineSpecific_Minimum + 49,
	kATIVertexKind		=	kQATag_EngineSpecific_Minimum + 50,
	kATIShadeMode		=	kQATag_EngineSpecific_Minimum + 51,
	kATIGLAlphaEnable	=	kQATag_EngineSpecific_Minimum + 52,
	kATIWaitVSync		=	kQATag_EngineSpecific_Minimum + 53,
	kATIPageFlip		=	kQATag_EngineSpecific_Minimum + 54,
	kATIGuardBand		=	kQATag_EngineSpecific_Minimum + 55,
	kATITags			=	56
} E_ATIVAR;

/*
 * ATI RAVE extension function pointers
 */
typedef TQAError (*TQATIClearDrawBuffer)( 
	const TQADrawContext	*drawContext, 
	const TQARect 			*dirtyRect );
	
typedef TQAError (*TQATIClearZBuffer)(
	const TQADrawContext	*drawContext,
	const TQARect 			*dirtyRect );

typedef TQAError (*TQATITextureUpdate)(
	unsigned long       	flags,             	/* Mask of kQATexture_xxx flags */  
	TQAImagePixelType		pixelType,          /* Depth, color space, etc. */
	const TQAImage 			*images,
	TQATexture 				*texture );  		/* Previously allocated by QATextureNew() */			/* new Image(s) for texture contents */

typedef TQAError (*TQATITextureBindCodeBook)(
	TQATexture 				*texture,
	void					*pCodebook );

typedef TQAError (*TQATIAccessBuffer) (
	const TQADrawContext	*drawContext,
	TQADevice 				*device );

typedef TQAError (*TQATISwapBuffers) (
	const TQADrawContext	*drawContext,					/* Draw context */
	TQARect 				*dirtyRect );

typedef void (*TQATIDrawTriGouraud) (
	const TQADrawContext	*drawContext,					/* Draw context */
	const GLDVertex			*v0,
	const GLDVertex			*v1,
	const GLDVertex			*v2,
	float					x_off,
	float					y_off,
	float					z_off,
	unsigned long			flags );

typedef void (*TQATIDrawTriTexture) (
	const TQADrawContext	*drawContext,					/* Draw context */
	const GLDVertex			*v0,
	const GLDVertex			*v1,
	const GLDVertex			*v2,
	float					x_off,
	float					y_off,
	float					z_off,
	unsigned long			flags );

typedef void (*TQATIDrawVGouraud) (
	const TQADrawContext	*drawContext,					/* Draw context */
	unsigned long			nVertices,
	TQAVertexMode			vertexMode,
	const GLDVertex 		*v_pivot,
	const GLDVertex 		vertices[],
	float					x_off,
	float					y_off,
	float					z_off,
	const unsigned long		flags[] );

typedef void (*TQATIDrawVTexture) (
	const TQADrawContext	*drawContext,					/* Draw context */
	unsigned long			nVertices,
	TQAVertexMode			vertexMode,
	const GLDVertex 		*v_pivot,
	const GLDVertex 		vertices[],
	float					x_off,
	float					y_off,
	float					z_off,
	const unsigned long		flags[] );

typedef unsigned long *(*TQATIAllocVertexBuffer) (
	const TQADrawContext	*drawContext,   				/* Draw context */
	unsigned long			VerticesRequested,
	unsigned long			firstCompiledVertex,
	unsigned long			*pVerticesAllocated );

typedef void (*TQATIFreeVertexBuffer) (
 	const TQADrawContext	*drawContext,   				/* Draw context */	
	unsigned long 			*pVertexBuffer );
								
typedef void (*TQATISubmitIndexList) (
	const TQADrawContext	*drawContext,   				/* Draw context */
	void					*indexList,
	unsigned long			indexType,
	unsigned long			numVertices,
	TQAVertexMode			vertexMode );					/* One of kQAVertexMode_xxx */

typedef struct RaveExtFuncs{
	TQATIClearDrawBuffer		clearDrawBuffer;
	TQATIClearZBuffer			clearZBuffer;
	TQATITextureUpdate			textureUpdate;
	TQATITextureBindCodeBook    bindCodeBook;
	TQATIAccessBuffer			accessBuffer;
	TQATISwapBuffers			swapBuffers;
	TQATIDrawTriGouraud			drawTriGouraud;
	TQATIDrawTriTexture			drawTriTexture;
	TQATIDrawVGouraud			drawVGouraud;
	TQATIDrawVTexture			drawVTexture;
	TQATIAllocVertexBuffer		allocGouraudVertexBuffer;
	TQATIAllocVertexBuffer		allocTextureVertexBuffer;
	TQATIFreeVertexBuffer		freeVertexBuffer;
	TQATISubmitIndexList		submitGouraudIndexList;
	TQATISubmitIndexList		submitTextureIndexList;
}RaveExtFuncs;

#endif
