//
// structs.h
//

#define	MAX_ANIMS			20
#define	MAX_KEYFRAMES		20
#define	MAX_JOINTS			20
#define	MAX_LIMBS	MAX_JOINTS


		/* CURRENT JOINT STATE */
		
typedef struct
{
	long		tick;					// time at which this state exists
	long		accelerationMode;		// mode of in/out acceleration
	TQ3Point3D	coord;					// current 3D coords of joint (relative to link)
	TQ3Vector3D	rotation;				// current rotation values of joint (relative to link)
	TQ3Vector3D	scale;					// current scale values of joint mesh
}JointKeyframeType;


		/* JOINT SPECIFICATIONS */
		
typedef struct
{
	long			limbGeometryIndex;			// index into g3DMFLimbList
	TQ3Vector3D		maxRot;						// max rot values of joint
	TQ3Vector3D		minRot;						// min rot values of joint
	long 			previousJointIndex; 		// index to previous link joint definition
	TQ3GroupObject	groupObject;				// reference to this joint's group object
	TQ3Matrix4x4		rotXYZMatrix;			// matrix that holds the following xyz xform
	TQ3TransformObject	rotXYZTransform;		// rotXYZTransform object
	TQ3TransformObject	transTransform;			// translate transform object
	TQ3TransformObject	scaleTransform;			// scaleTransform object
	long			numKeyFrames[MAX_ANIMS];	// # keyframes
	JointKeyframeType keyFrames[MAX_ANIMS][MAX_KEYFRAMES];		// keyframe data
}JointDefinitionType;

			/* ANIM EVENT TYPE */
			
typedef struct
{
	short	time;
	Byte	type;
	Byte	value;
}AnimEventType;


			/*  OBJECT RECORD STRUCTURE */


struct ObjNode
{
	long			NodeNum;		// node # in array (for internal use)
	struct ObjNode	*PrevNode;		// address of previous node in linked list
	struct ObjNode	*NextNode;		// address of next node in linked list
	long		Genre;				// obj genre: 0=sprite, 1=nonsprite
	long		SortSlot;			// sort value
	long		Type;				// obj type
	long		AnimNum;			// sub type (anim type)
	long		ModeFlags;		
	void		(*MoveCall)(void);	// pointer to object's move routine
	TQ3Point3D	Coord;
	TQ3Vector3D	Velocity;
	float		RotX,RotY,RotZ;
	float		ScaleX,ScaleY,ScaleZ;
	Boolean				Flag[4];
	long				Special[4];
	float				Floats[4];
	unsigned long		CType;		// collision type bits
	unsigned long		CBits;		// collision attribute bits
	long			Kind;			// kind
	long			Health;			// health
	Byte		ControlMode;		// keyboard,mouse,network, etc.
	
	TQ3Matrix4x4		BaseTransformMatrix;
	TQ3TransformObject	BaseTransformObject;		// illegal ref to BaseTransformMatrix

	TQ3GroupObject	BaseGroup;		// group containing all geometry,etc. for this object (for drawing)

					/***************************/
					/* SKELETON SPECIFIC STUFF */
					/***************************/
					
	long			NumJoints;						// # joints in this skeleton object
	JointDefinitionType	*JointDefinitions;			// array of joint definitions
	JointKeyframeType	*JointCurrentPosition;		// for each joint, holds current interpolated keyframe values

	long			NumLimbs;						// # limb geometries in skeleton obj
	TQ3GroupObject	*LimbObjects;					// references to each limb's geometry
	
	long			NumAnims;						// # animations in this skeleton object
	short			*NumAnimEventsList;				// ptr to array of NumAnimEvents.  Each entry is the # of animevents in each anim
	AnimEventType	**AnimEventsList;				// 2 dimensional array which holds a anim event list for each anim
	float			CurrentAnimTime;				// current time index for animation	
	float			LoopBackTime;					// time to loop or zigzag back to (default = 0 unless set by a setmarker)
	float			MaxAnimTime;					// duration of current anim
	float			AnimSpeed;						// time factor for speed of executing current anim (1.0 = normal time)
	short			AnimEventIndex;					// current index into anim event list
	Byte			AnimDirection;					// if going forward in timeline or backward			
	Byte			EndMode;						// what to do when reach end of animation
	Boolean			AnimHasStopped;					// flag gets set when anim has reached end of sequence (looping anims don't set this!)

	Boolean			IsMorphing;						// flag set when morphing from an anim to another
	JointKeyframeType	*MorphStart,*MorphEnd;		// morph start & end keyframes for each joint
	float			MorphSpeed;						// speed of morphing (1.0 = normal)
	float			MorphPercent;					// percentage of morph from kf1 to kf2 (0.0 - 1.0)

};
typedef struct ObjNode ObjNode;


		/* NEW OBJECT DEFINITION TYPE */
		
typedef struct
{
	long	genre,type,animNum;
	TQ3Point3D	coord;
	unsigned long	flags;
	long	slot;
	void	(*moveCall)(void);
}NewObjectDefinitionType;
