/****************************/
/*    3DMF OPT - MAIN 		*/
/* (c)1995-99 Pangea Software */
/* By Brian Greenstone      */
/****************************/


/****************************/
/*    EXTERNALS             */
/****************************/

#include 	<Fonts.h>
#include 	<Folders.h>
#include 	<Resources.h>
#include 	<Movies.h>
#include 	<Script.h>
#include 	<Gestalt.h>
#include <Rave.h>
#include <QD3D.h>
#include <QD3DCamera.h>
#include <QD3DDrawContext.h>
#include <QD3DErrors.h>
#include <QD3DGeometry.h>
#include <QD3DGroup.h>
#include <QD3DIO.h>
#include "QD3DMath.h"
#include	<QD3DTransform.h>
#include	<QD3DStyle.h>
#include "myglobals.h"
#include "qd3d_support.h"
#include "objects.h"
#include "mymenus.h"
#include "mywindows.h"
#include "myevents.h"
#include "main.h"
#include "misc.h"
#include "3dmf.h"
#include "process.h"
#include "dialogStuff.h"
#include "optimize.h"
#include "triangulator.h"

extern	NewObjectDefinitionType	gNewObjectDefinition;
extern	QD3DSetupOutputType		gModelViewInfo;
extern	Boolean	gShowContentStructure,gSaveTextToFile,gReduce32BitTextures;
extern	Boolean	gSaveAsBinary,gConvertToMipmap,gUseFaceNormals,gUseVertexNormals;
extern	Boolean	gBlackIsTransparent,gUseCullingGroup;
extern	Boolean	gAutoCenter,gCenterYAtZero,gAverageNormals,gMakeStripsFlag;
extern	Boolean	gCompressTextures,gKeepVertexColors;
extern	CodecQ			gCodecQuality;
extern	CodecType		gCodecType;
extern	CodecComponent 	gCodecComponent;
extern	long			gMaxTrianglesPerMaterial;
extern	TQ3SubdivisionStyleData	gSubdivData;


/****************************/
/*    PROTOTYPES            */
/****************************/

static	void ToolBoxInit(void);


/****************************/
/*    CONSTANTS             */
/****************************/

#define	MAX_FILES_IN_BATCH	500

#define	kQAVendor_ATI	1

typedef struct
{
	Boolean	showContentStructure,saveTextToFile;
	Boolean	saveAsBinary,convertToMipmap,useFaceNormals;
	Boolean	makeTexturesSquare,blackIsTransparent;
	Boolean	autoCenter,centerYAtZero,averageNormals,makeStripsFlag;
	Boolean	useBatchSuffix;
	char	batchSuffix[32];
	Boolean	compressTextures;
	CodecQ	codecQuality;
	CodecType		codecType;
	CodecComponent	codecComponent;
	long	numTrianglesPerMaterial;
	Boolean	useCullingGroups;
	float	tesselation;
	Boolean	keepVertexColors;
	Boolean	useVertexNormals;
	Boolean	reduce32BitTextures;
	Byte	reserved[100];
}PrefBlockType;


/****************************/
/*    VARIABLES             */
/****************************/

Boolean	gQD3DInitialized = false;

short	gPrefsFolderVRefNum;
long	gPrefsFolderDirID;

short		gMainAppRezFile;
FSSpec		gCurrentBatchSpec;

Boolean		gCanDo_frsqrte = false;

Boolean		gUseBatchSuffix = false;
Str255		gBatchSuffix;

#if ATI_LICENSE
Boolean		gExpertMode = false;
#else
Boolean		gExpertMode = true;
#endif


/*********************** BATCH PROCESS ********************************/

void BatchProcess(void)
{
StandardFileReply	reply;
SFTypeList	typeList;
short		numTypes;
FSSpec		spec,*specList;
OSErr		iErr;
CInfoPBRec	paramBlock;
Str255		name;
short		numFilesInDirectory,fileNum,numSpecs;
FInfo		fndrInfo;



			/* DO SAVE OPTIONS */
			
	typeList[0] = '3DMF';
	numTypes = 1;
		
	StandardGetFile(nil,numTypes,typeList,&reply);
	if (!reply.sfGood)								// see if cancelled 
		return;

	spec = reply.sfFile;							// copy FSSpec 

	ShowWaitingWindow();

	gSaveTextToFile = false;
	gShowContentStructure = false;

			/**********************/
			/* GET DIRECTORY INFO */
			/**********************/
			
	paramBlock.dirInfo.ioCompletion = nil;
	paramBlock.dirInfo.ioNamePtr = nil;
	paramBlock.dirInfo.ioVRefNum = spec.vRefNum;
	paramBlock.dirInfo.ioFDirIndex = -1;					// get dir info
	paramBlock.hFileInfo.ioDirID = spec.parID;
	
	iErr = PBGetCatInfoAsync(&paramBlock);
	if (iErr)
		DoFatalAlert("\pBatchProcess: PBGetCatInfo(Dir) failed!");

	numFilesInDirectory = paramBlock.dirInfo.ioDrNmFls;					// get # files & subdirs in selected directory

			/***************************/
			/* CREATE MY LIST OF FILES */
			/***************************/
			
	specList = (FSSpec *)AllocPtr(sizeof(FSSpec) * MAX_FILES_IN_BATCH);	// alloc mem for FSSpecs
	if (specList == nil)
		DoFatalAlert("\pBatchProcess: AllocPtr failed!");
			
	numSpecs = 0;
	for (fileNum = 1; fileNum <= numFilesInDirectory; fileNum++)
	{
				/* GET FILE INFO */
				
		paramBlock.dirInfo.ioCompletion = nil;
		paramBlock.dirInfo.ioNamePtr = name;
		paramBlock.dirInfo.ioVRefNum = spec.vRefNum;
		paramBlock.dirInfo.ioFDirIndex = fileNum;								// get file info
		paramBlock.hFileInfo.ioDirID = spec.parID;

		iErr = PBGetCatInfoAsync(&paramBlock);
		if (iErr)
			DoFatalAlert("\pBatchProcess: PBGetCatInfo(File) failed!");

		if (paramBlock.dirInfo.ioFlAttrib & (1<<4))										// if bit 4 is set, then this is a directory
			continue;

				/* SKIP NON-3DMF FILES */

		iErr = HGetFInfo(spec.vRefNum, spec.parID, name, &fndrInfo);			// get file type
		if (iErr)
		{
			DoAlert("\pBatchProcess: HGetFInfo failed!");
			DoAlert(name);
			ShowSystemErr(iErr);
		}

		if (fndrInfo.fdType != '3DMF')					// make sure its 3DMF file
			continue;
			
				/* MAKE FSSPEC */
				
		if (numSpecs > MAX_FILES_IN_BATCH)
		{
			DoAlert("\pSorry, but numFilesInDirectory > MAX_FILES_IN_BATCH");
			goto bye;
		}
		
		iErr = FSMakeFSSpec(spec.vRefNum,spec.parID, name, &specList[numSpecs++]);
		if (iErr)
			DoFatalAlert("\pBatchProcess: FSMakeFSSpec failed!");

	}
				/***********************************/
				/* PROCESS EACH 3DMF FILE IN THERE */
				/***********************************/
				
	for (fileNum = 0; fileNum < numSpecs; fileNum++)
	{

				/* PROCESS THIS FILE */
				

		OpenNewModel(true,&specList[fileNum],false);


				/* PRE TRANSFORM MESH VERTECIES */
				
		OptimizeGeometry();

			
				/* SAVE THE FILE OUT */

		if (gUseBatchSuffix)
		{
			FSSpec	spec2 = specList[fileNum];
			short	a = spec2.name[0];
			short	b = gBatchSuffix[0];
			short	n;
			
			n = a+b;
			if (n > 31)									// see if filename will be too long
			{
				a -= n-31;
				spec2.name[0] = a;
			}
				
			for (n = 0; n < b; n++)						// append suffix to filename
			{
				spec2.name[a+n+1] = gBatchSuffix[1+n];
				spec2.name[0]++;
			}
			
		
			Save3DMFModel(&gModelViewInfo,&spec2,WriteModelGeometry);	
		}
		else
			Save3DMFModel(&gModelViewInfo,&specList[fileNum],WriteModelGeometry);	
			
		DoModelWindowNullEvent();
	}

bye:
	KillWaitingWindow();
	DisposePtr((Ptr)specList);
}


/*****************/
/* TOOLBOX INIT  */
/*****************/

static void ToolBoxInit(void)
{
TQ3Status	myStatus;
long		response;
OSErr		iErr;

 	MaxApplZone();
	MoreMasters(); MoreMasters(); MoreMasters(); MoreMasters();
	InitGraf(&qd.thePort);
	FlushEvents ( everyEvent, REMOVE_ALL_EVENTS);
	InitFonts();
	InitWindows();
	InitDialogs(nil);
	InitCursor();
	InitMenus();
	TEInit();
	
	gMainAppRezFile = CurResFile();
	
	
			/* SEE IF QD3D AVAILABLE */
	
	if((void *)Q3Initialize == (void *)kUnresolvedCFragSymbolAddress)
		DoFatalAlert("\pQuickDraw 3D version 1.5 or better is required to run this application!");

#ifdef gestaltQD3DPresent
	Gestalt(gestaltQD3D, &response);
	if (response & (1<<gestaltQD3DPresent))
	{	
		Gestalt(gestaltQD3DVersion,&response);
		if (response < 0x10600)						// must be using 1.6 or better
		{
err:		
			DoFatalAlert("\pQuickDraw 3D version 1.6 or better is required to run this application!");
		}
	}
	else
		goto err;
#endif	
	
	myStatus = Q3Initialize();
	if ( myStatus == kQ3Failure )
		DoFatalAlert("\pQ3Initialize returned failure.");
		
	gQD3DInitialized = true;
				
		/* SEE IF PROCESSOR SUPPORTS frsqrte */
	
	if (!Gestalt(gestaltNativeCPUtype, &response))
	{
		switch(response)
		{
			case	gestaltCPU603:
			case	gestaltCPU604:
			case	gestaltCPU603e:
					gCanDo_frsqrte = true;
					break;
					
			default:	
					gCanDo_frsqrte = false;
		}
	}
	
	
			/* INIT APPLE EVENTS */
			                                    
	iErr = AEInstallEventHandler(kCoreEventClass, kAEOpenDocuments,			// install an AE
						 NewAEEventHandlerProc(MyAE_OpenDocument), 0, false);			

	iErr |= AEInstallEventHandler(kCoreEventClass, kAEQuitApplication,
						 NewAEEventHandlerProc(MyAE_QuitApplication), 0, false);			

	if (iErr)
		DoAlert("\pUnable to install AppleEvent!");
		
		
		
			/* CHECK PREFERENCES FOLDER */
			
	iErr = FindFolder(kOnSystemDisk,kPreferencesFolderType,kDontCreateFolder,		// locate the folder
					&gPrefsFolderVRefNum,&gPrefsFolderDirID);
	if (iErr != noErr)
		DoAlert("\pWarning: Cannot locate Preferences folder.");

//	iErr = DirCreate(gPrefsFolderVRefNum,gPrefsFolderDirID,"\p3DMF Optimizer",&createdDirID);	// make PillowF folder in there


}



/******************** SAVE PREFS **********************/

void SavePrefs(void)
{
FSSpec				file;
OSErr				iErr;
short				refNum,i;
long				count;
PrefBlockType		pref;
			
				/* CREATE BLANK FILE */
				
	FSMakeFSSpec(gPrefsFolderVRefNum, gPrefsFolderDirID, "\p3DMFOptimizerPrefs", &file);
	FSpDelete(&file);															// delete any existing file
	iErr = FSpCreate(&file, 'OP40', 'Pref', smSystemScript);					// create blank file
	if (iErr)
		return;

				/* OPEN FILE */
					
	iErr = FSpOpenDF(&file, fsRdWrPerm, &refNum);
	if (iErr)
	{
		FSpDelete(&file);
		return;
	}
	
			/* CREATE DATA BLOCK */
			
	pref.showContentStructure = gShowContentStructure;
	pref.saveTextToFile = gSaveTextToFile;
	pref.saveAsBinary = gSaveAsBinary;
	pref.convertToMipmap = gConvertToMipmap;
	pref.useFaceNormals = gUseFaceNormals;
	pref.makeTexturesSquare = false;					// no longer used
	pref.blackIsTransparent = gBlackIsTransparent;
	pref.autoCenter = gAutoCenter;
	pref.centerYAtZero = gCenterYAtZero;
	pref.averageNormals = gAverageNormals;
	pref.makeStripsFlag = gMakeStripsFlag;
	pref.useBatchSuffix = gUseBatchSuffix;
	
	pref.compressTextures 	= gCompressTextures;
	pref.codecQuality		= gCodecQuality;
	pref.codecType 			= gCodecType;
	pref.codecComponent		= gCodecComponent;
	
	pref.numTrianglesPerMaterial = gMaxTrianglesPerMaterial;	
	pref.useCullingGroups 	= gUseCullingGroup;	
	pref.tesselation = gSubdivData.c1;
	
	pref.useVertexNormals = gUseVertexNormals;
	pref.keepVertexColors = gKeepVertexColors;
	
	
	for (i=0; i < 32; i++)
		pref.batchSuffix[i] = gBatchSuffix[i];
	
	pref.reduce32BitTextures = gReduce32BitTextures;
			
	for (i=0; i < 100; i++)
		pref.reserved[i] = 0;

		
				/* WRITE DATA */

	count = sizeof(PrefBlockType);
	FSWrite(refNum, &count, (Ptr)&pref);	
	FSClose(refNum);
}


/******************** LOAD PREFS **********************/

void LoadPrefs(void)
{
OSErr				iErr;
short				refNum,i;
FSSpec				file;
long				count;
PrefBlockType		pref;
				
				/*************/
				/* READ FILE */
				/*************/
					
	FSMakeFSSpec(gPrefsFolderVRefNum, gPrefsFolderDirID, "\p3DMFOptimizerPrefs", &file);
	iErr = FSpOpenDF(&file, fsRdPerm, &refNum);	
	if (iErr)
		return;

	count = sizeof(PrefBlockType);
	iErr = FSRead(refNum, &count,  (Ptr)&pref);		// read data from file
	if (iErr)
	{
		FSClose(refNum);			
		return;
	}
	
	FSClose(refNum);			
	
	
			/* READ DATA BLOCK */
			
	gShowContentStructure = pref.showContentStructure;
	gSaveTextToFile = pref.saveTextToFile;
	gSaveAsBinary = pref.saveAsBinary;
	gConvertToMipmap = pref.convertToMipmap;
	gUseFaceNormals = pref.useFaceNormals;
	gBlackIsTransparent = pref.blackIsTransparent;
	gAutoCenter = pref.autoCenter;
	gCenterYAtZero = pref.centerYAtZero;
	gAverageNormals = pref.averageNormals;
	gMakeStripsFlag = pref.makeStripsFlag;
	gUseBatchSuffix = pref.useBatchSuffix;
	
	gCompressTextures = pref.compressTextures;
	gCodecQuality = pref.codecQuality;
	gCodecType = pref.codecType;
	gCodecComponent = pref.codecComponent;
	
	gMaxTrianglesPerMaterial = pref.numTrianglesPerMaterial;
	
	gUseCullingGroup = pref.useCullingGroups;
	
	gUseVertexNormals = pref.useVertexNormals;
	gKeepVertexColors = pref.keepVertexColors;
	
	if (pref.tesselation == 0.0f)
		pref.tesselation = DEFAULT_TESS;
	gSubdivData.c1 = gSubdivData.c2 = pref.tesselation;
	
	gReduce32BitTextures = pref.reduce32BitTextures;
	
	
	for (i=0; i < 32; i++)
		gBatchSuffix[i] = pref.batchSuffix[i];
}



/************************************************************/
/******************** PROGRAM MAIN ENTRY  *******************/
/************************************************************/


void main(void)
{
unsigned long	someLong;

	ToolBoxInit();
 		
 		
 		/* FOR ATI VERSION, MAKE SURE ONLY WORKS WITH ATI CARD INSTALLED */
 		
#if ATI_LICENSE
	{
		TQADevice		raveDevice;
		TQAEngine		*myEngine;

		raveDevice.deviceType = kQADeviceGDevice;
		raveDevice.device.gDevice = GetGDevice();

		for (myEngine = QADeviceGetFirstEngine(&raveDevice); myEngine; myEngine = QADeviceGetNextEngine(&raveDevice,myEngine))
		{
			if (QAEngineGestalt(myEngine, kQAGestalt_VendorID, &someLong) == kQANoErr)
				if (someLong == kQAVendor_ATI)
					goto ati_exists;
					
		}
		DoFatalAlert("\pSorry, but this version of 3DMF Optimizer requires that you have an ATI 3D card and the appropriate drivers installed on your machine.");
		
ati_exists:;		
	}
	

#endif 		
 		
 		
 		
	InitWindowStuff();
	InitMenuBar();

	GetDateTime ((unsigned long *)(&someLong));		// init random seed
	SetMyRandomSeed(someLong);

	CopyPStr("\p.opt",gBatchSuffix);				// init batch suffix string


				/* INIT STUFF */


	LoadPrefs();
	InitObjectManager();
	InitModelWindow();								// create QD3D window environment	
	CreateMyTriangulatorRenderer();					// (must do this after InitModelWindow)

				/* PROGRAM MAIN LOOP */

	while (true)
		HandleEvents();
}











