/****************************/
/*        MENUS             */
/* (c)1995 Pangea Software  */
/* By Brian Greenstone      */
/****************************/


/***************/
/* EXTERNALS   */
/***************/

#include <ToolUtils.h>
#include <Devices.h>
#include <QD3D.h>
#include <QD3DCamera.h>
#include <QD3DDrawContext.h>
#include <QD3DErrors.h>
#include <QD3DGeometry.h>
#include <QD3DGroup.h>
#include <QD3DIO.h>
#include	<QD3DTransform.h>
#include	"myglobals.h"
#include	 "qd3d_support.h"
#include	"mymenus.h"
#include	"misc.h"
#include	"main.h"
#include	"process.h"
#include	"objects.h"
#include	"dialogstuff.h"
#include	"mywindows.h"
#include	"optimize.h"
#include	"wavefrontobj.h"
#include "bg3d.h"

extern	EventRecord			gTheEvent;
extern	QD3DSetupOutputType		gModelViewInfo;
extern Boolean		gShowAsReflectionMap,gUseNULLShader,gUseGLBlending,gPreMultiply32BitTextures;
extern	WindowPtr	gBreakdownWindow;


/******************/
/*  PROTOTYPES    */
/******************/

static void HandleSpecialMenuChoice(short theItem);


/****************************/
/*    CONSTANTS             */
/****************************/

									// MENU BAR ITEMS //
#define	NOT_A_NORMAL_MENU	-1
#define	APPLE_MENU_ID		400
#define	FILE_MENU_ID		401
#define EDIT_MENU_ID		402
#define SPECIAL_MENU_ID		128


									// EDIT MENU ITEMS //
									
enum
{
	UNDO_ITEM	=	1,
	CUT_ITEM	=	3,
	COPY_ITEM	=	4,	
	PASTE_ITEM	=	5,
	CLEAR_ITEM	=	6,
	BATCHPREFS_ITEM = 8,
	OPTPREFS_ITEM = 9,
	OPTIMIZE_ITEM = 10
};

									// FILE MENU ITEMS //
enum
{
	OPEN_ITEM			= 1,
	SAVE_ITEM			= 2,
	BATCH_ITEM			= 3,
	OPEN_OPT_OBJ_ITEM 	= 4,
	EXPORT_BG3D_ITEM	= 5,
	QUIT_ITEM			= 7
};


									// APPLE MENU ITEMS //
enum									
{									
	ABOUT_ITEM = 1,
	HELP_ITEM = 2
};

									// SPECIAL MENU ITEMS //
enum									
{									
	REFLECTIONMAP_ITEM = 1,
	SCALE_ITEM			= 2,
	NULL_SHADER_ITEM	= 3,
	GL_BLENDING_ITEM	= 4,
	PREMULT_ITEM		= 5
};


/**********************/
/*     VARIABLES      */
/**********************/

MenuHandle	gAppleMenu;


/******************/
/* INIT MENU BAR  */
/******************/

void InitMenuBar(void)
{
Handle	myMenuBar;
		
					/* ALLOCATE MAIN MENU BAR */
	
#if ATI_LICENSE
	if ((myMenuBar	=	GetNewMBar(128)) == nil)
			DoFatalAlert("\pWhered the menu bar go?!");;
#else
	if ((myMenuBar	=	GetNewMBar(400)) == nil)
			DoFatalAlert("\pWhered the menu bar go?!");;
#endif						
	SetMenuBar(myMenuBar);
	
					/* SET APPLE MENU */

	if ((gAppleMenu	=	GetMenuHandle(400)) == nil)
			DoFatalAlert("\pGetMHandle failed!");
	AppendResMenu (gAppleMenu, 'DRVR');						// APPEND DESK ACCESSORIES
	
	DrawMenuBar();
}


/***************************/
/* HANDLE MENU BAR CHOICE  */
/***************************/

void HandleMenuChoice(long menuChoice)
{
short	theMenu;
short	theItem;
		
	if	(menuChoice != 0)
	{
		theMenu	=	HiWord(menuChoice);
		theItem	=	LoWord(menuChoice);
		switch	(theMenu)
		{
			case	APPLE_MENU_ID:
					HandleAppleChoice(theItem);
					break;
					
			case	FILE_MENU_ID:
					HandleFileChoice(theItem);
					break;
					
			case	EDIT_MENU_ID:
					HandleEditChoice(theItem);
					break;
					
			case	SPECIAL_MENU_ID:
					HandleSpecialMenuChoice(theItem);
					break;
					
		}
		HiliteMenu(0);
	}
}


/*****************************/
/* HANDLE APPLE MENU CHOICE  */
/*****************************/

void HandleAppleChoice(short theItem)
{
Str255	accName;
short		accNumber;
		
	switch	(theItem)
	{
		case	ABOUT_ITEM:
		
#if ATI_LICENSE
				Alert(129,nil);
#elif DEMO_VERSION			
				Alert(400,nil);
#else
				Alert(402,nil);
#endif				
				break;
		case	HELP_ITEM:
				Alert(128,nil);
				break;
			
		default:
				GetMenuItemText(gAppleMenu,theItem,accName);
				accNumber	=	OpenDeskAcc(accName);
				break;
	}
}

/****************************/
/* HANDLE FILE MENU CHOICE  */
/****************************/

void HandleFileChoice(short theItem)
{

	switch(theItem)
	{
		case	OPEN_ITEM:
				OpenNewModel(false,nil,true);
				break;				

		case	SAVE_ITEM:
				SaveOptimizedModel();
				break;				

		case	BATCH_ITEM:
				BatchProcess();
				break;
				
		case	OPEN_OPT_OBJ_ITEM:
				OpenAndOptimizeOBJFile();
				break;

		case	EXPORT_BG3D_ITEM:
				ExportBG3D();
				break;

		case	QUIT_ITEM:
				DeleteAllObjects();
				QD3D_DisposeWindowSetup(&gModelViewInfo);
				DisposeWindow(gModelViewInfo.window);
				CleanQuit();
				break;				
	}
}


/****************************/
/* HANDLE SPECIAL MENU CHOICE  */
/****************************/

static void HandleSpecialMenuChoice(short theItem)
{
MenuHandle	m;

	switch(theItem)
	{				
		case	REFLECTIONMAP_ITEM:
				gShowAsReflectionMap = !gShowAsReflectionMap;
				break;
				
		case	SCALE_ITEM:
				if (!DoScaleDialog())
				{
					ShowWaitingWindow();
					OptimizeGeometry();
					KillWaitingWindow();
				}
				break;
				
		case	NULL_SHADER_ITEM:
				gUseNULLShader = !gUseNULLShader;
				m = GetMenuHandle(SPECIAL_MENU_ID);
				SetItemMark(m, theItem, checkMark * gUseNULLShader);
				break;
				
		case	GL_BLENDING_ITEM:		
				gUseGLBlending = !gUseGLBlending;
				m = GetMenuHandle(SPECIAL_MENU_ID);
				SetItemMark(m, theItem, checkMark * gUseGLBlending);
				DoMyRAVECalls();
				DrawModelWindow();
				break;
				
		case	PREMULT_ITEM:
				gPreMultiply32BitTextures = true;
				ShowWaitingWindow();
				OptimizeGeometry();
				KillWaitingWindow();
				gPreMultiply32BitTextures = false;
				break;
		
	}
}


/****************************/
/* HANDLE EDIT MENU CHOICE  */
/****************************/

void HandleEditChoice(short theItem)
{
	SystemEdit(theItem-1);
		
	switch(theItem)
	{
		case	BATCHPREFS_ITEM:
				DoBatchPrefsDialog();
				break;
				
		case	OPTPREFS_ITEM:
				DoOptimizationPrefsDialog();
				break;
				
		case	OPTIMIZE_ITEM:
				ShowWaitingWindow();
				OptimizeGeometry();
				KillWaitingWindow();
				break;
	}
}



