/****************************/
/*   	OPTIMIZE.C     		*/
/* (c)1996 Pangea Software  */
/* By Brian Greenstone      */
/****************************/


/****************************/
/*    EXTERNALS             */
/****************************/

#include <QD3D.h>
#include <QD3DErrors.h>
#include <QD3DGeometry.h>
#include <QD3DGroup.h>
#include "QD3DMath.h"
#include "QD3DShader.h"
#include "QD3DStyle.h"
#include	<QD3DTransform.h>

#include "myglobals.h"
#include "objects.h"
#include "mymenus.h"
#include "mywindows.h"
#include "myevents.h"
#include "misc.h"
#include "process.h"
#include "qd3d_support.h"
#include "3dmf.h"
#include "dialogStuff.h"
#include "optimize.h"
#include "textureOpt.h"
#include "triangulator.h"

extern	EventRecord			gTheEvent;
extern	NewObjectDefinitionType	gNewObjectDefinition;
extern	ObjNode				*gCurrentNode,*gModelObjNode;
extern	float				gFramesPerSecond;
extern	TQ3Object			gModelGroup;
extern	QD3DSetupOutputType		gModelViewInfo;
extern	FSSpec				gLast3DMF_FSSpec;
extern	Boolean					gAutoCenter;
extern	Boolean					gCenterYAtZero,gCurrentIsOptimized;
extern	Boolean			gShowContentStructure,gSaveTextToFile;
extern	long			gTotalNumVertecies,gTotalNumFaces,gTotalMeshes;
extern	float					gAutoCenterXOff,gAutoCenterYOff,gAutoCenterZOff;

/****************************/
/*    PROTOTYPES            */
/****************************/

static void RecurseGeometry(TQ3Object obj);
static void PreTransformAttribute(TQ3AttributeSet theAttribute);
static void ConvertAttributeSetToMaterialType(TQ3AttributeSet compositeAttribSet,MaterialType *materialType);
static TQ3AttributeSet CreateAttributeSetFromMaterial(MaterialType *material);
static void OptimizeTriangleListTextures(void);

static void StripifyTriMesh(TQ3Object triMesh);
static void Stripify_Forward(long numTriangles, TQ3TriMeshTriangleData *triangles, TQ3Vector3D *faceNormals);
static void Stripify_Backward(long numTriangles, TQ3TriMeshTriangleData *triangles, TQ3Vector3D *faceNormals);

static void RemoveUnwantedAttribs(TQ3AttributeSet attribs);


/****************************/
/*    CONSTANTS             */
/****************************/

#define	MAX_MATERIAL_TYPES			512

/*********************/
/*    VARIABLES      */
/*********************/

TQ3Matrix4x4 		gWorkMatrix;
TQ3AttributeSet		gWorkAttribSet;
TQ3Object			gWorkShader;

short				gNumMaterialTypes = 0;
MaterialType		gMaterialTypeList[MAX_MATERIAL_TYPES];

short				gNumTriMeshes = 0;
TQ3GeometryObject	gTriMeshObjects[MAX_MATERIAL_TYPES];

float				gBoundsLeft,gBoundsRight,gBoundsFront,gBoundsBack;

Boolean				gUseFaceNormals = true,gFoundBadGeneralPolygon;
Boolean				gMakeStripsFlag = false;

Boolean				gAverageNormals = false;
Boolean				gUseCullingGroup = false;

Boolean				gUseVertexNormals = true;
Boolean				gKeepVertexColors = false;

unsigned short		gNumStripBreaks;

long				gMaxTrianglesPerMaterial = 1500;

float				gScale = 1.0;



/******************** OPTIMIZATION INIT ************************/

void OptimizationInit(void)
{			
int		i;

	InitTextureList();

	gTotalNumVertecies = gTotalNumFaces = 0;
	gFoundBadGeneralPolygon= false;
			
	for (i=0; i < gNumTriMeshes; i++)											// dispose old stuff
		Q3Object_Dispose(gTriMeshObjects[i]);
	gNumTriMeshes = 0;					
				
	for (i=0; i < gNumMaterialTypes; i++)										// free old material memory
	{
		if (gMaterialTypeList[i].triangleList)
		{
			DisposePtr((Ptr)gMaterialTypeList[i].triangleList);
			gMaterialTypeList[i].triangleList = nil;
		}
			
		if (gMaterialTypeList[i].vertexColors)
		{
			DisposePtr((Ptr)gMaterialTypeList[i].vertexColors);
			gMaterialTypeList[i].vertexColors = nil;
		}
	}
	gNumMaterialTypes = 0;	
}


/****************** OPTIMIZE GEOMETRY ***********************/
//
// This is the basic call which optimizes the current 3DMF file.
// The geometry is in gModelGroup
//

void OptimizeGeometry(void)
{
TQ3Matrix4x4	m;
TQ3BoundingBox	boundingBox;

	if (gModelObjNode == nil)													// make sure there's something to optimize
		return;
	
				/*******************/
				/* INIT SOME STUFF */
				/*******************/

	OptimizationInit();
	
	
		/***********************************************************/
		/* RECURSE THRU GEOMETRY WHICH CREATES LISTS OF TRI-MESHES */
		/***********************************************************/
			
				/* INIT ACCUMULATION STUFF */

	Q3Matrix4x4_SetScale(&gWorkMatrix, gScale, gScale, gScale);				

	if (gAutoCenter)
	{
		Q3Matrix4x4_SetTranslate(&m, gAutoCenterXOff, gAutoCenterYOff, gAutoCenterZOff);
		Q3Matrix4x4_Multiply(&m, &gWorkMatrix, &gWorkMatrix);
	}

	gWorkAttribSet = Q3AttributeSet_New();										// init empty attrib set
	gWorkShader = nil;															// init nil shader
	
				/* RECURSE IT */
				
	RecurseGeometry(gModelGroup);												// recursivly parse source geometry (pass base group)

				
			/* BUILD TRIMESHES */
				
	OptimizeTriangleListTextures();												// verify and optimize all textures
	ConvertTriangleListsToTriMeshes();											// convert triangle lists into tri-mesh objects






			/* CHECK FOR ERRORS */
			
	if (gFoundBadGeneralPolygon)
		DoAlert("\pThere was at least 1 general polygon in the model which was too complex to be converted into TriMesh - it was removed.");


		/* PUT ALL TRI-MESHES INTO A NEW DISPLAY GROUP */

	BuildDisplayModel();
	
	gCurrentIsOptimized = true;
	gScale = 1.0;
	
	
				/* RECALC AUTO-CENTER OFFSET */

	QD3D_CalcObjectBoundingBox(&gModelViewInfo, gModelGroup, &boundingBox);
	gAutoCenterXOff = -(boundingBox.max.x + boundingBox.min.x) / 2.0f;		// calc center X
	gAutoCenterZOff = -(boundingBox.max.z + boundingBox.min.z) / 2.0f;		// calc center Z
	if (gCenterYAtZero)
		gAutoCenterYOff = -boundingBox.min.y;								// keey bottom/y @ y=0
	else
		gAutoCenterYOff = -(boundingBox.max.y + boundingBox.min.y) / 2.0f;	// calc center Y
	
}


/********************* BUILD DISPLAY MODEL ***********************/

void BuildDisplayModel(void)
{
long	i;
ObjNode	*newObj;
float	scale;
TQ3GroupPosition	position;
TQ3DisplayGroupState	state;

	if (gModelGroup)
		Q3Object_Dispose(gModelGroup);												// nuke old geometry
		
	gModelGroup = Q3DisplayGroup_New();											// make new group
	if (gModelGroup == nil)
		DoFatalAlert("\pOptimizeGeometry: Q3DisplayGroup_New failed!");

	for (i=0; i < gNumTriMeshes; i++)											// add each trimesh to the group
	{
		position = Q3Group_AddObject(gModelGroup, gTriMeshObjects[i]);
		if (position == nil)
			DoFatalAlert("\pOptimizeGeometry: Q3Group_AddObject failed!");
	}
	
	Q3DisplayGroup_GetState(gModelGroup,&state);								// set group as in-line
	state |= kQ3DisplayGroupStateMaskIsInline;
	Q3DisplayGroup_SetState(gModelGroup,state);
	
	if (gUseCullingGroup)														// see if should calc group bbox
	{
		Q3DisplayGroup_CalcAndUseBoundingBox(gModelGroup, kQ3ComputeBoundsExact, gModelViewInfo.viewObject);
	}
		
	
			/* SET NEW GROUP AS DISPLAYED OBJNODE */
			
	DeleteObject(gModelObjNode);														// delete old ObjNode
	
	gNewObjectDefinition.genre = DISPLAY_GROUP_GENRE;									// make new one
	gNewObjectDefinition.coord.x = 0;
	gNewObjectDefinition.coord.y = 0;
	gNewObjectDefinition.coord.z = 0;
	gNewObjectDefinition.flags = OBJ_MODE_DRAW;
	gNewObjectDefinition.slot = 100;
	gNewObjectDefinition.moveCall = MoveModel;
	newObj = MakeNewDisplayGroupObject(&gNewObjectDefinition);

	gModelObjNode = newObj;																// remember this
	scale = CalcScaleToMakeFit(&gModelViewInfo,gModelGroup, TARGET_SIZE);				// calc scale 
	newObj->ScaleX = newObj->ScaleY = newObj->ScaleZ = scale;

	AttachGeometryToDisplayGroupObject(newObj,gModelGroup);


			/* BUILD BASIC MATRIX */
			
	UpdateObjectTransforms(newObj);

	gTotalMeshes = gNumTriMeshes;
	DrawModelWindowHeader();									// update stats

}


/****************** RECURSE GEOMETRY ***********************/
//
// Recursive call which takes current TQ3Object and recurses it while executing any relevant objects.
//

static void RecurseGeometry(TQ3Object obj)
{
TQ3Matrix4x4		transform;
TQ3GroupPosition	position;
TQ3Object   		object,stashShader;
TQ3Matrix4x4  		stashMatrix;
TQ3AttributeSet		stashAttribSet;
TQ3ObjectType		oType;


				/*******************************/
				/* SEE IF ACCUMULATE TRANSFORM */
				/*******************************/
				
	if (Q3Object_IsType(obj,kQ3ShapeTypeTransform))
	{
  		Q3Transform_GetMatrix(obj,&transform);
		Q3Matrix4x4_Multiply(&transform,&gWorkMatrix,&gWorkMatrix);
 	}
	else
				/*************************/
				/* SEE IF FOUND GEOMETRY */
				/*************************/

	if (Q3Object_IsType(obj,kQ3ShapeTypeGeometry))
	{
		oType = Q3Geometry_GetType(obj);						// get geometry type
		

		switch(oType)
		{
						/* TRIANGLE */
						
			case	kQ3GeometryTypeTriangle:
					PreTransformTri(obj,nil);
					AddTriToTriangleLists(obj,nil);
					break;
					
					
			default:
					DecomposeGeometry(obj);
					
		}
	}
	else
				/**************************/
				/* SEE IF FOUND ATTRIBUTE */
				/**************************/

	if (Q3Object_IsType(obj,kQ3SetTypeAttribute))
	{
		RemoveUnwantedAttribs(obj);
		Q3AttributeSet_Inherit(gWorkAttribSet, obj, gWorkAttribSet);					// add obj's attribs to gWorkAttribSet
	}
	else
				/***********************/
				/* SEE IF FOUND SHADER */
				/***********************/

	if (Q3Object_IsType(obj,kQ3ShapeTypeShader))
	{	
		if (Q3Shader_GetType(obj) == kQ3ShaderTypeSurface)								// must be texture surface shader
		{
			gWorkShader = obj;
			Q3AttributeSet_Add(gWorkAttribSet, kQ3AttributeTypeSurfaceShader, &gWorkShader);	// add shader to current gWorkAttribSet
		}
	}
	else
	
			/****************************/
			/* SEE IF RECURSE SUB-GROUP */
			/****************************/

	if (Q3Object_IsType(obj,kQ3ShapeTypeGroup))
 	{
  		stashMatrix 	= gWorkMatrix;													// push matrix
  		stashAttribSet	= QD3D_MakeDuplicateAttributeSet(gWorkAttribSet);				// push attribute set
  		stashShader 	= gWorkShader;													// push shader
  		
  		Q3Group_GetFirstPosition(obj, &position);										// get 1st object in group
  		while(position != nil)															// scan all objects in group
 		{
   			Q3Group_GetPositionObject (obj, position, &object);							// get object from group
			if (object != nil)
   			{
   				/* FILTER OUT UNWANTED OBJECTS */
   				
				if (Q3Object_IsType(object,kQ3ObjectTypeView))							// filter out View Objects
				{
					Q3Object_Dispose(Q3Group_RemovePosition(obj, position));
				}
					/* RECURSE THIS OBJ */
					
				else
				{				
	    			RecurseGeometry(object);											// recurse this object
    			}
    			Q3Object_Dispose(object);												// dispose local ref
   			}
   			Q3Group_GetNextPosition(obj, &position);									// get next object in group
  		}
  		gWorkMatrix = stashMatrix;														// pop matrix
  		Q3Object_Dispose(gWorkAttribSet);												// pop attribute set
  		gWorkAttribSet = stashAttribSet;
  		gWorkShader = stashShader;														// pop shader
	}
}





/*************************** PRETRANSFORM TRIANGLE GEOMETRY **********************************/
//
// If theTriangle == nil, then immediate data is in inData.
//

void PreTransformTri(TQ3GeometryObject theTriangle, TQ3TriangleData *triDataPtr)
{
TQ3TriangleData	triData;
TQ3Status 		status;
Boolean			gotData = false;

				/* GET TRIANGLE DATA */
				
	if (triDataPtr == nil)
	{
		status = Q3Triangle_GetData(theTriangle, &triData);
		if (status != kQ3Success)
			DoFatalAlert("\pPreTransformTri: Q3Triangle_GetData failed!");
		gotData = true;
		triDataPtr = &triData;								// point to it
	}

			/* TRANSFORM VERTICES */
			
	Q3Point3D_Transform(&triDataPtr->vertices[0].point, &gWorkMatrix, &triDataPtr->vertices[0].point);				
	Q3Point3D_Transform(&triDataPtr->vertices[1].point, &gWorkMatrix, &triDataPtr->vertices[1].point);
	Q3Point3D_Transform(&triDataPtr->vertices[2].point, &gWorkMatrix, &triDataPtr->vertices[2].point);


			/* TRANSFORM ATTRIBUTES */
				
	if (triDataPtr->triangleAttributeSet)
		PreTransformAttribute(triDataPtr->triangleAttributeSet);


			/* UPDATE THE TRIANGLE */
			
	if (theTriangle)
	{
		status = Q3Triangle_SetData(theTriangle, triDataPtr);
		if (status != kQ3Success)
			DoFatalAlert("\pPreTransformTri: Q3Triangle_SetData failed!");
	}

	if (gotData)
		Q3Triangle_EmptyData(triDataPtr);												// nuke it
}



/****************** PRETRANSFORM ATTRIBUTE *********************/
//
// Applies gWorkMatrix to any applicable attributes such as tangents and normals.
//

static void PreTransformAttribute(TQ3AttributeSet theAttribute)
{
TQ3Status			status;
TQ3AttributeType	attribType;
TQ3Vector3D			vector;
TQ3Tangent2D		tangent;

	if (theAttribute == nil)
		return;

				/* GET 1ST ATTRIBUTE TYPE */
				
	attribType = kQ3AttributeTypeNone;
	status = Q3AttributeSet_GetNextAttributeType(theAttribute, &attribType);

			/* ITERATE THRU ALL ATTRIBS LOOKING FOR SOMETHING TO ROTATE */
				
	if ((status == kQ3Success) && (attribType != kQ3AttributeTypeNone))
	{
		do
		{
			switch(attribType)
			{
				case	kQ3AttributeTypeNormal:
						Q3AttributeSet_Get(theAttribute, attribType, &vector);	
						if ((vector.x == 0) && (vector.y == 0) && (vector.z == 0))		// error check for illegal vectors (from FormZ) & default to 1 if bad
						{
							vector.x = 1;						
						}
						else
							Q3Vector3D_Transform(&vector, &gWorkMatrix, &vector);	
						Q3Vector3D_Normalize(&vector,&vector);
						if(fabs(1.0 - (vector.x * vector.x + vector.y * vector.y + vector.z * vector.z)) > 0.00001)
						{
							vector.x = 1.0;						
							vector.y = 0.0;						
							vector.z = 0.0;						
						}
						
						status = Q3AttributeSet_Add(theAttribute, attribType, &vector);
						if (status != kQ3Success)
							DoFatalAlert("\pQ3AttributeSet_Add failed!");
						break;	

				case	kQ3AttributeTypeSurfaceTangent:
						Q3AttributeSet_Get(theAttribute, attribType, &tangent);
						Q3Vector3D_Transform(&tangent.uTangent, &gWorkMatrix, &tangent.uTangent);				
						Q3Vector3D_Transform(&tangent.vTangent, &gWorkMatrix, &tangent.vTangent);				
						status = Q3AttributeSet_Add(theAttribute, attribType, &tangent);
						if (status != kQ3Success)
							DoFatalAlert("\pQ3AttributeSet_Add failed!");
						break;	
						
												
			}
			status = Q3AttributeSet_GetNextAttributeType(theAttribute, &attribType);	// get next attrib type
			if (status != kQ3Success)
				break;
		}while(attribType != kQ3AttributeTypeNone);
	}
}


/*********************** REMOVE UNWANTED ATTRIBUTES *******************/

static void RemoveUnwantedAttribs(TQ3AttributeSet attribs)
{

	if (attribs == nil)
		return;
		
				/* REMOVE HIGHLIGH INFO */
				
	Q3AttributeSet_Clear(attribs,  kQ3AttributeTypeHighlightState);		
		
}






/************************** ADD TRIANGLE GEOMETRY TO TRIANGLE LISTS ***********************************/
//
// If theTri == nil, then data is in inData.
//

void AddTriToTriangleLists(TQ3Object theTri, TQ3TriangleData *triDataPtr)
{
TQ3TriangleData	data;
TQ3Status 		status;
TQ3Point3D			coords[3];
TQ3Vector3D			normals[3],faceNormal;
TQ3Param2D			uvs[3];
MaterialType		materialType;
long				i;
TQ3AttributeSet		compositeAttribSet;
TQ3ColorRGB			vertexColors[3];


				/* GET TRIANGLE DATA */
		
	if (theTri)
	{		
		status = Q3Triangle_GetData(theTri, &data);
		if (status != kQ3Success)
			DoFatalAlert("\pAddTriToTriangleLists: Q3Triangle_GetData failed!");
			
		triDataPtr = &data;
	}

	
			/* INHERIT ATTRIBS */

	RemoveUnwantedAttribs(gWorkAttribSet);
	compositeAttribSet = QD3D_MakeDuplicateAttributeSet(gWorkAttribSet);							// start with these attribs
	if (triDataPtr->triangleAttributeSet)
	{
		RemoveUnwantedAttribs(triDataPtr->triangleAttributeSet);
		Q3AttributeSet_Inherit(compositeAttribSet, triDataPtr->triangleAttributeSet, compositeAttribSet);	// add to composite attribs
	}

	
				/*****************/
				/* MAKE MATERIAL */
				/*****************/
					
	ConvertAttributeSetToMaterialType(compositeAttribSet,&materialType);							// calc material	
	
	Q3Object_Dispose(compositeAttribSet);															// dispose attrib set ref


		/* GET SHADING UV ATTRIBUTES FROM VERTICES (IF ANY) */
				
	for (i=0; i < 3; i++)
	{
		status = Q3AttributeSet_Get(triDataPtr->vertices[i].attributeSet, kQ3AttributeTypeShadingUV, &uvs[i]);
		if (status == kQ3Failure)
			Q3AttributeSet_Get(triDataPtr->vertices[i].attributeSet, kQ3AttributeTypeSurfaceUV, &uvs[i]);
	}
	
			/* GET COORDS & CALC FACE NORMAL */
	
	coords[0] = triDataPtr->vertices[0].point;
	coords[1] = triDataPtr->vertices[1].point;
	coords[2] = triDataPtr->vertices[2].point;
	Q3Point3D_CrossProductTri(&coords[0], &coords[1], &coords[2], &faceNormal);


			/* GET VERTEX NORMAL ATTRIBUTES (IF ANY) */
			
	for (i=0; i < 3; i++)
	{
		status = Q3AttributeSet_Get(triDataPtr->vertices[i].attributeSet, kQ3AttributeTypeNormal, &normals[i]);
		if (status == kQ3Failure)
			normals[i] = faceNormal;
	}


			/* GET VERTEX COLOR ATTRIBUTES (IF ANY) */
			
	if (gKeepVertexColors)
	{
		for (i=0; i < 3; i++)
		{
			status = Q3AttributeSet_Get(triDataPtr->vertices[i].attributeSet, kQ3AttributeTypeDiffuseColor, &vertexColors[i]);
			if (status == kQ3Success)
				materialType.hasVertexColors = true;
		}
	}	
	
				/* ADD TRIANGLE TO LIST */
				
	AddTriangleToTriangleList(&coords[0],&normals[0],&faceNormal,&uvs[0],&materialType, &vertexColors[0]);



			/* CLEANUP */

	if (theTri)
		Q3Triangle_EmptyData(triDataPtr);
}






/********************** CONVERT ATTRIBUTE SET TO MATERIAL TYPE ***************************/
//
// Returns a material type definition based on the input attribute values.
//

static void ConvertAttributeSetToMaterialType(TQ3AttributeSet compositeAttribSet,MaterialType *materialType)
{
TQ3Status				status;
TQ3ColorRGB				color,defaultColor = {1,1,1};
TQ3SurfaceShaderObject	shader;


			/* ATTEMPT TO EXTRACT SURFACE SHADER INFO */

	status = Q3AttributeSet_Get(compositeAttribSet, kQ3AttributeTypeSurfaceShader, &shader);	
	if (status == kQ3Success)
	{
		materialType->isTextured = true;
		materialType->textureRef = shader;
	}
	
			/* ATTEMPT TO EXTRACT COLOR INFO */
	else
	{		
		materialType->isTextured = false;
		
		status = Q3AttributeSet_Get(compositeAttribSet, kQ3AttributeTypeDiffuseColor, &color);	
		if (status == kQ3Success)
			materialType->color = color;
		
				/* NOT COLORED NOR TEXTURED, SO APPLY DEFAULT COLOR */
		else
			materialType->color = defaultColor;	
	}
	
			/* ATTEMPT TO EXTRACT TRANSPARENCY INFO */
			
	status = Q3AttributeSet_Get(compositeAttribSet, kQ3AttributeTypeTransparencyColor, &color);	
	if (status == kQ3Success)
	{
		materialType->isTransparent = true;
		materialType->transColor = color;
	}
	else
		materialType->isTransparent = false;
		
		
			/* FOR NOW, ASSUME NO VERTEX COLORS */
			
	materialType->hasVertexColors = false;
		
}



/********************** ADD TRIANGLE TO TRIANGLE LIST *********************************/
//
// This takes raw triangle data and inserts it into the appropriate triangle list, or if none appropriate
// exists, then it creates a new one.
//

void AddTriangleToTriangleList(TQ3Point3D *coords, TQ3Vector3D *vertexNormals, TQ3Vector3D *faceNormal,
								TQ3Param2D *vertexUVs, MaterialType *materialType, TQ3ColorRGB *vertexColors)
{
long	materialNum;
long	triangleNum;


			/*******************************************/
			/* SEARCH FOR EXISTING LIST TO APPEND ONTO */
			/*******************************************/

	for (materialNum = 0; materialNum < gNumMaterialTypes; materialNum++)
	{
				/* CHECK VERTEX COLOR FLAG */
			
		if (materialType->hasVertexColors != gMaterialTypeList[materialNum].hasVertexColors)	// see if both true or false
			continue;
	
	
				/* CHECK TRANSPARENCY SIMILARITY */
				
		if (materialType->isTransparent != gMaterialTypeList[materialNum].isTransparent)		// see if both same on/off
			continue;

		if (materialType->isTransparent)
		{
			if ((materialType->transColor.r != gMaterialTypeList[materialNum].transColor.r) || 	// matching RGB value
				(materialType->transColor.g != gMaterialTypeList[materialNum].transColor.g) ||
				(materialType->transColor.b != gMaterialTypeList[materialNum].transColor.b))
				continue;
		}	
			
				/* TEXTURED */
				
		if (materialType->isTextured)
		{
			if (gMaterialTypeList[materialNum].isTextured)										// matching type
			{
				if (materialType->textureRef == gMaterialTypeList[materialNum].textureRef)		// matching texture shader
				{
					goto add_it;
				}
			}
		}
				/* COLORED */
		else
		{
			if (!gMaterialTypeList[materialNum].isTextured)										// matching type
			{
				if ((materialType->color.r == gMaterialTypeList[materialNum].color.r) && 		// matching RGB value
					(materialType->color.g == gMaterialTypeList[materialNum].color.g) &&
					(materialType->color.b == gMaterialTypeList[materialNum].color.b))
				{
					goto	add_it;
				}
			}
		}
	}

			/*********************************************************/
			/* NO MATCH, SO CREATE NEW MATERIAL TYPE & TRIANGLE LIST */
			/*********************************************************/

			/* SHOW TEXTURE IN WINDOW */
			
//	if (materialType->isTextured)
//	{
//		win = NewCWindow(nil,&r,"\p",true,noGrowDocProc,(WindowPtr)-1L,false,0L);
//	
//	DumpGWorld(gTextureList[gNumTextures].gworld, gTextureList[gNumTextures].window);
//	}

	materialNum = gNumMaterialTypes++;
	gMaterialTypeList[materialNum] = *materialType;												// copy material def into list
	gMaterialTypeList[materialNum].merged = false;												// not merged yet
	
	gMaterialTypeList[materialNum].triangleList = (TriangleDefType *)AllocPtr(sizeof(TriangleDefType)*gMaxTrianglesPerMaterial);
	if (gMaterialTypeList[materialNum].triangleList == nil)
		DoFatalAlert("\pAddTriangleToTriangleList: Out of Memory (triangleList)!");
	
	
			/* ALLOC VERTEX COLOR ARRAY IF NEEDED */
				
	if (materialType->hasVertexColors)
	{
		gMaterialTypeList[materialNum].vertexColors = (TriangleVertexColorsType *)AllocPtr(sizeof(TriangleVertexColorsType)*gMaxTrianglesPerMaterial);
		if (gMaterialTypeList[materialNum].vertexColors == nil)
			DoFatalAlert("\pAddTriangleToTriangleList: Out of Memory (vertexColors)!");
	}
	else
	{
		gMaterialTypeList[materialNum].vertexColors = nil;	
	}
		
				
	gMaterialTypeList[materialNum].numTrianglesInList = 0;
	

			/******************************/
			/* ADD IT TO THE CURRENT LIST */
			/******************************/
add_it:

	triangleNum = gMaterialTypeList[materialNum].numTrianglesInList;								// get # triangles in list

	if (triangleNum >= gMaxTrianglesPerMaterial)													// check for overflow
		DoFatalAlert("\pMAX TRIANGLES PER MATERIAL exceeded!");

	gMaterialTypeList[materialNum].triangleList[triangleNum].vertexCoords[0] = coords[0];			// set vertex coords
	gMaterialTypeList[materialNum].triangleList[triangleNum].vertexCoords[1] = coords[1];
	gMaterialTypeList[materialNum].triangleList[triangleNum].vertexCoords[2] = coords[2];

	Q3Vector3D_Normalize(&vertexNormals[0],&vertexNormals[0]);										// make sure these are normalized
	Q3Vector3D_Normalize(&vertexNormals[1],&vertexNormals[1]);
	Q3Vector3D_Normalize(&vertexNormals[2],&vertexNormals[2]);

	gMaterialTypeList[materialNum].triangleList[triangleNum].vertexNormals[0] = vertexNormals[0];	// set vertex normals
	gMaterialTypeList[materialNum].triangleList[triangleNum].vertexNormals[1] = vertexNormals[1];
	gMaterialTypeList[materialNum].triangleList[triangleNum].vertexNormals[2] = vertexNormals[2];

	gMaterialTypeList[materialNum].triangleList[triangleNum].vertexUVs[0] = vertexUVs[0];			// set vertex UVs
	gMaterialTypeList[materialNum].triangleList[triangleNum].vertexUVs[1] = vertexUVs[1];
	gMaterialTypeList[materialNum].triangleList[triangleNum].vertexUVs[2] = vertexUVs[2];

	gMaterialTypeList[materialNum].triangleList[triangleNum].faceNormal = *faceNormal;				// set face normal


					/* ADD VERTEX COLORS */
					
	if (gMaterialTypeList[materialNum].hasVertexColors)
	{
		gMaterialTypeList[materialNum].vertexColors[triangleNum].color[0] = vertexColors[0];		// set vertex Colors
		gMaterialTypeList[materialNum].vertexColors[triangleNum].color[1] = vertexColors[1];
		gMaterialTypeList[materialNum].vertexColors[triangleNum].color[2] = vertexColors[2];
	}


	gMaterialTypeList[materialNum].numTrianglesInList++;											// inc count
}


/******************** CONVERT TRIANGLE LISTS TO TRI-MESHES ****************************/
//
// Scans thru gMaterialTypeList and converts each triangle list into a tri-mesh object.
//

void ConvertTriangleListsToTriMeshes(void)
{
long				i,j,v;
TQ3TriMeshData		triMeshData;
TQ3TriMeshTriangleData	*triangles;
TQ3TriMeshAttributeData	triangleAttributes,vertexAttributes[3];
TQ3Point3D			*points,thisPoint;
long				numPoints,numTriangles,k;
TQ3Vector3D			*faceNormals,*vertexNormals,thisVertNormal;
TQ3Param2D			*uvs,thisVertUV;
long				showMeCount;
TQ3ColorRGB			*vertexColors, thisVertColor;

	showMeCount = 0;
	gNumTriMeshes = 0;
	
	TextMode(srcCopy);
	
				/********************************************/
				/* CONVERT EACH LIST INTO A TRI-MESH OBJECT */
				/********************************************/
				//
				// Note: each material type is 1 tri-mesh.
				//
				
	for (i = 0; i < gNumMaterialTypes; i++)
	{
		numTriangles =  gMaterialTypeList[i].numTrianglesInList;								// get # triangles
	
						/* ALLOC BUFFERS */
						
		points = (TQ3Point3D *)AllocPtr(sizeof(TQ3Point3D)*numTriangles*3);						// alloc worst-case vertex/points list buffer
		if (points == nil)
			DoFatalAlert("\pConvertTriangleListsToTriMeshes: Out of Memory! (points)");

		vertexNormals = (TQ3Vector3D *)AllocPtr(sizeof(TQ3Vector3D)*numTriangles*3);			// alloc worst-case vertex normal buffer
		if (vertexNormals == nil)
			DoFatalAlert("\pConvertTriangleListsToTriMeshes: Out of Memory! (vertexNormals)");

		vertexColors = (TQ3ColorRGB *)AllocPtr(sizeof(TQ3ColorRGB)*numTriangles*3);				// alloc worst-case vertex color buffer
		if (vertexColors == nil)
			DoFatalAlert("\pConvertTriangleListsToTriMeshes: Out of Memory! (vertexColors)");

		triangles = (TQ3TriMeshTriangleData *)AllocPtr(sizeof(TQ3TriMeshTriangleData)*numTriangles);	// alloc triangle list buffer
		if (triangles == nil)
			DoFatalAlert("\pConvertTriangleListsToTriMeshes: Out of Memory! (triangles)");

		faceNormals = (TQ3Vector3D *)AllocPtr(sizeof(TQ3Vector3D)*numTriangles);				// alloc face normal buffer
		if (faceNormals == nil)
			DoFatalAlert("\pConvertTriangleListsToTriMeshes: Out of Memory! (faceNormals)");

		uvs = (TQ3Param2D *)AllocPtr(sizeof(TQ3Param2D)*numTriangles*3);						// alloc worst-case vertex uv buffer
		if (uvs == nil)
			DoFatalAlert("\pConvertTriangleListsToTriMeshes: Out of Memory! (uv)");



					/*********************************/
					/* SET ALL TRIANGLE & POINT INFO */
					/*********************************/
					//
					// while adding vertices, it checks for duplicates already in list.
					//

		numPoints = 0;	
		for (j=0; j < numTriangles; j++)
		{
			TQ3Vector3D	thisFaceNormal;
			
			if (gAverageNormals)
			{
				QD3D_CalcFaceNormal(&gMaterialTypeList[i].triangleList[j].vertexCoords[0],
									&gMaterialTypeList[i].triangleList[j].vertexCoords[1],
									&gMaterialTypeList[i].triangleList[j].vertexCoords[2],
									&thisFaceNormal);
			}
		
			for (v = 0; v < 3; v++)																				// do all 3 verts in this triangle/face
			{
				thisPoint 		= gMaterialTypeList[i].triangleList[j].vertexCoords[v];							// get vert coord
				thisVertUV 		= gMaterialTypeList[i].triangleList[j].vertexUVs[v];							// get vert UV
				if (gMaterialTypeList[i].hasVertexColors)
					thisVertColor	= gMaterialTypeList[i].vertexColors[j].color[v];							// get vert color
				
				if (gAverageNormals)
					thisVertNormal = thisFaceNormal;
				else
					thisVertNormal = gMaterialTypeList[i].triangleList[j].vertexNormals[v];						// get vert normal
				
				
						/**********************************/
						/* LOOK FOR EXACT MATCHING VERTEX */
						/**********************************/
						
				for (k = 0; k < numPoints; k++)														
				{
					if (PointsAreCloseEnough(&points[k],&thisPoint))											// see if this point already in list
					{
						if (VectorsAreCloseEnough(&vertexNormals[k],&thisVertNormal) || (!gUseVertexNormals))	// vertex normals must match also if we want them
						{
ok_normals:					if ((!gMaterialTypeList[i].isTextured) || UVsAreCloseEnough(&uvs[k],&thisVertUV))	// also see if UV's match (textured only)
							{
								if (!gMaterialTypeList[i].hasVertexColors)										// see if need to check vertex colors
									goto found_dupe;
			
								if (ColorsAreCloseEnough(&vertexColors[k],&thisVertColor))						// also see if vertex colors match
								{
found_dupe:							triangles[j].pointIndices[v] = k;											// FOUND DUPE, so use it
									vertexNormals[k] = thisVertNormal;											// update normal (if was averaged)
									goto next_v;
								}
							}
						}
						else
						if (gAverageNormals)																	// see if average normals
						{
							thisVertNormal.x += vertexNormals[k].x;
							thisVertNormal.y += vertexNormals[k].y;
							thisVertNormal.z += vertexNormals[k].z;
							Q3Vector3D_Normalize(&thisVertNormal,&thisVertNormal);
							goto ok_normals;
						}
					}
				}		
				
						/*****************************/
						/* ADD NEW VERTEX TO TRIMESH */
						/*****************************/
							
				points[numPoints] 			= thisPoint;									// none found, so add new vertex to point list
				vertexNormals[numPoints] 	= thisVertNormal;								// also build vertex normal list
				uvs[numPoints] 				= thisVertUV;									// also build vertex uv list
				triangles[j].pointIndices[v] 	= numPoints;								// assign to triangle
				vertexColors[numPoints] 	= thisVertColor;								// also build vertex colors list
				
				numPoints++;																// inc count
next_v:;				
			}
			
			showMeCount++;
			if ((showMeCount & 0xff) == 0)
			{
				Str32	s;
				MoveTo(10,10);
				NumToString(showMeCount, s);
				DrawString(s);
			}
		}	// for numTriangles
		
		
				/* SET FINAL TRIMESH DATA */
				
		triMeshData.numPoints = numPoints;													// set nPoints
		triMeshData.points = points;														// point to vertex list
		triMeshData.numTriangles = numTriangles;												// set nTriangles
		triMeshData.triangles = triangles;													// point to triangle list
			

					/*********************************************/
					/* CALC BOUNDING BOX MIN/MAX FROM POINT LIST */
					/*********************************************/
			
		Q3BoundingBox_SetFromPoints3D(&triMeshData.bBox, &points[0], numPoints, sizeof(TQ3Point3D));

			
					/******************************************/
					/* SET ALL TRIANGLE FACE NORMAL ATTRIBUTS */
					/******************************************/

		if (gUseFaceNormals)
		{
			triMeshData.numTriangleAttributeTypes = 1;												// # attribute types
			triMeshData.triangleAttributeTypes = &triangleAttributes;								// point to attribute list

			for (j=0; j < numTriangles; j++)
			{
			
				QD3D_CalcFaceNormal(&gMaterialTypeList[i].triangleList[j].vertexCoords[0],			// calc new face normal (don't use one in material list)
									&gMaterialTypeList[i].triangleList[j].vertexCoords[1],
									&gMaterialTypeList[i].triangleList[j].vertexCoords[2],
									&faceNormals[j]);
			
			}
			
			triangleAttributes.attributeType = kQ3AttributeTypeNormal;								// set attribute Type
			triangleAttributes.data = faceNormals;													// point to attribute data
			triangleAttributes.attributeUseArray = nil;												// (not used)
		}
		else
		{
			triMeshData.numTriangleAttributeTypes = 0;												// no face normal attrib
			triMeshData.triangleAttributeTypes = nil;	
		}		

					/**********************************/
					/* SET ALL VERTEX/POINT ATTRIBUTS */
					/**********************************/


		if (gMaterialTypeList[i].isTextured)
		{
			triMeshData.numVertexAttributeTypes = 2;												// set # vertex attrib types (normals & UV)
			triMeshData.vertexAttributeTypes = &vertexAttributes[0];							// point to vertex attrib list
	
			vertexAttributes[0].attributeType = kQ3AttributeTypeNormal;							// set attrib type == normal
			vertexAttributes[0].data = vertexNormals;											// point to vertex normals
			vertexAttributes[0].attributeUseArray = nil;										// (not used)

			vertexAttributes[1].attributeType = kQ3AttributeTypeShadingUV;						// set attrib type == shading UV
			vertexAttributes[1].data = uvs;														// point to vertex UV's
			vertexAttributes[1].attributeUseArray = nil;										// (not used)		
		}
		else
		{
			triMeshData.numVertexAttributeTypes = 1;												// set # vertex attrib types (normals)
			triMeshData.vertexAttributeTypes = &vertexAttributes[0];							// point to vertex attrib list
		
			vertexAttributes[0].attributeType = kQ3AttributeTypeNormal;							// set attrib type == normal
			vertexAttributes[0].data = vertexNormals;											// point to vertex normals
			vertexAttributes[0].attributeUseArray = nil;										// (not used)
		}	
		
		
					/* SEE IF ALSO USE VERTEX COLORS */
					
		if (gMaterialTypeList[i].hasVertexColors)
		{
			int	m;
			
			m = triMeshData.numVertexAttributeTypes++;											// get index & increment num
			
			vertexAttributes[m].attributeType = kQ3AttributeTypeDiffuseColor;					// set attrib type == diffuse color
			vertexAttributes[m].data = vertexColors;											// point to vertex colors
			vertexAttributes[m].attributeUseArray = nil;										// (not used)
		}	
		

			/*********************************/
			/* SETUP REMAINING TRI-MESH DATA */
			/*********************************/

		triMeshData.triMeshAttributeSet = CreateAttributeSetFromMaterial(&gMaterialTypeList[i]);// assign material attribs
				
		triMeshData.numEdges = 0;																	// set # edges
		triMeshData.edges = nil;																// point to edge list
		triMeshData.numEdgeAttributeTypes = 0;													// set # edge attrib types
		triMeshData.edgeAttributeTypes = nil;													// point to edge attrib list	
	
	
				/* MAKE THE TRI-MESH */
						
		gTriMeshObjects[i] = Q3TriMesh_New(&triMeshData);										// create the tri-mesh
		if (gTriMeshObjects[i] == nil)
		{
			DoAlert("\pConvertTriangleListsToTriMeshes: Q3TriMesh_New failed!");
			QD3D_ShowRecentError();
		}
		
			/* OPTIMIZE FOR STRIPS */

		if (gMakeStripsFlag)
			StripifyTriMesh(gTriMeshObjects[i]);
		
		
		gNumTriMeshes++;																		// inc count
	
				/* CLEANUP */
					
		DisposePtr((Ptr)points);																// free buffers
		DisposePtr((Ptr)triangles);
		DisposePtr((Ptr)faceNormals);
		DisposePtr((Ptr)vertexNormals);
		DisposePtr((Ptr)uvs);
		DisposePtr((Ptr)vertexColors);
			
		gTotalNumVertecies += numPoints;	
		gTotalNumFaces += numTriangles;		
	}
}


/************************** CREATE ATTRIBUTE SET FROM MATERIAL ********************************/
//
// Given a material definition, create it's equivaluent attribute set.
//

static TQ3AttributeSet CreateAttributeSetFromMaterial(MaterialType *material)
{
TQ3AttributeSet	attributeSet;
TQ3Status		status;

	attributeSet = Q3AttributeSet_New();
	if (attributeSet == nil)
		DoFatalAlert("\pCreateAttributeSetFromMaterial: Q3AttributeSet_New failed!");
	
				/* DEAL WITH TEXTURES & COLORS */
				
	if (material->isTextured)
	{
		status = Q3AttributeSet_Add(attributeSet, kQ3AttributeTypeSurfaceShader, &material->textureRef);
		if (status != kQ3Success)
			DoFatalAlert("\pCreateAttributeSetFromMaterial: Q3AttributeSet_Add failed!");
	}
	else
	{
		status = Q3AttributeSet_Add(attributeSet, kQ3AttributeTypeDiffuseColor, &material->color);
		if (status != kQ3Success)
			DoFatalAlert("\pCreateAttributeSetFromMaterial: Q3AttributeSet_Add (color) failed!");
	}
	
			/* DEAL WITH TRANSPARENCY */
			
	if (material->isTransparent)
	{
		status = Q3AttributeSet_Add(attributeSet, kQ3AttributeTypeTransparencyColor, &material->transColor);
		if (status != kQ3Success)
			DoFatalAlert("\pCreateAttributeSetFromMaterial: Q3AttributeSet_Add failed!");
	}

	return(attributeSet);
}


/*************************** OPTIMIZE TRIANGLE LIST TEXTURES ************************************/

static void OptimizeTriangleListTextures(void)
{
long					materialNum;
TQ3SurfaceShaderObject	surface;

			/* OPTIMIZE AND REMOVE DUPLICATE TEXTURES */
			
	for (materialNum = 0; materialNum < gNumMaterialTypes; materialNum++)								// scan thru all materials...
	{
		if (gMaterialTypeList[materialNum].isTextured)													// ... looking for textured materials
		{
			surface = gMaterialTypeList[materialNum].textureRef;										// get ref to surface shader
			gMaterialTypeList[materialNum].textureRef = OptimizeTexture(surface);
		}
	}
}


/************************** STRIP TRIMESH ********************************/
//
// Optimize input trimesh to have strips
//

static void StripifyTriMesh(TQ3Object triMesh)
{
TQ3TriMeshData			triMeshData;
TQ3TriMeshTriangleData	*triangles;
TQ3Vector3D				*faceNormals;
long					i,numTriangles;

			/* GET THE TRIMESH'S DATA */
			
	Q3TriMesh_GetData(triMesh, &triMeshData);
	numTriangles = triMeshData.numTriangles;									// get # triangles
	triangles = triMeshData.triangles;											// get ptr to triangle list

	if (triMeshData.numTriangleAttributeTypes)									// see if has face normals
		faceNormals = triMeshData.triangleAttributeTypes[0].data;				// point to face normals
	else
		goto bail;
		

			/**********************************************/
			/* DO TRAVERSAL MULTIPLE TIMES FOR REFINEMENT */
			/**********************************************/
			
	
	for (i = 0; i < 8; i++)
	{
		gNumStripBreaks = 0;
		Stripify_Backward(numTriangles, triangles, faceNormals);
		gNumStripBreaks = 0;
		Stripify_Forward(numTriangles, triangles, faceNormals);
	}
  	

#if 0
	{
		Str255	s;
		NumToString(gNumStripBreaks, s);
		DoAlert (s);
	}
#endif	


			/* UPDATE THE TRIMESH */
			
	Q3TriMesh_SetData(triMesh, &triMeshData);
bail:	
	Q3TriMesh_EmptyData(&triMeshData);
}


/********************** STRIP: FORWARD ***********************/

static void Stripify_Forward(long numTriangles, TQ3TriMeshTriangleData *triangles, TQ3Vector3D *faceNormals)
{
long	i,j,stripLen,attempt;
unsigned long	t[2],m[3],q1,q2;
long		numMatches;
TQ3TriMeshTriangleData	temp;
TQ3Vector3D	temp2;

	stripLen = 1;
	for (i = 0; i < (numTriangles-1); i++)
	{
		attempt = 0;			

another_attempt:				

			/* GET LAST 2 POINT INDECIES FOR TRIANGLE */
				
		t[0] = triangles[i].pointIndices[1];
		t[1] = triangles[i].pointIndices[2];
		
		
			/***************************************************************/
			/* SCAN FOR A TRIANGLE WHO'S EDGE MATCHES 2ND EDGE OF THIS ONE */
			/***************************************************************/
			
		for (j = i+1; j < numTriangles; j++)
		{	
			
				/* GET ALL 3 POINT INDECIES FOR TRIANGLE */
				
			m[0] = triangles[j].pointIndices[0];
			m[1] = triangles[j].pointIndices[1];
			m[2] = triangles[j].pointIndices[2];


				/* SEE IF THERE IS A MATCH */
				
			numMatches = 0;
			
			if (t[0] == m[0])		{ numMatches++; q1 = 0; }
			else if (t[0] == m[1])	{ numMatches++; q1 = 1; }
			else if (t[0] == m[2])	{ numMatches++; q1 = 2; }
	
			if (t[1] == m[0])		{ numMatches++; q2 = 0; }
			else if (t[1] == m[1])	{ numMatches++; q2 = 1; }
			else if (t[1] == m[2])	{ numMatches++; q2 = 2; }
	
	
				/* WHICH EDGES MATCHED */
				//
				// We alter the triangle indecies such that the 1st 2 indecies always match the last 2 indecies
				// and in the same order.  Thus if #1 is 0,1,2 then #2 must be 1,2,x.
				//
				
			if (numMatches == 2)
			{
				if ((q1 == 0) && (q2 == 1))					// see if 2nd triangle is alredy ordered correctly (1st 2 points match last 2 points of other)
					goto do_swap;	
				else
				if ((q1 == 1) && (q2 == 0))					// need to reverse 1st 2 points 
				{
					triangles[j].pointIndices[0] = m[1];
					triangles[j].pointIndices[1] = m[0];
					goto do_swap;
				}
				else
				if ((q1 == 1) && (q2 == 2))					// need to rotate indecies
				{
					triangles[j].pointIndices[0] = m[1];
					triangles[j].pointIndices[1] = m[2];
					triangles[j].pointIndices[2] = m[0];
					goto do_swap;
				}
				else
				if ((q1 == 2) && (q2 == 1))					// need to rotate indecies 
				{
					triangles[j].pointIndices[0] = m[2];
					triangles[j].pointIndices[1] = m[1];
					triangles[j].pointIndices[2] = m[0];
					goto do_swap;
				}
				else
				if ((q1 == 0) && (q2 == 2))					// need to rotate indecies 
				{
					triangles[j].pointIndices[1] = m[2];
					triangles[j].pointIndices[2] = m[1];
					goto do_swap;
				}
				else
				if ((q1 == 2) && (q2 == 0))					// need to rotate indecies
				{
					triangles[j].pointIndices[0] = m[2];
					triangles[j].pointIndices[1] = m[0];
					triangles[j].pointIndices[2] = m[1];
					goto do_swap;
				}
				else
					goto no_match;
//					DoFatalAlert("\pStripify_Forward: I messed up!");
	
	
				/* SWAP NEW TRIANGLE WITH "NEXT" */
do_swap:
				temp = triangles[i+1];						// swap triangle data
				triangles[i+1] = triangles[j];
				triangles[j] = temp;

				temp2 = faceNormals[i+1];					// swap face normals
				faceNormals[i+1] = faceNormals[j];
				faceNormals[j] = temp2;
				
				stripLen++;
				break;										// break inner loop and check next main triangle
			}
			
				/* NO MATCH, SO TRY NEXT COMPARE */
no_match:;				
		}
		
			/* NO MATCHING EDGE AT ALL, SO THIS IS END OF THIS STRIP */
			
		if (stripLen == 1)									// if was single triangle, then fudge it to it'll maybe be better next pass
		{
			triangles[i].pointIndices[1] = triangles[i].pointIndices[0];
			triangles[i].pointIndices[0] = t[1];
			triangles[i].pointIndices[2] = t[0];
			if (++attempt < 3)
				goto another_attempt;
		}
			
		if (j >= numTriangles)
		{
			gNumStripBreaks++;
			stripLen = 1;									// reset for next strip to begin
		}
		
	} // for i
}




/********************** STRIP: BACKWARD ***********************/

static void Stripify_Backward(long numTriangles, TQ3TriMeshTriangleData *triangles, TQ3Vector3D *faceNormals)
{
long	i,j,stripLen,attempt;
unsigned long	t[2],m[3],q1,q2;
long	numMatches;
TQ3TriMeshTriangleData	temp;
TQ3Vector3D	temp2;

	stripLen = 1;
	
	for (i = numTriangles-1; i > 0; i--)
	{
		attempt = 0;			

another_attempt:				
	
			/* GET FIRST 2 POINT INDECIES FOR TRIANGLE */
				
		t[0] = triangles[i].pointIndices[0];
		t[1] = triangles[i].pointIndices[1];
		
		
			/***************************************************************/
			/* SCAN FOR A TRIANGLE WHO'S EDGE MATCHES 1st EDGE OF THIS ONE */
			/***************************************************************/
				
		for (j = i-1; j >= 0; j--)
		{				
				/* GET ALL 3 POINT INDECIES FOR TRIANGLE */

			m[0] = triangles[j].pointIndices[0];
			m[1] = triangles[j].pointIndices[1];
			m[2] = triangles[j].pointIndices[2];


				/* SEE IF THERE IS A MATCH */
				
			numMatches = 0;
			
			if (t[0] == m[0])		{ numMatches=1; q1 = 0; }
			else if (t[0] == m[1])	{ numMatches=1; q1 = 1; }
			else if (t[0] == m[2])	{ numMatches=1; q1 = 2; }
	
			if (t[1] == m[0])		{ numMatches++; q2 = 0; }
			else if (t[1] == m[1])	{ numMatches++; q2 = 1; }
			else if (t[1] == m[2])	{ numMatches++; q2 = 2; }
	
	
				/* WHICH EDGES MATCHED */
				//
				// We alter the triangle indecies such that the 1st 2 indecies always match the last 2 indecies
				// and in the same order.  Thus if #1 is 0,1,2 then #2 must be 1,2,x.
				//
				
			if (numMatches == 2)
			{
				if ((q1 == 1) && (q2 == 2))		
					goto do_swap;	
				else
				if ((q1 == 2) && (q2 == 1))					// need to reverse
				{
					triangles[j].pointIndices[1] = m[2];
					triangles[j].pointIndices[2] = m[1];
					goto do_swap;
				}
				else
				if ((q1 == 0) && (q2 == 1))					// need to rotate indecies
				{
					triangles[j].pointIndices[0] = m[2];
					triangles[j].pointIndices[1] = m[0];
					triangles[j].pointIndices[2] = m[1];
					goto do_swap;
				}
				else
				if ((q1 == 1) && (q2 == 0))					// need to rotate indecies 
				{
					triangles[j].pointIndices[0] = m[2];
					triangles[j].pointIndices[2] = m[0];
					goto do_swap;
				}
				else
				if ((q1 == 0) && (q2 == 2))					// need to rotate indecies 
				{
					triangles[j].pointIndices[0] = m[1];
					triangles[j].pointIndices[1] = m[0];
					goto do_swap;
				}
				else
				if ((q1 == 2) && (q2 == 0))					// need to rotate indecies
				{
					triangles[j].pointIndices[0] = m[1];
					triangles[j].pointIndices[1] = m[2];
					triangles[j].pointIndices[2] = m[0];
					goto do_swap;
				}
				else
					goto no_match;
//					DoFatalAlert("\pStripify_Backward: I messed up!");
	
	
				/* SWAP NEW TRIANGLE WITH "NEXT" */
do_swap:
				temp = triangles[i-1];						// swap triangle data
				triangles[i-1] = triangles[j];
				triangles[j] = temp;

				temp2 = faceNormals[i-1];					// swap face normals
				faceNormals[i-1] = faceNormals[j];
				faceNormals[j] = temp2;
				
				stripLen++;
				break;										// break inner loop and check next main triangle
			}
			
				/* NO MATCH, SO TRY NEXT COMPARE */
no_match:;				
		}
		
			/* NO MATCHING EDGE AT ALL, SO THIS IS END OF THIS STRIP */
			
		if (stripLen == 1)									// if was single triangle, then fudge it to it'll maybe be better next pass
		{
			triangles[i].pointIndices[1] = triangles[i].pointIndices[2];
			triangles[i].pointIndices[0] = t[1];
			triangles[i].pointIndices[2] = t[0];
			if (++attempt < 3)
				goto another_attempt;
		}
			
		if (j < 0)
		{
			gNumStripBreaks++;		
			stripLen = 1;									// reset for next strip to begin
		}
	}
}










