/****************************/
/*   	PROCESS.C     		*/
/* (c)1996 Pangea Software  */
/* By Brian Greenstone      */
/****************************/


/****************************/
/*    EXTERNALS             */
/****************************/

#include <TextUtils.h>
#include <Fonts.h>
#include <QD3D.h>
#include <QD3DCamera.h>
#include <QD3DDrawContext.h>
#include <QD3DErrors.h>
#include <QD3DGeometry.h>
#include <QD3DGroup.h>
#include <QD3DIO.h>
#include <QD3DLight.h>
#include <QD3DMath.h>
#include <QD3DPick.h>
#include <QD3DRenderer.h>
#include <QD3DShader.h>
#include <QD3DStyle.h>
#include <QD3DView.h>
#include	<QD3DTransform.h>
#include	<drag.h>
#include <Resources.h>

#include <timer.h>
#include "myglobals.h"
#include "objects.h"
#include "mymenus.h"
#include "mywindows.h"
#include "myevents.h"
#include "misc.h"
#include "process.h"
#include "qd3d_support.h"
#include "3dmf.h"
#include "dialogStuff.h"
#include "optimize.h"
#include "environmentmap.h"

extern	EventRecord			gTheEvent;
extern	NewObjectDefinitionType	gNewObjectDefinition;
extern	ObjNode				*gCurrentNode,*gFirstNodePtr;
extern	float				gFramesPerSecond;
extern	TQ3Point3D			gRoadSplinePoints[],gCamCoord;
extern	short				gNumTrTQ3AttributeSet,gNumTriMeshes;
extern	TQ3Object			gWorkShader;
extern	TQ3AttributeSet		gWorkAttribSet;
extern	long	gTotalNumVertecies,gTotalNumFaces,gTotalMeshes;
extern	Boolean				gShowAsReflectionMap,gExpertMode,gDropped3DMF,gUseCullingGroup;


/****************************/
/*    PROTOTYPES            */
/****************************/

static Boolean InitTextStreamOutput(void);
static pascal OSErr MyDragTrackingFunc(DragTrackingMessage theMessage, WindowPtr theWindow, void *theRefCon, DragReference theDragRef);
static pascal OSErr MyDragReceiveFunc(WindowPtr theWindow, void *theRefCon, DragReference theDragRef);
static Boolean IsDrag3DMF(DragReference theDrag);


/****************************/
/*    CONSTANTS             */
/****************************/

#define	MODEL_WIND_ID			400
#define	HEADER_HEIGHT			16


/*********************/
/*    VARIABLES      */
/*********************/

WindowPtr				gModelWindow = nil;
QD3DSetupOutputType		gModelViewInfo;
Boolean					gShowContentStructure = false;
Boolean					gSaveTextToFile = false;
FSSpec					gStreamFSSpec,gDragFSSpec;
short					gStreamRefNum;
Boolean					gAutoCenter = false;
Boolean					gCenterYAtZero = false;
float					gAutoCenterXOff,gAutoCenterYOff,gAutoCenterZOff;

Boolean					gSaveAsBinary = true;

ObjNode					*gModelObjNode = nil;

TQ3Object			gModelGroup = nil;

Boolean					gCurrentIsOptimized = false;

float					gRotDX = 1, gRotDY = .4, gRotDZ = 0;


/************** INIT MODEL WINDOW *******************/

void InitModelWindow(void)
{
QD3DSetupInputType	viewDef;
TQ3Point3D			cameraFrom = { 0,0, 8 };
TQ3Point3D			cameraTo = { 0.0, 0, 0 };
TQ3Vector3D			cameraUp = { 0.0, 1.0, 0 };
TQ3ColorARGB		clearColor = {1,0,0,.5};
TQ3ColorRGB			ambientColor = { 1.0, 1.0, 1.0 };
TQ3Vector3D			fillDirection1 = {.4, -.2, -1 };
TQ3Vector3D			fillDirection2 = { -1, -1, -.2 };

OSErr				iErr;


			/* CREATE THE WINDOW */
			
	gModelWindow = GetNewCWindow(MODEL_WIND_ID, nil,MOVE_TO_FRONT);
	if (gModelWindow == nil)
		DoFatalAlert("\pWhere did the GameWindow window go?");					
	SetPort((GrafPtr)gModelWindow);


			/***************************/
			/* SETUP DRAG & DROP STUFF */
			/***************************/

	iErr = InstallTrackingHandler(NewDragTrackingHandlerProc(MyDragTrackingFunc), gModelWindow, nil);
	if (iErr)
		DoFatalAlert("\pInitModelWindow:  InstallTrackingHandler failed!");					

	iErr = InstallReceiveHandler(NewDragReceiveHandlerProc(MyDragReceiveFunc), gModelWindow, nil);
	if (iErr)
		DoFatalAlert("\pInitModelWindow:  InstallReceiveHandler failed!");					


			/***********************/
			/* SET QD3D PARAMETERS */
			/***********************/
			
	viewDef.view.displayWindow 		= gModelWindow;
	viewDef.view.rendererType 		= kQ3RendererTypeInteractive;
	viewDef.view.clearColor 		= clearColor;
	viewDef.view.paneClip.left 		= 0;
	viewDef.view.paneClip.right 	= 0;
	viewDef.view.paneClip.top 		= 16;
	viewDef.view.paneClip.bottom 	= 0; 

	viewDef.styles.interpolation 	= kQ3InterpolationStyleVertex; 
	viewDef.styles.backfacing 		= kQ3BackfacingStyleRemove;
	viewDef.styles.fill				= kQ3FillStyleFilled; 
	viewDef.styles.shaderType 		= kQ3IlluminationTypeLambert; 

	gCamCoord = viewDef.camera.from 			= cameraFrom;
	viewDef.camera.to 				= cameraTo;
	viewDef.camera.up 				= cameraUp;
	viewDef.camera.hither 			= 2;
	viewDef.camera.yon 				= 13;
	viewDef.camera.fov 				= 1.0;

	viewDef.lights.ambientBrightness = 0.2;
	viewDef.lights.ambientColor 	= ambientColor;
	viewDef.lights.numFillLights 	= 2;
	viewDef.lights.fillDirection[0] = fillDirection1;
	viewDef.lights.fillDirection[1] = fillDirection2;
	viewDef.lights.fillColor[0] 	= ambientColor;
	viewDef.lights.fillColor[1] 	= ambientColor;
	viewDef.lights.fillBrightness[0] = 1.1;
	viewDef.lights.fillBrightness[1] = 0.6;


	QD3D_SetupWindow(&viewDef, &gModelViewInfo);


		/* SET RAVE STUFF */
		
	GetMyRAVEInfo();
}


/********************** MY DRAG TRACKING FUNC **************************/
//
// Tracking handler for Drag & Drop support
//

static pascal OSErr MyDragTrackingFunc(DragTrackingMessage theMessage, WindowPtr theWindow, void *theRefCon, DragReference theDragRef)
{
OSErr			err = noErr;
DragAttributes	attributes;
GrafPtr			oldPort;
Rect			r;

	theRefCon;

	GetDragAttributes(theDragRef, &attributes);						// get attribs
	if (attributes & kDragInsideSenderWindow)						// see if drag originated from this window
		return(noErr);

	if (!IsDrag3DMF(theDragRef))									// see if it's a 3DMF
		return(noErr);

	switch(theMessage)
	{
		case	kDragTrackingEnterWindow:							/* drag has entered window*/
		case	kDragTrackingLeaveWindow:							/* drag has exited window*/
				GetPort(&oldPort);
				SetPort(theWindow);
				ForeColor(whiteColor);
				PenSize(4,4);
				PenMode(patXor);
				r = theWindow->portRect;
				if (gModelGroup)
					r.top += HEADER_HEIGHT;
				FrameRect(&r);
				PenNormal();
				SetPort(oldPort);
				break;
		
		case	kDragTrackingInWindow:								/* drag is moving within window*/
				break;
				
		case	kDragTrackingLeaveHandler:						/* drag has exited handler*/
				break;

		case	kDragTrackingEnterHandler:							/* drag has entered handler*/
				break;
	}
	
	return(err);
}



/********************** MY DRAG RECEIVE FUNC **************************/
//
// Receiving handler for Drag & Drop support
//

static pascal OSErr MyDragReceiveFunc(WindowPtr theWindow, void *theRefCon, DragReference theDragRef)
{
DragAttributes	attributes;

	theWindow;	theRefCon;

	GetDragAttributes(theDragRef, &attributes);						// get attribs
	if (attributes & kDragInsideSenderWindow)						// see if drag originated from this window
		return(noErr);

	if (!IsDrag3DMF(theDragRef))									// see if it's a 3DMF
		return(noErr);

	gDropped3DMF = true;											// set flag so event loop will know to do something
	return(noErr);
}


/*********************** IS DRAG 3DMF ***************************/
//
// Determines if a drag reference is a 3DMF file/object
//

static Boolean IsDrag3DMF(DragReference theDrag)
{
HFSFlavor	hfsData;
Size		dataSize = sizeof(HFSFlavor);
unsigned short	numFlavors,i;
FlavorType	flavorType;
ItemReference	theItemRef;

	if (GetDragItemReferenceNumber(theDrag, 1, &theItemRef))
		return(false);


	if (CountDragItemFlavors(theDrag, theItemRef, &numFlavors))
		return(false);


	for (i = 1; i <= numFlavors; i++)
	{
		if (GetFlavorType(theDrag, theItemRef, i, &flavorType))
			return(false);

			
		if (flavorType == flavorTypeHFS)
		{
			if (GetFlavorData(theDrag, theItemRef, flavorTypeHFS, (Ptr)&hfsData, &dataSize, 0))
				return(false);
	
			if (hfsData.fileType == '3DMF')
			{
				gDragFSSpec = hfsData.fileSpec;
				return(true);
			}
			else
				return(false);
		}
	}

	return(false);
}




/******************** DRAW MODEL WINDOW HEADER ***********************/

void DrawModelWindowHeader(void)
{
PicHandle p;
Rect		r;
Str255	s;

	SetPort(gModelWindow);
	
	TextSize(9);
	BackColor(blackColor);
	ForeColor(whiteColor);
	TextMode(srcCopy);
	TextFace(kFontIDGeneva);
	TextFace(0);
	
	p = GetPicture(128);
	r = (**p).picFrame;
	DrawPicture(p, &r);
	ReleaseResource((Handle)p);	
	
	
	MoveTo(120,10);								// show # tri
	NumToString(gTotalNumFaces,s);
	DrawString(s);
	DrawString("\p  ");

	MoveTo(230,10);								// show # verts
	NumToString(gTotalNumVertecies,s);
	DrawString(s);
	DrawString("\p  ");


	MoveTo(340,10);								// show # meshes
	NumToString(gTotalMeshes,s);
	DrawString(s);
	DrawString("\p  ");
	
	
}


/******************* DRAW MODEL WINDOW *********************/
//
// Calls the support function QD3D_DrawScene.
// It passes the view info and a pointer to the game draw callback.
//

void DrawModelWindow(void)
{
	InitReflectionMapQueue();

	if (gShowAsReflectionMap)
		CalcEnvironmentMappingCoords(gFirstNodePtr->BaseGroup);

	QD3D_DrawScene(&gModelViewInfo,(void *)DrawObjects);
}



/*************** DO MODEL WINDOW NULL EVENT **********************/

void DoModelWindowNullEvent(void)
{
	if (gModelObjNode)
	{
		SetPort(gModelWindow);
		QD3D_CalcFramesPerSecond();
		MoveObjects();
		DrawModelWindow();
	}
}



/****************** UPDATE MODEL WINDOW ******************/
//
// Called when there is an update event from the event manager
//

void UpdateModelWindow(void)
{
PicHandle	pic;
	
	if (gModelGroup)
	{
		DrawModelWindowHeader();	
		DrawModelWindow();
	}
	else
	{
#if ATI_LICENSE	
		pic = GetPicture(132);
#else
		pic = GetPicture(129);							// load ATI version of splash screen
#endif		
		SetPort(gModelWindow);
		DrawPicture(pic,&gModelWindow->portRect);
		ReleaseResource((Handle)pic);
	}
	
}




/****************** OPEN NEW MODEL **********************/
//
// INPUT: isBatching = true if open model in batch mode
//

void OpenNewModel(Boolean isBatching, FSSpec *desiredFile, Boolean expertOptions)
{
TQ3Object	newModel;
ObjNode		*newObj;
float 		scale;
TQ3BoundingBox	boundingBox;

			/* DO READ OPTIONS */

	if (gExpertMode && expertOptions)
	{			
		if (!isBatching)
			if (Do3DMFOpenOptions())
				return;
	}
	else
		gShowContentStructure = false;


			/* SEE IF INIT BREAKDOWN DISPLAY */
			
	if (gShowContentStructure)
	{
		InitBreakdownWindow();
		if (gSaveTextToFile)
		{
			if (InitTextStreamOutput())
				return;
		}
	}
	
				/* SEE IF GET RID OF ANY EXISTING MODEL */
				
	if (gModelObjNode != nil)
	{
		DeleteObject(gModelObjNode);
		gModelObjNode = nil;
	}

			/* LOAD THE MODEL & DISPLAY CONTENTS */
			
	newModel = Load3DMFModel(desiredFile);
	if (newModel == nil)
		return;

			/* CLOSE FILE (IF ANY) */
			
	if (gShowContentStructure && gSaveTextToFile)
		FSClose(gStreamRefNum);


				/* SHOW THE OBJECT */

	gNewObjectDefinition.genre = DISPLAY_GROUP_GENRE;
	gNewObjectDefinition.coord.x = 0;
	gNewObjectDefinition.coord.y = 0;
	gNewObjectDefinition.coord.z = 0;
	gNewObjectDefinition.flags = OBJ_MODE_DRAW;
	gNewObjectDefinition.slot = 100;
	gNewObjectDefinition.moveCall = MoveModel;
	newObj = MakeNewDisplayGroupObject(&gNewObjectDefinition);
	if (newObj == nil)
		return;

	gModelObjNode = newObj;									// remember this
	gModelGroup = newModel;

	scale = CalcScaleToMakeFit(&gModelViewInfo,newModel, TARGET_SIZE);			// calc scale 
	newObj->ScaleX = newObj->ScaleY = newObj->ScaleZ = scale;

	AttachGeometryToDisplayGroupObject(newObj,newModel);
	
				/* CALC AUTO-CENTER OFFSET */

		QD3D_CalcObjectBoundingBox(&gModelViewInfo, newModel, &boundingBox);
		gAutoCenterXOff = -(boundingBox.max.x + boundingBox.min.x) / 2.0f;		// calc center X
		gAutoCenterZOff = -(boundingBox.max.z + boundingBox.min.z) / 2.0f;		// calc center Z
		if (gCenterYAtZero)
			gAutoCenterYOff = -boundingBox.min.y;								// keey bottom/y @ y=0
		else
			gAutoCenterYOff = -(boundingBox.max.y + boundingBox.min.y) / 2.0f;	// calc center Y
	
	gCurrentIsOptimized	= false;
	
			/* BUILD BASIC MATRIX */
			
	UpdateObjectTransforms(newObj);
}


/***************** MOVE MODEL *********************/

void MoveModel(void)
{
ObjNode	*theNode = gCurrentNode;
TQ3Matrix4x4	m;

	Q3Matrix4x4_SetRotate_XYZ(&m,gRotDX/gFramesPerSecond,
								gRotDY/gFramesPerSecond,
								gRotDZ/gFramesPerSecond);


	Q3Matrix4x4_Multiply(&theNode->BaseTransformMatrix,	
						 &m,
						 &theNode->BaseTransformMatrix);


	SetObjectTransformMatrix(theNode);
}


/***************** INTERACTIVE ROTATE *********************/

void InteractiveRotate(void)
{
Point	p,oldP;
float	dx,dy,timer;

	if (!gModelGroup)
		return;

	GetMouse(&oldP);

	while(Button())
	{
		GetMouse(&p);
		
		dx = p.h - oldP.h;
		dy = p.v - oldP.v;

		if ((dx != 0) || (dy != 0))
		{
			gRotDX = dy * 2;
			gRotDY = dx * 2;

			gCurrentNode = gModelObjNode;
			MoveModel();						
			
			timer = 0;
		}
		else
		{
			timer += 1.0/gFramesPerSecond;
			if (timer > .08)
				gRotDX = gRotDY = 0;
		}
		
		QD3D_DrawScene(&gModelViewInfo,(void *)DrawObjects);
		QD3D_CalcFramesPerSecond();
		oldP = p;		
	}
	
	gRotDX *= .5;
	gRotDY *= .5;
	
}



/***************** INIT TEXT STREAM OUTPUT **********************/
//
// OUTPUT: true = cancelled
//

static Boolean InitTextStreamOutput(void)
{
StandardFileReply	reply;
OSErr		iErr;
		
	StandardPutFile("\pSave Stream TEXT as...","\pstream.txt",&reply);
	if (!reply.sfGood)								// see if cancelled 
		return(true);

	gStreamFSSpec = reply.sfFile;	
	
	if (reply.sfReplacing)
		FSpDelete(&gStreamFSSpec);									// delete any old one
	
				/* CREATE TEXT FILE */
				
	iErr = FSpCreate(&gStreamFSSpec, 'CWIE', 'TEXT', smSystemScript);		// create new file
	if (iErr)
	{
		DoAlert("\pError Creating Output text file.");
		return(true);
	}

			/* OPEN FILE FOR WRITING */
			
	iErr = FSpOpenDF(&gStreamFSSpec, fsCurPerm, &gStreamRefNum);
	if (iErr)
		DoFatalAlert("\pError Opening Output text file.");

	return(false);
}


/****************** SAVE OPTIMIZED MODEL **********************/

void SaveOptimizedModel(void)
{
	if (gModelObjNode == nil)
	{
		DoAlert("\pYou need to Open a 3DMF file before you can optimize and save it!");
		return;
	}

			/* PRE TRANSFORM MESH VERTECIES */
		
	if (!gCurrentIsOptimized)						// if not already optimized, then do it now
	{
		ShowWaitingWindow();
		OptimizeGeometry();
		KillWaitingWindow();
	}

			/*********************/
			/* SAVE THE FILE OUT */
			/*********************/

	Save3DMFModel(&gModelViewInfo,nil,WriteModelGeometry);				// save entire thing into 1 file
}


/**************** WRITE MODEL GEOMETRY ********************/

void WriteModelGeometry(QD3DSetupOutputType *setupInfo)
{
TQ3Status				status;
TQ3GroupPosition	position;
TQ3Object   		object;

	if (gModelObjNode == nil)
		return;

		
		
		/* IF BASE IS GROUP, THEN WRITE OUT OBJECTS IN GROUP SEPARATELY */
	
	if ((!gUseCullingGroup) && Q3Object_IsType(gModelGroup,kQ3ShapeTypeGroup))
 	{
  		for (Q3Group_GetFirstPosition(gModelGroup, &position); position != nil;
  			 Q3Group_GetNextPosition(gModelGroup, &position))				// scan all objects in group
 		{
   			Q3Group_GetPositionObject (gModelGroup, position, &object);		// get object from group
			if (object != NULL)
   			{
				status = Q3Object_Submit(object,setupInfo->viewObject);		// write it
				if (status == kQ3Failure)
				{
					TQ3Error e;
					
					e = Q3Error_Get(nil);
					DoAlert("\pWriting Model Failed!");
					ShowSystemErr(e);
				}
   				Q3Object_Dispose(object);
   			}
  		}
	}
	
		/* BASE ISNT A GROUP, SO JUST WRITE WHATEVER IT IS OUT */
	else
	{
		status = Q3DisplayGroup_Submit(gModelGroup,setupInfo->viewObject);		// write it
		if (status == kQ3Failure)
			DoFatalAlert("\pWriting Model Failed 2!");
	}
		
		
		
		
		
}











