/****************************/
/*   	REDUCTION.C    		*/
/* (c)1998 Pangea Software  */
/* By Brian Greenstone      */
/****************************/


/****************************/
/*    EXTERNALS             */
/****************************/

#include <QD3D.h>
#include <QD3DErrors.h>
#include <QD3DGeometry.h>
#include <QD3DGroup.h>
#include "QD3DMath.h"
#include "QD3DShader.h"
#include "QD3DStyle.h"
#include	<QD3DTransform.h>

#include "myglobals.h"
#include "objects.h"
#include "mymenus.h"
#include "mywindows.h"
#include "myevents.h"
#include "misc.h"
#include "reduction.h"
#include "qd3d_support.h"
#include "3dmf.h"
#include "dialogStuff.h"


/****************************/
/*    PROTOTYPES            */
/****************************/

static void DecomposeATriMesh(TQ3Object theTriMesh);
static void DecompRefMo_Recurse(TQ3Object inObj);
static void DecomposeReferenceModel(TQ3Object theModel);


/****************************/
/*    CONSTANTS             */
/****************************/



/*********************/
/*    VARIABLES      */
/*********************/

Byte					gNumDecomposedTriMeshes = 0;
TQ3TriMeshData			gDecomposedTriMeshes[MAX_DECOMPOSED_TRIMESHES];

long					gNumDecomposedPoints = 0;
DecomposedPointType		gDecomposedPointList[MAX_DECOMPOSED_POINTS];

long					gNumDecomposedNormals = 0;
DecomposedNormalType	gDecomposedNormalsList[MAX_DECOMPOSED_NORMALS];



#pragma mark ======== DECOMPOSITION ==========

/****************** DECOMPOSE REFERENCE MODEL ************************/
//
// Decomposes the master Object into lists of unique points and normals.
// All duplicates are removed.
//

static void DecomposeReferenceModel(TQ3Object theModel)
{
long				i;

			/* CLEAR OLD DATA */
			
	for (i = 0; i < gNumDecomposedTriMeshes; i++)
		Q3TriMesh_EmptyData(&gDecomposedTriMeshes[i]);	

	gNumDecomposedTriMeshes = 0;
	gNumDecomposedPoints = 0;
	gNumDecomposedNormals = 0;


		/* DO SUBRECURSIVE SCAN */
		
	DecompRefMo_Recurse(theModel);
}


/***************** DECOM REF MO: RECURSE ***************************/
//
// Sub-recursive scan to decompose each object.
//

static void DecompRefMo_Recurse(TQ3Object inObj)
{
TQ3GroupPosition	position;
TQ3Object   		newObj;
TQ3ObjectType		oType;

				/* SEE IF FOUND GEOMETRY */

	if (Q3Object_IsType(inObj,kQ3ShapeTypeGeometry))
	{
		oType = Q3Geometry_GetType(inObj);									// get geometry type
		switch(oType)
		{
			case	kQ3GeometryTypeTriMesh:
					DecomposeATriMesh(inObj);
					break;
		}
	}
	else
	
			/* SEE IF RECURSE SUB-GROUP */

	if (Q3Object_IsType(inObj,kQ3ShapeTypeGroup))
 	{
  		for (Q3Group_GetFirstPosition(inObj, &position); position != nil;	// scan all objects in group
  			 Q3Group_GetNextPosition(inObj, &position))	
 		{
   			Q3Group_GetPositionObject (inObj, position, &newObj);			// get object from group
			if (newObj != NULL)
   			{
    			DecompRefMo_Recurse(newObj);								// sub-recurse this object
    			Q3Object_Dispose(newObj);									// dispose local ref
   			}
  		}
	}
}


/******************* DECOMPOSE A TRIMESH ***********************/
//
// Decomposes a trimesh object
//

static void DecomposeATriMesh(TQ3Object theTriMesh)
{
TQ3Status			status;
unsigned long		numVertecies,vertNum;
TQ3Point3D			*vertexList;
long				i,n,refNum,pointNum;
TQ3Vector3D			*normalPtr;

	n = gNumDecomposedTriMeshes;												// get index into list of trimeshes
	if (n >= MAX_DECOMPOSED_TRIMESHES)
		DoFatalAlert("\pDecomposeATriMesh: gNumDecomposedTriMeshes > MAX_DECOMPOSED_TRIMESHES");


					/* GET TRIMESH'S DATA */
			
	status = Q3TriMesh_GetData(theTriMesh, &gDecomposedTriMeshes[n]);
	if (status != kQ3Success) 
		DoFatalAlert("\pPreTransformTriMesh: Q3TriMesh_GetData failed!");
		
	numVertecies = gDecomposedTriMeshes[n].numPoints;							// get # point
	vertexList = gDecomposedTriMeshes[n].points;								// point to points list
	normalPtr  = (TQ3Vector3D *)(gDecomposedTriMeshes[n].vertexAttributeTypes[0].data); // point to normals



				/*******************************/
				/* EXTRACT VERTECIES & NORMALS */
				/*******************************/
				
	for (vertNum = 0; vertNum < numVertecies; vertNum++)
	{				
			/* SEE IF THIS POINT IS ALREADY IN THERE */
				
		for (i=0; i < gNumDecomposedPoints; i++)
		{
			if (PointsAreCloseEnough(&vertexList[vertNum],&gDecomposedPointList[i].realPoint))	// see if close enough match
			{
					/* ADD ANOTHER REFERENCE */
					
				refNum = gDecomposedPointList[i].numRefs;							// get # refs for this point
				if (refNum >= MAX_POINT_REFS)
					DoFatalAlert("\pDecomposeATriMesh: MAX_POINT_REFS exceeded!");
					
				gDecomposedPointList[i].triMeshNum[refNum] = n;						// set triMesh #
				gDecomposedPointList[i].pointNum[refNum] = vertNum;					// set point #
				gDecomposedPointList[i].numRefs++;									// inc counter
				goto added_vert;
			}
		}
	
				/* IT'S A NEW POINT SO ADD TO LIST */
				
		i = gNumDecomposedPoints;
		if (i >= MAX_DECOMPOSED_POINTS)
			DoFatalAlert("\pDecomposeATriMesh: MAX_DECOMPOSED_POINTS exceeded!");
		
		refNum = 0;														// it's the 1st entry (need refNum for verts below).
		
		gDecomposedPointList[i].realPoint = vertexList[vertNum];		// add new point to list			
		gDecomposedPointList[i].triMeshNum[refNum] = n;					// set triMesh #
		gDecomposedPointList[i].pointNum[refNum] = vertNum;				// set point #
		gDecomposedPointList[i].numRefs = 1;							// set # refs to 1
		gDecomposedPointList[i].isAttached = false;	
		
		gNumDecomposedPoints++;											// inc # decomposed points
		
added_vert:
		pointNum = i;													// remember which point we are using
		
		
				/***********************************************/
				/* ADD THIS POINT'S NORMAL TO THE NORMALS LIST */
				/***********************************************/
				
					/* SEE IF NORMAL ALREADY IN LIST */
					
		for (i=0; i < gNumDecomposedNormals; i++)
		{
			if (VectorsAreCloseEnough(&normalPtr[vertNum],&gDecomposedNormalsList[i].realVector))	// if already in list, then dont add it again
				goto added_norm;
		}

				/* ADD NEW NORMAL TO LIST */
				
		i = gNumDecomposedNormals;
		if (i >= MAX_DECOMPOSED_NORMALS)
			DoFatalAlert("\pDecomposeATriMesh: MAX_DECOMPOSED_NORMALS exceeded!");
		
		gDecomposedNormalsList[i].realVector = normalPtr[vertNum];		// add new normal to list			
		gNumDecomposedNormals++;										// inc # decomposed normals
	
	
added_norm:
					/* KEEP REF TO NORMAL IN POINT LIST */

		gDecomposedPointList[pointNum].whichNormal[refNum] = i;			// save index to normal	
	}

	gNumDecomposedTriMeshes++;											// inc # of trimeshes in decomp list
}

