/****************************/
/*   TEXTURE OPTIMIZE.C	    */
/* (c)1996 Pangea Software  */
/* By Brian Greenstone      */
/****************************/


/****************************/
/*    EXTERNALS             */
/****************************/

#include <Resources.h>
#include "myglobals.h"
#include <QD3DGroup.h>
#include <QD3DLight.h>
#include <QD3DPick.h>
#include <QD3DTransform.h>
#include <QD3DStorage.h>
#include <QD3DMath.h>
#include 	<QuickTimeComponents.h>

#include <timer.h>
#include "misc.h"
#include "qd3d_support.h"
#include "textureOpt.h"
#include "mywindows.h"

extern	WindowPtr	gModelWindow;

/****************************/
/*    PROTOTYPES            */
/****************************/

static TQ3Object DoesTextureMatch(short pixelSize, void *newPixels, long w, long h);


/****************************/
/*    CONSTANTS             */
/****************************/

#define	MAX_TEXTURES		100

typedef struct
{
	long		width;
	long		height;
	short		depth;
	void		*pixels;
	TQ3Object	theTexture;			// illegal reference to the texture
}TextureType;


/*********************/
/*    VARIABLES      */
/*********************/

Boolean	gBlackIsTransparent = false;
Boolean	gConvertToMipmap 	= true;
Boolean	gCompressTextures 	= false;
Boolean	gReduce32BitTextures = true;

Boolean	gPreMultiply32BitTextures = false;

CodecQ			gCodecQuality 	= codecNormalQuality;
CodecType		gCodecType		= kJPEGCodecType;
CodecComponent 	gCodecComponent = anyCodec;
short	 		gCodecDepth		= 16;

long			gNumTextures = 0;
TextureType		gTextureList[MAX_TEXTURES];


/********************* INIT TEXTURE LIST **************************/

void InitTextureList(void)
{
long	i;

	for (i = 0 ; i < gNumTextures; i++)
	{
		DisposePtr((Ptr)gTextureList[i].pixels);
	}
	gNumTextures = 0;
}


/******************** OPTIMIZE TEXTURE ****************************/
//
// Optimizes by converting all textures to 16bit and making
// rowBytes == width*pixel (in other words, it removes the blank padding @ the end of each scan line)
//

TQ3SurfaceShaderObject OptimizeTexture(TQ3SurfaceShaderObject theShader)
{
TQ3TextureObject	texture;
TQ3StoragePixmap	pixmap;
TQ3Mipmap			mipmap;
TQ3CompressedPixmap	compPixmap;
TQ3Status			status;
Rect				r,r2;
GWorldPtr			textureGWorld,newGWorld;
OSErr				iErr;
unsigned short		*pix16Src,*pix16Dest,pixel16;
unsigned long		*pix32Src,*pix32Dest,pixel32;
void				*pixelBuffer;
PixMapHandle	gworldPixmap;
long			gworldRowBytes,x,y,targetDepth;
Ptr				gworldPixelPtr;
long			size,width,height;
Boolean			hasAlpha,srcIsMipmap;
TQ3SurfaceShaderObject		existingShader;


				/* GET INFO ON THIS TEXTURE */
				
	status = Q3TextureShader_GetTexture(theShader, &texture);								// get texture from shader
	if (status != kQ3Success)
		DoFatalAlert("\pOptimizeTexture: Q3TextureShader_GetTexture failed!");

	switch(Q3Texture_GetType(texture))														// must be pixmap or mipmap src
	{
		case	kQ3TextureTypePixmap:
				srcIsMipmap = false;
				status = Q3PixmapTexture_GetPixmap(texture, &pixmap);						// get texture's pixmap
				if (status != kQ3Success)
					DoFatalAlert("\pOptimizeTexture: Q3PixmapTexture_GetPixmap failed!");
				textureGWorld = QD3D_PixmapToGWorld(&pixmap,&r);					
				Q3Object_Dispose(pixmap.image);												// nuke old image storage
				break;
				
		case	kQ3TextureTypeMipmap:
				srcIsMipmap = true;
				status = Q3MipmapTexture_GetMipmap(texture, &mipmap);						// get texture's mipmap
				if (status != kQ3Success)
					DoFatalAlert("\pOptimizeTexture: Q3MipmapTexture_GetMipmap failed!");
				textureGWorld = QD3D_MipmapToGWorld(&mipmap,&r);							// draw into gworld	
				Q3Object_Dispose(mipmap.image);												// nuke old image storage
				break;
				
		case	kQ3TextureTypeCompressedPixmap:
				srcIsMipmap = false;
				status = Q3CompressedPixmapTexture_GetCompressedPixmap(texture, &compPixmap);	// get texture's mipmap
				if (status != kQ3Success)
					DoFatalAlert("\pOptimizeTexture: Q3CompressedPixmapTexture_GetCompressedPixmap failed!");
				textureGWorld = QD3D_CompressedPixmapToGWorld(&compPixmap,&r);					// draw into gworld	
				Q3Object_Dispose(compPixmap.compressedImage);								// nuke old image storage
				Q3Object_Dispose(compPixmap.imageDesc);
				break;
				
		default:					
				DoAlert("\pWarning, OptimizeTexture enountered a non-pixmap texture.");
				return(nil);
	}
		
		
			/* ROUND DOWN TO POWER OF 2 IN SIZE */
		
	r2 = r;
		

	if (r2.right >= 512)									// pin textures as max 512x512
		r2.right = 512;
	else
	if (r2.right >= 256)				
		r2.right = 256;
	else
	if (r2.right >= 128)
		r2.right = 128;
	else
	if (r2.right >= 64)
		r2.right = 64;
	else
	if (r2.right >= 32)
		r2.right = 32;
	else
	if (r2.right >= 16)
		r2.right = 16;
	else
	if (r2.right >= 8)
		r2.right = 8;
	else
	if (r2.right >= 4)
		r2.right = 4;
	else
		r2.right = 2;
		
	if (r2.bottom >= 512)									// pin textures as max 512x512
		r2.bottom = 512;
	else
	if (r2.bottom >= 256)			
		r2.bottom = 256;
	else
	if (r2.bottom >= 128)
		r2.bottom = 128;
	else
	if (r2.bottom >= 64)
		r2.bottom = 64;
	else
	if (r2.bottom >= 32)
		r2.bottom = 32;
	else
	if (r2.bottom >= 16)
		r2.bottom = 16;
	else
	if (r2.bottom >= 8)
		r2.bottom = 8;
	else
	if (r2.bottom >= 4)
		r2.bottom = 4;
	else
		r2.bottom = 2;


			/* MAKE GWORLD TO CONVERT INTO */

	if (gReduce32BitTextures)								// see if force 16-bit or use original depth
		targetDepth = 16;
	else
	{
		gworldPixmap = GetGWorldPixMap(textureGWorld);
		targetDepth = (*gworldPixmap)->pixelSize;
	}

	iErr = NewGWorld(&newGWorld, targetDepth, &r2, nil, nil, useTempMem);
	if (iErr)
		DoFatalAlert("\pOptimizeTexture: NewGWorld failed!");
	DoLockPixels(newGWorld);
		
		
				/* COPY FROM SRC TO 16BIT GWORLD */
								
	DumpGWorldToGWorld(textureGWorld, newGWorld, &r, &r2);
	DisposeGWorld(textureGWorld);							// nuke original

	width = r2.right;
	height = r2.bottom;
	hasAlpha = false;										// assume no alpha in this texture
		

				/* GET GWORLD INFO */
				
	gworldPixmap = GetGWorldPixMap(newGWorld);
	LockPixels(gworldPixmap);
	gworldRowBytes = (**gworldPixmap).rowBytes & 0x3fff;	// get GWorld's rowbytes
	gworldPixelPtr = GetPixBaseAddr(gworldPixmap);			// get ptr to pixels


			/***************************************/		
			/* MAKE OPTIMIZED PIXMAP/MIPMAP BUFFER */
			/***************************************/		
				
	if (!gCompressTextures)
	{
					/* 16-BIT VERSION */
					
		if (targetDepth == 16)
		{
			size = width * height * 2;
			pixelBuffer = (unsigned short *)AllocPtr(size);		// alloc memory for 16bit buffer
			if (pixelBuffer == nil)
				DoFatalAlert("\pOptimizeTexture: AllocPtr failed!");
			
			
					/* COPY FROM GWORLD TO BUFFER */
			
			pix16Dest = pixelBuffer;										// dest == buffer
			pix16Src = (unsigned short *)gworldPixelPtr;					// src == gworld
			for (y = 0; y <  height; y++)
			{
				for (x = 0; x < width; x++)
				{
					pixel16 = pix16Src[x] & 0x7fff;							// get 16bit pixel (clear alpha)
					if (gBlackIsTransparent)
					{
						if (pixel16 != 0x0000)								// see if activate alpha on $0000 pure black
							pixel16 |= 0x8000;								// set bit to make solid/non-xparent
						else
							hasAlpha = true;								// leave bit15 off to make xparent
					}
					
					*pix16Dest++ = pixel16;
				}
			
				pix16Src += gworldRowBytes/2;
			}
		}
		
					/* 32BIT VERSION */
		else
		{
			size = width * height * 4;
			pixelBuffer = (unsigned short *)AllocPtr(size);					// alloc memory for 32bit buffer
			if (pixelBuffer == nil)
				DoFatalAlert("\pOptimizeTexture: AllocPtr failed!");
			
			
					/* COPY FROM GWORLD TO BUFFER */
			
			pix32Dest = pixelBuffer;										// dest == buffer
			pix32Src = (unsigned long *)gworldPixelPtr;						// src == gworld
			for (y = 0; y <  height; y++)
			{
				for (x = 0; x < width; x++)
				{
					pixel32 = pix32Src[x];									// get pixel
					if (pixel32 & 0xff000000)								// see if has an alpha value other than $00
						hasAlpha = true;
					*pix32Dest++ = pixel32;
				}			
				pix32Src += gworldRowBytes/4;
			}
			
					/* DO BLENDING PREMULTIPLY ON PIXELS */
					
			if (hasAlpha && gPreMultiply32BitTextures)
			{
				pix32Dest = pixelBuffer;									// point to buffer from above
				for (y = 0; y <  height; y++)
				{
					for (x = 0; x < width; x++)
					{
						double	a,r,g,b;
						
						pixel32 = *pix32Dest;								// get pixel							
						a = (double)((pixel32 >> 24) & 0xff) / 255.0;		// get alpha
						r = (double)((pixel32 >> 16) & 0xff) / 255.0;		// get r
						g = (double)((pixel32 >> 8)  & 0xff) / 255.0;		// get g
						b = (double)(pixel32 		 & 0xff) / 255.0;		// get b
						
						r *= a;												// do r,g,b premultiply
						g *= a;
						b *= a;
						
						pixel32 &= 0xff000000;								// keep old alpha value
						pixel32 |= (u_long)(r*255.0+.5) << 16;				// insert new r,g,b
						pixel32 |= (u_long)(g*255.0+.5) << 8;
						pixel32 |= (u_long)(b*255.0+.5);
						*pix32Dest++ = pixel32;
					}			
				}
			}
		}
		
		DisposeGWorld(newGWorld);										// nuke gworld
		
		
		/***********************************************/
		/* SEE IF THIS PIXEL BUFFER MATCHES ANY OTHERS */
		/***********************************************/

		existingShader = DoesTextureMatch(targetDepth, pixelBuffer, width, height);		
		if (existingShader)
			return(existingShader);										// pass illegal ref to existing shader
		

				/* TO USE MIPMAPS USE PIXMAP, NOT THEN USE MIPMAP - BACKWARDS, EH? */
				
		if (gConvertToMipmap)
		{
 			mipmap.image = Q3MemoryStorage_New((unsigned char *)pixelBuffer, size);	// make new storage from pixel buffer
			if (mipmap.image == nil)
				DoFatalAlert("\pOptimizeTexture: Q3MemoryStorage_New failed!");

			mipmap.useMipmapping = kQ3False;							// not actually using mipmaps
			if (targetDepth == 16)
			{
				mipmap.mipmaps[0].rowBytes 	= width*2;
				if (hasAlpha)
					mipmap.pixelType = kQ3PixelTypeARGB16;						
				else
					mipmap.pixelType = kQ3PixelTypeRGB16;						
			}
			else
			{
				mipmap.mipmaps[0].rowBytes 	= width*4;
				if (hasAlpha)
					mipmap.pixelType = kQ3PixelTypeARGB32;						
				else
					mipmap.pixelType = kQ3PixelTypeRGB32;						
			}
			
			mipmap.bitOrder 			= kQ3EndianBig;
			mipmap.byteOrder 			= kQ3EndianBig;
			mipmap.reserved 			= nil;
			mipmap.mipmaps[0].width 	= width;
			mipmap.mipmaps[0].height 	= height;
			mipmap.mipmaps[0].offset 	= 0;

			texture = Q3MipmapTexture_New(&mipmap);							// make new mipmap	
			if (texture == nil)
				DoFatalAlert("\pOptimizeTexture: Q3MipmapTexture_New failed!");
		}
		else	
		{
			pixmap.image 		= Q3MemoryStorage_New((unsigned char *)pixelBuffer, size);		// make new storage from pixel buffer
			pixmap.width 		= width;
			pixmap.height 		= height;
			pixmap.pixelSize 	= targetDepth;
			pixmap.bitOrder		= kQ3EndianBig;
			pixmap.byteOrder	= kQ3EndianBig;
			
			if (targetDepth == 16)
			{
				pixmap.rowBytes 	= 	width*2;
				if (hasAlpha)
					pixmap.pixelType = kQ3PixelTypeARGB16;	
				else
					pixmap.pixelType = kQ3PixelTypeRGB16;
			}
			else
			{
				pixmap.rowBytes 	= 	width*4;
				if (hasAlpha)
					pixmap.pixelType = kQ3PixelTypeARGB32;	
				else
					pixmap.pixelType = kQ3PixelTypeRGB32;
			}

			texture = Q3PixmapTexture_New(&pixmap);							// make new pixmap	
			if (texture == nil)
				DoFatalAlert("\pOptimizeTexture: Q3PixmapTexture_New failed!");
		}
	}
	
				/*******************************************/
				/* WE WANT THIS TO BE A COMPRESSED TEXTURE */
				/*******************************************/
	else
	{
					/* FIX THE ALPHA INFORMATION IN GWORLD */
					
		if (targetDepth == 16)
		{
			pix16Src = (unsigned short *)gworldPixelPtr;					// src == gworld
			for (y = 0; y <  height; y++)
			{
				for (x = 0; x < width; x++)
				{
					pixel16 = pix16Src[x] & 0x7fff;							// get 16bit pixel (clear alpha)
					if (gBlackIsTransparent)
					{
						if (pixel16 != 0x0000)								// see if activate alpha on $0000 pure black
							pixel16 |= 0x8000;								// set bit to make solid/non-xparent
						else
							hasAlpha = true;								// leave bit15 off to make xparent
					}
					pix16Src[x] = pixel16;									// draw pixel back into gworld
				}
				pix16Src += gworldRowBytes/2;
			}
		}
		else
		{
			pix32Src = (unsigned long *)gworldPixelPtr;						// src == gworld
			for (y = 0; y <  height; y++)
			{
				for (x = 0; x < width; x++)
				{
					pixel32 = pix32Src[x];									// get pixel
					if (pixel32 & 0xff000000)								// see if alpha != $00
						hasAlpha = true;									
					pix32Src[x] = pixel32;									// draw pixel back into gworld
				}
				pix32Src += gworldRowBytes/4;
			}
		}
		
		
					/* FILL OUT COMPRESSED PIXMAP STRUCTURE */

		compPixmap.makeMipmaps 			= (TQ3Boolean)!gConvertToMipmap;	
		compPixmap.width 				= width;
		compPixmap.height				= height;
		compPixmap.pixelSize 			= targetDepth;
		if (targetDepth == 16)
		{
			if (hasAlpha)
				compPixmap.pixelType = kQ3PixelTypeARGB16;	
			else
				compPixmap.pixelType = kQ3PixelTypeRGB16;
		}
		else
		{
			if (hasAlpha)
				compPixmap.pixelType = kQ3PixelTypeARGB32;	
			else
				compPixmap.pixelType = kQ3PixelTypeRGB32;
		}


				/* COMPRESS IMAGE AND FILL OUT REMAINING RECORDS IN STRUCTURE */
				
		status = Q3CompressedPixmapTexture_CompressImage(&compPixmap, gworldPixmap, gCodecType,
														gCodecComponent, gCodecDepth,gCodecQuality);
		if (status == kQ3Failure)
			DoFatalAlert("\pOptimizeTexture: Q3CompressedPixmapTexture_CompressImage Failed!");	
		DisposeGWorld (textureGWorld);		
		
		
			/* MAKE NEW COMPRESSED PIXMAP TEXTUE OBJECT */
					
		texture = Q3CompressedPixmapTexture_New (&compPixmap);
		if (texture == nil)
			DoFatalAlert("\pOptimizeTexture: Q3CompressedPixmapTexture_New failed!");		
	}


				/*************************/
				/* UPDATE TEXTURE SHADER */
				/*************************/
				
	status = Q3TextureShader_SetTexture(theShader, texture);	
	if (status != kQ3Success)
		DoFatalAlert("\pOptimizeTexture: Q3TextureShader_SetTexture failed!");
	
			/* ADD TO TEXTURE LIST */
			
			
	gTextureList[gNumTextures].width = width;
	gTextureList[gNumTextures].height = height;
	gTextureList[gNumTextures].depth = targetDepth;
	gTextureList[gNumTextures].pixels = pixelBuffer;
	gTextureList[gNumTextures].theTexture = theShader;	// keep illegal ref
	gNumTextures++;
	
				/* CLEANUP */
				
	Q3Object_Dispose(texture);				
	
	return(theShader);
}


/************************* DOES TEXTURE MATCH ****************************/

static TQ3Object DoesTextureMatch(short pixelSize, void *newPixels, long w, long h)
{
long	i,j;
u_short	*p1,*p2;
u_long	*pl1,*pl2;

	for (i = 0; i < gNumTextures; i++)
	{
		if (gTextureList[i].width != w)							// see if dimensions match
			continue;
		if (gTextureList[i].height != h)
			continue;

		if (gTextureList[i].depth != pixelSize)					// pixel depth must match
			continue;

				/* COMPARE 16-BIT */
				
		if (pixelSize == 16)
		{	
			p1 = newPixels;
			p2 = gTextureList[i].pixels;
		
			for (j = 0; j < (h*w); j++)							// compare pixels
			{
				if (p1[j] != p2[j])
					goto next;					
			}
		}
		
				/* COMPARE 32-BIT */
		else
		{
			pl1 = newPixels;
			pl2 = gTextureList[i].pixels;
		
			for (j = 0; j < (h*w); j++)							// compare pixels
			{
				if (pl1[j] != pl2[j])
					goto next;					
			}
		}
		
		return(gTextureList[i].theTexture);						// return illegal reference to the matching texture shader object
next:;		
	}
	
	
	return(nil);
}



/****************** GET MY COMPRESSION INFO ************************/

void GetMyCompressionInfo(void)
{
ComponentInstance	aComponent;
OSErr				iErr;
PicHandle			picture;
ComponentResult		aCompRes;
Boolean				cancel = false;
SCSpatialSettings	spatialInfo;

			/* CREATE COMPONENT */
			
	aComponent = OpenDefaultComponent('scdi', 'imag');
	if (aComponent == nil)
		DoFatalAlert("\pSelectCompressor: OpenDefaultComponent failed!");

			/* SET TEST PICT */
			
	picture = GetPicture(129);	
	aCompRes = SCSetTestImagePictHandle(aComponent, picture, nil, 0);

			/* DO THE DIALOG */
			
	aCompRes = SCRequestImageSettings(aComponent);
	if (aCompRes == scUserCancelled)
	{
		cancel = true;
		goto cleanup;
	}
	if (aCompRes < 0)
		DoFatalAlert("\pSelectCompressor: SCRequestSequenceSettings failed!");
	
	
			/* EXTRACT INFO */
	
	aCompRes = SCGetInfo(aComponent, scSpatialSettingsType, &spatialInfo);		
	if (aCompRes < 0)
		DoFatalAlert("\pSelectCompressor: SCGetInfo failed!");

	gCodecType 		= spatialInfo.codecType;					// get codec type
	gCodecComponent = spatialInfo.codec;						// get codec component
	gCodecQuality 	= spatialInfo.spatialQuality;				// get quality
	gCodecDepth 	= spatialInfo.depth;						// get depth


			/* CLEANUP */
			
cleanup:			
	ReleaseResource((Handle)picture);										// nuke pict
	iErr = CloseComponent(aComponent);										// nuke component
	if (iErr)
		DoFatalAlert("\pSelectCompressor: CloseComponent failed!");
}













