/****************************/
/*   	WAVEFRONTMTL.C 		*/
/* (c)1998 Pangea Software  */
/* By Brian Greenstone      */
/****************************/


/****************************/
/*    EXTERNALS             */
/****************************/

#include <QD3D.h>
#include <QD3DErrors.h>
#include <QD3DGeometry.h>
#include <QD3DGroup.h>
#include "QD3DMath.h"
#include "QD3DShader.h"
#include "QD3DStyle.h"
#include <QD3DTransform.h>

#include "myglobals.h"
#include "objects.h"
#include "mymenus.h"
#include "mywindows.h"
#include "myevents.h"
#include "misc.h"
#include "wavefrontobj.h"
#include "qd3d_support.h"
#include "3dmf.h"
#include "dialogStuff.h"
#include "optimize.h"


extern	ObjNode					*gModelObjNode;
extern	Boolean					gCurrentIsOptimized;
extern	FSSpec 					gTheOBJFileSpec;

/****************************/
/*    PROTOTYPES            */
/****************************/

static OSErr LoadMTLFile(FSSpec *spec);

static OSErr DecomposeMTL(void);
static Boolean SkipMTLSpaces(void);
static Boolean NextMTLLine(void);
static int GetMTLToken(void);
static Boolean SkipMTLUnwanted(void);
static Boolean ReadMTLWord(void);
static Boolean ReadMTLFloat(float *value);
static Boolean ReadMTLInt(long *value);
static void ReadMTLAmbientColor(void);
static void ReadMTLDiffuseColor(void);
static void ReadMTLSpecularColor(void);
static void NewMTLMaterial(void);
static void ReadMTLDiffuseCoeff(void);
static void ReadMTLSpecularCoeff(void);
static void ReadMTLDiffuseMap(void);


/****************************/
/*    CONSTANTS             */
/****************************/


enum
{
	TOKEN_END		=	0,
	TOKEN_NEWMTL,
	TOKEN_AMBIENT_COLOR,
	TOKEN_DIFFUSE_COLOR,
	TOKEN_SPECULAR_COLOR,
	TOKEN_TF,					// ????
	TOKEN_ILLUM,				// ????
	TOKEN_DIFFUSE_COEFF,
	TOKEN_SPECULAR_COEFF,
	TOKEN_SHARPNESS,			// ?????
	TOKEN_NI,					// ????
	TOKEN_AMBIENT_MAP,
	TOKEN_DIFFUSE_MAP,
	TOKEN_SPECULAR_MAP,
	
	TOKEN_UNKNOWN
};


#define	MAX_OBJ_MATERIALS	100


/*********************/
/*    VARIABLES      */
/*********************/

static long		gMTLFileIndex,gFileSize;
static Ptr		gTheMTLData = nil;
static	Str32	gMTLWord;

long			gNumOBJMaterials;
OBJMaterialType	*gOBJMaterials = nil;

long			gCurrentActiveMTLMaterial;


/*********** INIT OBJ MATERIALS ******************/

void InitOBJMaterials(void)
{
int		i;

			/* SEE IF ALLOCATE ARRAY */

	if (gOBJMaterials == nil)
		gOBJMaterials = (OBJMaterialType *)AllocPtr(sizeof(OBJMaterialType) * MAX_OBJ_MATERIALS);
		
			/* NUKE OLD MATERIALS */
			
	else
	{
		for (i = 0; i < gNumOBJMaterials; i++)
		{
			if (gOBJMaterials[i].texture)							// nuke texture shader
				Q3Object_Dispose(gOBJMaterials[i].texture);
		}
	}


			/* INIT COUNTS */
			
	gNumOBJMaterials = 0;
	gCurrentActiveMTLMaterial = 0;

}


/******************* PROCESS MTL FILE ***********************/

void ProcessMTLFile(FSSpec	*theFile)
{
			/* SELECT AND LOAD THE FILE */
			
	LoadMTLFile(theFile);


		/* DECOMPOSE THE MTL FILE */

	if (DecomposeMTL())
		return;

	
			/* CLEANUP */
			
	if (gTheMTLData)
	{
		DisposePtr(gTheMTLData);
		gTheMTLData = nil;
	}
}



/********************** LOAD MTL FILE **************************/

static OSErr LoadMTLFile(FSSpec *spec)
{
OSErr	iErr;
short	refNum = 0;


	if (gTheMTLData)								// see if nuke old data
	{
		DisposePtr(gTheMTLData);
		gTheMTLData = nil;
	}


			/****************************/
			/* OPEN & LOAD THE MTL FILE */
			/****************************/

	iErr = FSpOpenDF(spec, fsRdPerm, &refNum);			// open the file
	if (iErr)
	{
		DoAlert("\pDecomposeMTL: cannot open MTL file");
		return(1);
	}
	
	iErr = GetEOF(refNum, &gFileSize);						// get size of file
	if (iErr)
	{
		DoAlert("\pDecomposeMTL: GetEOF failed");
		goto bail;
	}
	
	gTheMTLData = AllocPtr(gFileSize);						// alloc memory for the file's data
	if (gTheMTLData == nil)
	{
		DoAlert("\pDecomposeMTL: Cannot alloc memory for file");
		goto bail;
	}
	
	iErr = FSRead(refNum, &gFileSize, gTheMTLData);		// read the file into memory
	if (iErr)
	{
		DoAlert("\pDecomposeMTL: Reading file failed");
		goto bail;
	}

	return(noErr);
	
	
bail:
	if (refNum)
		FSClose(refNum);
	if (gTheMTLData)
		DisposePtr(gTheMTLData);
		
	return(1);

}


/********************** DECOMPOSE MTL *****************************/

static OSErr DecomposeMTL(void)
{			

			/* POINT TO 1ST TOKEN */
							
	gMTLFileIndex = 0;									// start @ front of file
	if (SkipMTLUnwanted())
		goto badfile;
		
	if (gTheMTLData[gMTLFileIndex] == '#')				// see if 1st line is a comment
	{
		if (NextMTLLine())
		{
badfile:		
			DoAlert("\pDecomposeMTL: the file doesnt contain anything useful");
			return(1);								// nothing found
		}
	}	


		/*******************************/
		/* SCAN AND PROCESS EACH TOKEN */
		/*******************************/

	do
	{
		switch(GetMTLToken())
		{
			case	TOKEN_END:
					goto eof;
				
			case	TOKEN_NEWMTL:
					NewMTLMaterial();
					break;
					
			case	TOKEN_AMBIENT_COLOR:
					ReadMTLAmbientColor();
					break;
					
			case	TOKEN_DIFFUSE_COLOR:
					ReadMTLDiffuseColor();
					break;
					
			case	TOKEN_SPECULAR_COLOR:
					ReadMTLSpecularColor();
					break;
					
			case	TOKEN_TF:
					break;
					
			case	TOKEN_ILLUM:
					break;
					
			case	TOKEN_DIFFUSE_COEFF:
					ReadMTLDiffuseCoeff();
					break;
					
			case	TOKEN_SPECULAR_COEFF:
					ReadMTLSpecularCoeff();
					break;
					
			case	TOKEN_SHARPNESS:
					break;
					
			case	TOKEN_NI:
					break;
					
			case	TOKEN_AMBIENT_MAP:
					break;
					
			case	TOKEN_DIFFUSE_MAP:
					ReadMTLDiffuseMap();
					break;
					
			case	TOKEN_SPECULAR_MAP:
					break;
					
		}
	}while (!NextMTLLine());
eof:

	
	
	return(noErr);
}

#if 0
/****************** READ MTL VERTEX DATA ************************/
//
// The current token is "v", so read 3 floats and store into vertex array
//

static Boolean ReadMTLVertexData(void)
{
float	x,y,z;

		/* CHECK IF OVERFLOW */
		
	if (gMTLNumVertices >= MAX_MTL_VERTS)
	{
		DoAlert("\pReadMTLVertexData: vertex array overflow");
		return(true);
	}


		/* TRY TO READ 3 FLOATS */
		
	if (ReadMTLFloat(&x))
		return(true);
	if (ReadMTLFloat(&y))
		return(true);
	if (ReadMTLFloat(&z))
		return(true);


		/* STORE FLOATS */
		
	(*gMTLVertexArray)[gMTLNumVertices].x = x;
	(*gMTLVertexArray)[gMTLNumVertices].y = y;
	(*gMTLVertexArray)[gMTLNumVertices].z = z;
	gMTLNumVertices++;

	return(false);
}
#endif

//============================================================================================
//============================================================================================
//============================================================================================

#pragma mark ========= TEXT PARSING ===========


/***************** READ MTL WORD *******************/
//
// Read current text until cr, lf, sp, or /
//

static Boolean ReadMTLWord(void)
{
short	i;

			/**********************/
			/* READ IN THE STRING */
			/**********************/

	gMTLWord[0] = 0;											// init string
	i = 1;
	
	while((gTheMTLData[gMTLFileIndex] != ' ') &&				// copy until <sp>, CR, or LF or /
			(gTheMTLData[gMTLFileIndex] != CHAR_RETURN) &&
			(gTheMTLData[gMTLFileIndex] != CHAR_LF) &&	
			(gTheMTLData[gMTLFileIndex] != '/'))
	{
		gMTLWord[i] = gTheMTLData[gMTLFileIndex];					// copy char
		gMTLWord[0]++;
		i++;

		if (++gMTLFileIndex >= gFileSize)
			return(true);
	}
	
	
	SkipMTLSpaces();										// after reading token, skip any spaces

	return(false);
}



/****************** READ MTL FLOAT *******************/

static Boolean ReadMTLFloat(float *value)
{
	if (ReadMTLWord())
		return(true);
		
	*value = StringToFloat(gMTLWord);
	
	return(false);
}


/****************** READ MTL INT *******************/

static Boolean ReadMTLInt(long *value)
{
	if (ReadMTLWord())
		return(true);
		
	if (gMTLWord[0] == 0)			// see if empty
	{
		*value = 0;
		return(false);
	}
	
	StringToNum(gMTLWord, value);
	
	return(false);
}


/************************ GET MTL TOKEN ************************/
//
// Assumes the current file pointer is pointing to the start of the next token
//

static int GetMTLToken(void)
{

			/**********************/
			/* READ IN THE STRING */
			/**********************/

	if (ReadMTLWord())
		return(TOKEN_END);

	
		/************************/
		/* INTERPRET THE STRING */
		/************************/
			
	if (EqualString(gMTLWord, "\pnewmtl", false, false))
		return(TOKEN_NEWMTL);

	if (EqualString(gMTLWord, "\pKa", false, false))
		return(TOKEN_AMBIENT_COLOR);
		
	if (EqualString(gMTLWord, "\pKd", false, false))
		return(TOKEN_DIFFUSE_COLOR);
	
	if (EqualString(gMTLWord, "\pKs", false, false))
		return(TOKEN_SPECULAR_COLOR);

	if (EqualString(gMTLWord, "\pTf", false, false))
		return(TOKEN_TF);

	if (EqualString(gMTLWord, "\pillum", false, false))
		return(TOKEN_ILLUM);

	if (EqualString(gMTLWord, "\pd", false, false))
		return(TOKEN_DIFFUSE_COEFF);

	if (EqualString(gMTLWord, "\pNs", false, false))
		return(TOKEN_SPECULAR_COEFF);

	if (EqualString(gMTLWord, "\psharpness", false, false))
		return(TOKEN_SHARPNESS);

	if (EqualString(gMTLWord, "\pNi", false, false))
		return(TOKEN_NI);

	if (EqualString(gMTLWord, "\pmap_Ka", false, false))
		return(TOKEN_AMBIENT_MAP);

	if (EqualString(gMTLWord, "\pmap_Kd", false, false))
		return(TOKEN_DIFFUSE_MAP);

	if (EqualString(gMTLWord, "\pmap_Ks", false, false))
		return(TOKEN_SPECULAR_MAP);


	DoAlert("\pUnrecognized MTL Token");
	DoAlert(gMTLWord);
	return(TOKEN_UNKNOWN);
}



/********************* NEXT LINE ********************/
//
// Skips to next qualified line and skips comment lines and blank lines
//
// Returns true if end of file
//

static Boolean NextMTLLine(void)
{
	do
	{
				/* FIND END OF LINE */
				
		while((gTheMTLData[gMTLFileIndex] != CHAR_RETURN) &&
				(gTheMTLData[gMTLFileIndex] != CHAR_LF))
		{
			if (++gMTLFileIndex >= gFileSize)
				return(true);
		}

				/* FIND START OF NEXT LINE */
				
		while((gTheMTLData[gMTLFileIndex] == CHAR_RETURN) ||
				(gTheMTLData[gMTLFileIndex] == CHAR_LF))
		{
			if (++gMTLFileIndex >= gFileSize)
				return(true);
		}

				/* POINT TO TOKEN */
				
		if (SkipMTLUnwanted())
			return(true);
			
			
			/* IF TOKEN IS A COMMENT, THEN SKIP THIS LINE */			
	
	}while(gTheMTLData[gMTLFileIndex] == '#');


	return(false);
}


/********************** SKIP SPACES ***********************/
//
// skips spaces CR, and LF.
//

static Boolean SkipMTLSpaces(void)
{
	while(gTheMTLData[gMTLFileIndex] == ' ')
	{
		if (++gMTLFileIndex >= gFileSize)
			return(true);
	}
	
	return(false);
}

/********************** SKIP UNWANTED ***********************/
//
// skips spaces CR, and LF.
//

static Boolean SkipMTLUnwanted(void)
{
	while((gTheMTLData[gMTLFileIndex] == ' ') ||
		(gTheMTLData[gMTLFileIndex] == CHAR_RETURN) || 	// also skip CR / LF's
		(gTheMTLData[gMTLFileIndex] == CHAR_LF))
	{
		if (++gMTLFileIndex >= gFileSize)
			return(true);
	}
	
	return(false);
}


//==============================================================================================================
//==============================================================================================================
//==============================================================================================================

#pragma mark ========== MATERIALS ==============


/**************** NEW MTL MATERIAL *******************/

static void NewMTLMaterial(void)
{
	if (gNumOBJMaterials >= MAX_OBJ_MATERIALS)
	{
		DoAlert("\pNewMTLMaterial: exceeded max allowed materials!");
		return;
	}

			/* GET MATERIAL NAME */

	if (ReadMTLWord())											// get name string
		return;		
	CopyPStr(gMTLWord, gOBJMaterials[gNumOBJMaterials].name);	// save into struct


			/* INIT FIELDS */
			
	gOBJMaterials[gNumOBJMaterials].ambientColor.r = .5;
	gOBJMaterials[gNumOBJMaterials].ambientColor.g = .5;
	gOBJMaterials[gNumOBJMaterials].ambientColor.b = .5;

	gOBJMaterials[gNumOBJMaterials].diffuseColor.r = .5;
	gOBJMaterials[gNumOBJMaterials].diffuseColor.g = .5;
	gOBJMaterials[gNumOBJMaterials].diffuseColor.b = .5;

	gOBJMaterials[gNumOBJMaterials].specularColor.r = 1;
	gOBJMaterials[gNumOBJMaterials].specularColor.g = 1;
	gOBJMaterials[gNumOBJMaterials].specularColor.b = 1;

	gOBJMaterials[gNumOBJMaterials].diffuseCoeff = 1;
	gOBJMaterials[gNumOBJMaterials].specularCoeff = 1;

	gOBJMaterials[gNumOBJMaterials].texture = nil;


	gNumOBJMaterials++;
}



/****************** READ MTL AMBIENT COLOR *******************/

static void ReadMTLAmbientColor(void)
{
float	r,g,b;
int		i;

			/* READ R,G,B VALUES */
			
	if (ReadMTLFloat(&r))
		return;
	if (ReadMTLFloat(&g))
		return;
	if (ReadMTLFloat(&b))
		return;


	i = gNumOBJMaterials-1;						// get index into material list

	gOBJMaterials[i].ambientColor.r = r;
	gOBJMaterials[i].ambientColor.g = g;
	gOBJMaterials[i].ambientColor.b = b;
}


/****************** READ MTL DIFFUSE COLOR *******************/

static void ReadMTLDiffuseColor(void)
{
float	r,g,b;
int		i;

			/* READ R,G,B VALUES */
			
	if (ReadMTLFloat(&r))
		return;
	if (ReadMTLFloat(&g))
		return;
	if (ReadMTLFloat(&b))
		return;


	i = gNumOBJMaterials-1;						// get index into material list

	gOBJMaterials[i].diffuseColor.r = r;
	gOBJMaterials[i].diffuseColor.g = g;
	gOBJMaterials[i].diffuseColor.b = b;
}

/****************** READ MTL SPECULAR COLOR *******************/

static void ReadMTLSpecularColor(void)
{
float	r,g,b;
int		i;

			/* READ R,G,B VALUES */
			
	if (ReadMTLFloat(&r))
		return;
	if (ReadMTLFloat(&g))
		return;
	if (ReadMTLFloat(&b))
		return;


	i = gNumOBJMaterials-1;						// get index into material list

	gOBJMaterials[i].specularColor.r = r;
	gOBJMaterials[i].specularColor.g = g;
	gOBJMaterials[i].specularColor.b = b;
}



/****************** READ MTL DIFFUSE COEFF *******************/

static void ReadMTLDiffuseCoeff(void)
{
int		i;

	i = gNumOBJMaterials-1;						// get index into material list

	if (ReadMTLFloat(&gOBJMaterials[i].diffuseCoeff))
		return;
}




/****************** READ MTL SPECULAR COEFF *******************/

static void ReadMTLSpecularCoeff(void)
{
int		i;

	i = gNumOBJMaterials-1;						// get index into material list

	if (ReadMTLFloat(&gOBJMaterials[i].specularCoeff))
		return;
}


/******************* READ MTL DIFFUSE MAP ***********************/

static void ReadMTLDiffuseMap(void)
{
OSErr	iErr;
FSSpec	spec;
TQ3SurfaceShaderObject	theShader;
int		i;

	if (ReadMTLWord())									// read the filename of the texture file
		return;
						
	iErr = FSMakeFSSpec(gTheOBJFileSpec.vRefNum,		// create FSSpec for the file
						 gTheOBJFileSpec.parID,
						 gMTLWord, &spec);	
	if (iErr != noErr)
	{
		DoAlert("\pCannot find texture map library needed by this MTL file");
		DoAlert(gMTLWord);
		return;
	}
	
			/* LOAD THE MAP */
	
	theShader = QD3D_GetQTTextureMap(&spec);
	
	
			/* SAVE INTO LIST */
			
	i = gNumOBJMaterials-1;					
	gOBJMaterials[i].texture = theShader;
}


/******************** SELECT ACTIVE MTL MATERIAL *******************/

void SelectActiveMTLMaterial(unsigned char *matName)
{
long	i,j;

			/* SCAN MATERIALS LIST FOR A MATCH */
			
	for (i = 0; i < gNumOBJMaterials; i++)
	{
		if (gOBJMaterials[i].name[0] != matName[0])			// lengths must match
			continue;
			
		for (j = 1; j <= matName[0]; j++)					// match strings
			if (gOBJMaterials[i].name[j] != matName[j])
				goto nextmat;
	
			/* IF GETS HERE, THEN WE HAVE A MATCH */
			
		gCurrentActiveMTLMaterial = i;
		return;
	
nextmat:;	
	}

	DoAlert("\pSelectActiveMTLMaterial: cannot find matching material in the MTL files");	
}











