/****************************/
/*   	WAVEFRONTOBJ.C 		*/
/* (c)1998 Pangea Software  */
/* By Brian Greenstone      */
/****************************/


/****************************/
/*    EXTERNALS             */
/****************************/

#include <QD3D.h>
#include <QD3DErrors.h>
#include <QD3DGeometry.h>
#include <QD3DGroup.h>
#include "QD3DMath.h"
#include "QD3DShader.h"
#include "QD3DStyle.h"
#include	<QD3DTransform.h>

#include "myglobals.h"
#include "objects.h"
#include "mymenus.h"
#include "mywindows.h"
#include "myevents.h"
#include "misc.h"
#include "wavefrontobj.h"
#include "qd3d_support.h"
#include "3dmf.h"
#include "dialogStuff.h"
#include "optimize.h"


extern	ObjNode					*gModelObjNode;
extern	Boolean					gCurrentIsOptimized;
extern	OBJMaterialType			*gOBJMaterials;
extern	long					gCurrentActiveMTLMaterial;


/****************************/
/*    PROTOTYPES            */
/****************************/

static OSErr DecomposeOBJ(void);
static OSErr LoadOBJFile(void);
static Boolean SkipSpaces(void);
static Boolean NextLine(void);
static int GetToken(void);
static Boolean SkipUnwanted(void);
static Boolean ReadOBJVertexData(void);
static Boolean ReadWord(void);
static Boolean ReadFloat(float *value);
static Boolean ReadInt(long *value);
static Boolean ReadOBJFace(void);
static Boolean ReadOBJVertexNormalData(void);
static Boolean ReadOBJUVData(void);


/****************************/
/*    CONSTANTS             */
/****************************/

#define	MAX_OBJ_VERTS	10000
#define	MAX_OBJ_UVS		10000

#define	MAX_POLY_SIZE	256

enum
{
	TOKEN_END		=	0,
	TOKEN_GROUP,
	TOKEN_VERTEX,
	TOKEN_UV,
	TOKEN_USEMTL,
	TOKEN_MTLLIB,
	TOKEN_FACE,
	TOKEN_OBJECT,
	TOKEN_VERTEX_NORMAL,
	
	TOKEN_UNKNOWN
};


/*********************/
/*    VARIABLES      */
/*********************/

static long	gFileIndex,gFileSize;
static Ptr		gTheOBJData = nil;

int		gOBJNumVertices, gOBJNumUVs, gObjNumVertexNormals;

static TQ3Point3D	**gOBJVertexArray = nil;
static TQ3Param2D	**gOBJUVArray = nil;
static TQ3Vector3D	**gOBJVertexNormalArray = nil;

static	Str32	gWord;

FSSpec 	gTheOBJFileSpec;



/******************* OPEN AND OPTIMIZE OBJ FILE ***********************/

void OpenAndOptimizeOBJFile(void)
{
	InitOBJMaterials();

			/* SELECT AND LOAD THE FILE */
			
	LoadOBJFile();


		/* DECOMPOSE THE OBJ FILE INTO TRIANGLES */

	if (DecomposeOBJ())
		return;


			/* CONVERT THE THING INTO A TRIMESH */
			
	ConvertTriangleListsToTriMeshes();

			/* BUILD SOMETHING TO SHOW */
			
	BuildDisplayModel();

	gCurrentIsOptimized = true;
	
	
			/* CLEANUP */
			
	if (gTheOBJData)
	{
		DisposePtr(gTheOBJData);
		gTheOBJData = nil;
	}


	if (gOBJVertexArray)
	{
		DisposeHandle((Handle)gOBJVertexArray);
		gOBJVertexArray = nil;
	}

	if (gOBJUVArray)
	{
		DisposeHandle((Handle)gOBJUVArray);
		gOBJUVArray = nil;
	}
	
	if (gOBJVertexNormalArray)
	{
		DisposeHandle((Handle)gOBJVertexNormalArray);
		gOBJVertexNormalArray = nil;
	}
}



/********************** LOAD OBJ FILE **************************/

OSErr LoadOBJFile(void)
{
OSErr	iErr;
short	refNum = 0;
StandardFileReply	reply;


				/* SEE IF GET RID OF ANY EXISTING MODEL */
				
	if (gModelObjNode)
	{
		DeleteObject(gModelObjNode);
		gModelObjNode = nil;
	}


			/********************************/
			/* LET USER SELECT THE OBJ FILE */
			/********************************/
		
	StandardGetFile(nil,-1,nil,&reply);
	if (!reply.sfGood)								// see if cancelled 
		return(1);
	gTheOBJFileSpec = reply.sfFile;

	if (gTheOBJData)								// see if nuke old data
	{
		DisposePtr(gTheOBJData);
		gTheOBJData = nil;
	}


			/****************************/
			/* OPEN & LOAD THE OBJ FILE */
			/****************************/

	iErr = FSpOpenDF(&gTheOBJFileSpec, fsRdPerm, &refNum);			// open the file
	if (iErr)
	{
		DoAlert("\pDecomposeOBJ: cannot open OBJ file");
		return(1);
	}
	
	iErr = GetEOF(refNum, &gFileSize);						// get size of file
	if (iErr)
	{
		DoAlert("\pDecomposeOBJ: GetEOF failed");
		goto bail;
	}
	
	gTheOBJData = AllocPtr(gFileSize);						// alloc memory for the file's data
	if (gTheOBJData == nil)
	{
		DoAlert("\pDecomposeOBJ: Cannot alloc memory for file");
		goto bail;
	}
	
	iErr = FSRead(refNum, &gFileSize, gTheOBJData);		// read the file into memory
	if (iErr)
	{
		DoAlert("\pDecomposeOBJ: Reading file failed");
		goto bail;
	}

	return(noErr);
	
	
bail:
	if (refNum)
		FSClose(refNum);
	if (gTheOBJData)
		DisposePtr(gTheOBJData);
		
	return(1);

}


/********************** DECOMPOSE OBJ *****************************/

static OSErr DecomposeOBJ(void)
{
FSSpec	spec;
OSErr	iErr;
	
			/********************/
			/* INITIALIZE STUFF */
			/********************/

			
		/* ALLOCATE MEMORY FOR THE ARRAYS */

	gOBJNumVertices = 0;
	if (!gOBJVertexArray)
	{
		gOBJVertexArray = (TQ3Point3D **)AllocHandle(sizeof(TQ3Point3D) * MAX_OBJ_VERTS);
		if (gOBJVertexArray == nil)
			DoFatalAlert("\pDecomposeOBJ: Cannot allocate vertex array");
	}

	gOBJNumUVs = 0;
	if (!gOBJUVArray)
	{
		gOBJUVArray = (TQ3Param2D **)AllocHandle(sizeof(TQ3Param2D) * MAX_OBJ_UVS);
		if (gOBJUVArray == nil)
			DoFatalAlert("\pDecomposeOBJ: Cannot allocate UV array");
	}
	
	gObjNumVertexNormals = 0;
	if (!gOBJVertexNormalArray)
	{
		gOBJVertexNormalArray = (TQ3Vector3D **)AllocHandle(sizeof(TQ3Vector3D) * MAX_OBJ_VERTS);
		if (gOBJVertexNormalArray == nil)
			DoFatalAlert("\pDecomposeOBJ: Cannot allocate vertex normal array");
	}

			

			/* POINT TO 1ST TOKEN */
							
	gFileIndex = 0;									// start @ front of file
	if (SkipUnwanted())
		goto badfile;
		
	if (gTheOBJData[gFileIndex] == '#')				// see if 1st line is a comment
	{
		if (NextLine())
		{
badfile:		
			DoAlert("\pDecomposeOBJ: the file doesnt contain anything useful");
			return(1);								// nothing found
		}
	}	


		/*******************************/
		/* SCAN AND PROCESS EACH TOKEN */
		/*******************************/

	do
	{
		switch(GetToken())
		{
			case	TOKEN_END:
					goto eof;
					
			case	TOKEN_GROUP:
					break;
					
			case	TOKEN_OBJECT:
					gOBJNumVertices = 0;		// reset this since starting new object
					gOBJNumUVs = 0;
					gObjNumVertexNormals = 0;
					break;
					
			case	TOKEN_VERTEX:
					if (ReadOBJVertexData())
						return(1);
					break;
					
			case	TOKEN_VERTEX_NORMAL:
					if (ReadOBJVertexNormalData())
						return(1);
					break;
					
			case	TOKEN_UV:
					if (ReadOBJUVData())
						return(1);
					break;
					
			case	TOKEN_USEMTL:
					if (ReadWord())				// read the name of the material to activate
						return(1);
					SelectActiveMTLMaterial(&gWord[0]);
					break;
					
			case	TOKEN_MTLLIB:
					if (ReadWord())				// read the filename of the material lib file
						return(1);
										
					iErr = FSMakeFSSpec(gTheOBJFileSpec.vRefNum,	// create FSSpec for the file
										 gTheOBJFileSpec.parID,
										 gWord, &spec);	
					if (iErr != noErr)
					{
						DoAlert("\pCannot find .MTL library needed by this .OBJ file");
						break;
					}
					
					ProcessMTLFile(&spec);		// read the materials out of the file			
					break;
					
			case	TOKEN_FACE:
					if (ReadOBJFace())
						return(1);
					break;
				
		}
	}while (!NextLine());
eof:

	
	
	return(noErr);
}


/****************** READ OBJ VERTEX DATA ************************/
//
// The current token is "v", so read 3 floats and store into vertex array
//

static Boolean ReadOBJVertexData(void)
{
float	x,y,z;

		/* CHECK IF OVERFLOW */
		
	if (gOBJNumVertices >= MAX_OBJ_VERTS)
	{
		DoAlert("\pReadOBJVertexData: vertex array overflow");
		return(true);
	}


		/* TRY TO READ 3 FLOATS */
		
	if (ReadFloat(&x))
		return(true);
	if (ReadFloat(&y))
		return(true);
	if (ReadFloat(&z))
		return(true);


		/* STORE FLOATS */
		
	(*gOBJVertexArray)[gOBJNumVertices].x = x;
	(*gOBJVertexArray)[gOBJNumVertices].y = y;
	(*gOBJVertexArray)[gOBJNumVertices].z = z;
	gOBJNumVertices++;

	return(false);
}


/****************** READ OBJ VERTEX NORMAL DATA ************************/

static Boolean ReadOBJVertexNormalData(void)
{
float	x,y,z;

		/* CHECK IF OVERFLOW */
		
	if (gObjNumVertexNormals >= MAX_OBJ_VERTS)
	{
		DoAlert("\pReadOBJVertexNormalData: array overflow");
		return(true);
	}


		/* TRY TO READ 3 FLOATS */
		
	if (ReadFloat(&x))
		return(true);
	if (ReadFloat(&y))
		return(true);
	if (ReadFloat(&z))
		return(true);


		/* STORE FLOATS */
		
	(*gOBJVertexNormalArray)[gObjNumVertexNormals].x = x;
	(*gOBJVertexNormalArray)[gObjNumVertexNormals].y = y;
	(*gOBJVertexNormalArray)[gObjNumVertexNormals].z = z;
	gObjNumVertexNormals++;

	return(false);
}



/****************** READ OBJ UV DATA ************************/

static Boolean ReadOBJUVData(void)
{
float	u,v;

		/* CHECK IF OVERFLOW */
		
	if (gOBJNumUVs >= MAX_OBJ_VERTS)
	{
		DoAlert("\pReadOBJUVData: array overflow");
		return(true);
	}


		/* TRY TO READ 2 FLOATS */
		
	if (ReadFloat(&u))
		return(true);
	if (ReadFloat(&v))
		return(true);


		/* STORE FLOATS */
		
	(*gOBJUVArray)[gOBJNumUVs].u = u;
	(*gOBJUVArray)[gOBJNumUVs].v = v;
	gOBJNumUVs++;

	return(false);
}




/*********************** READ OBJ FACE ****************************/
//
// This reads in a face and tesselates to triangles, and adds to triangle list in Optimizer
//

static Boolean ReadOBJFace(void)
{
long		iVertex[MAX_POLY_SIZE], iUV[MAX_POLY_SIZE], iNormal[MAX_POLY_SIZE];
short		i,numTriangles,t;
TQ3Point3D	p[3];
TQ3Vector3D vn[3],fn;
TQ3Param2D	uv[3];
TQ3ColorRGB	vertexColors[3];

MaterialType materialType;


		/***********************/
		/* BUILD MATERIAL TYPE */
		/***********************/
	
	materialType.hasVertexColors = false; 
	
	if (gOBJMaterials[gCurrentActiveMTLMaterial].texture)
	{
		materialType.isTextured = true;	
		materialType.textureRef = gOBJMaterials[gCurrentActiveMTLMaterial].texture;
	}
	else
	{
		materialType.isTextured = false;	
		materialType.color.r = gOBJMaterials[gCurrentActiveMTLMaterial].diffuseColor.r;	
		materialType.color.g = gOBJMaterials[gCurrentActiveMTLMaterial].diffuseColor.g;	
		materialType.color.b = gOBJMaterials[gCurrentActiveMTLMaterial].diffuseColor.b;	
	}
	
	materialType.isTransparent = false;
		



	
			/*********************/
			/* READ ALL VERTICES */
			/*********************/

	i = 0;
	do
	{
		if (i >= MAX_POLY_SIZE)
			DoFatalAlert("\pReadOBJFace: # verts on face exceeded max allowed");

			/* READ VERTEX INDEX */
			
		ReadInt(&iVertex[i]);
		if (iVertex[i] == 0)									// see if didnt read it
		{
			DoAlert("\pReadOBJFace: blank vertex index");
			return(true);
		}
		
		if (iVertex[i] < 0)										// see if negative index
			iVertex[i] = gOBJNumVertices+iVertex[i];
		else
			iVertex[i]--;										// convert to 0-based index

		if (gTheOBJData[gFileIndex] == '/')						// see if has another element
		{
			gFileIndex++;
			
				/* READ UV INDEX */
				
			ReadInt(&iUV[i]);
			if (iUV[i] < 0)										// see if negative index
				iUV[i] = gOBJNumUVs+iUV[i];
			else
				iUV[i]--;						

			if (gTheOBJData[gFileIndex] == '/')					// see if has another element
			{
				gFileIndex++;
				
					/* READ NORMAL INDEX */
					
				ReadInt(&iNormal[i]);
				if (iNormal[i] < 0)								// see if negative index
					iNormal[i] = gObjNumVertexNormals+iNormal[i];
				else
					iNormal[i]--;
			}
		}
		
		
		i++;
					
	}while(((gTheOBJData[gFileIndex] >= '0') && (gTheOBJData[gFileIndex] <= '9') ||
			gTheOBJData[gFileIndex] == '-'));


		/************************/
		/*		TESSELATE	    */
		/************************/

	numTriangles = i - 2;
		
	for (t = 0; t < numTriangles; t++)
	{			
//		int	a = 0, b = 1, c = 2;
		int	a = 2, b = 1, c = 0;
	
			/* SET TRIANGLE POINTS */
			
		p[a] 	= (*gOBJVertexArray)[iVertex[0]];
		p[b] 	= (*gOBJVertexArray)[iVertex[t+1]];
		p[c] 	= (*gOBJVertexArray)[iVertex[t+2]];

		
			/* SET TRIANGLE UVS */
			
		uv[a] 	= (*gOBJUVArray)[iUV[0]];
		uv[b] 	= (*gOBJUVArray)[iUV[t+1]];
		uv[c] 	= (*gOBJUVArray)[iUV[t+2]];
	
	
			/* CALC FACE NORMAL */
	
		QD3D_CalcFaceNormal(&p[0], &p[1], &p[2], &fn);
		
			/* SET VERTEX NORMALS */
			
		vn[a] 	= (*gOBJVertexNormalArray)[iNormal[0]];
		vn[b] 	= (*gOBJVertexNormalArray)[iNormal[t+1]];
		vn[c] 	= (*gOBJVertexNormalArray)[iNormal[t+2]];


			/* ADD THIS TRIANGLE TO THE LIST */
		
		AddTriangleToTriangleList(&p[0], &vn[0], &fn, &uv[0], &materialType, &vertexColors[0]);

	}
	
	return(false);
}



//============================================================================================
//============================================================================================
//============================================================================================

#pragma mark ========= TEXT PARSING ===========


/***************** READ WORD *******************/
//
// Read current text until cr, lf, sp, or /
//

static Boolean ReadWord(void)
{
short	i;

			/**********************/
			/* READ IN THE STRING */
			/**********************/

	gWord[0] = 0;											// init string
	i = 1;
	
	while((gTheOBJData[gFileIndex] != ' ') &&				// copy until <sp>, CR, or LF or /
			(gTheOBJData[gFileIndex] != CHAR_RETURN) &&
			(gTheOBJData[gFileIndex] != CHAR_LF) &&	
			(gTheOBJData[gFileIndex] != '/'))
	{
		gWord[i] = gTheOBJData[gFileIndex];					// copy char
		gWord[0]++;
		i++;

		if (++gFileIndex >= gFileSize)
			return(true);
	}
	
	
	SkipSpaces();										// after reading token, skip any spaces

	return(false);
}



/****************** READ FLOAT *******************/

static Boolean ReadFloat(float *value)
{
	if (ReadWord())
		return(true);
		
	*value = StringToFloat(gWord);
	
	return(false);
}

/****************** READ INT *******************/

static Boolean ReadInt(long *value)
{
	if (ReadWord())
		return(true);
		
	if (gWord[0] == 0)			// see if empty
	{
		*value = 0;
		return(false);
	}
	
	StringToNum(gWord, value);
	
	return(false);
}


/************************ GET TOKEN ************************/
//
// Assumes the current file pointer is pointing to the start of the next token
//

static int GetToken(void)
{

			/**********************/
			/* READ IN THE STRING */
			/**********************/

	if (ReadWord())
		return(TOKEN_END);

	
		/************************/
		/* INTERPRET THE STRING */
		/************************/
			
	if (EqualString(gWord, "\pg", false, false))
		return(TOKEN_GROUP);

	if (EqualString(gWord, "\po", false, false))
		return(TOKEN_OBJECT);

	if (EqualString(gWord, "\pv", false, false))
		return(TOKEN_VERTEX);

	if (EqualString(gWord, "\pvn", false, false))
		return(TOKEN_VERTEX_NORMAL);
				
	if (EqualString(gWord, "\pvt", false, false))
		return(TOKEN_UV);
	
	if (EqualString(gWord, "\pusemtl", false, false))
		return(TOKEN_USEMTL);

	if (EqualString(gWord, "\pmtllib", false, false))
		return(TOKEN_MTLLIB);

	if (EqualString(gWord, "\pf", false, false))
		return(TOKEN_FACE);

	DoAlert("\pUnrecognized Token");
	DoAlert(gWord);
	return(TOKEN_UNKNOWN);
}



/********************* NEXT LINE ********************/
//
// Skips to next qualified line and skips comment lines and blank lines
//
// Returns true if end of file
//

static Boolean NextLine(void)
{
	do
	{
				/* FIND END OF LINE */
				
		while((gTheOBJData[gFileIndex] != CHAR_RETURN) &&
				(gTheOBJData[gFileIndex] != CHAR_LF))
		{
			if (++gFileIndex >= gFileSize)
				return(true);
		}

				/* FIND START OF NEXT LINE */
				
		while((gTheOBJData[gFileIndex] == CHAR_RETURN) ||
				(gTheOBJData[gFileIndex] == CHAR_LF))
		{
			if (++gFileIndex >= gFileSize)
				return(true);
		}

				/* POINT TO TOKEN */
				
		if (SkipUnwanted())
			return(true);
			
			
			/* IF TOKEN IS A COMMENT, THEN SKIP THIS LINE */			
	
	}while(gTheOBJData[gFileIndex] == '#');


	return(false);
}


/********************** SKIP SPACES ***********************/
//
// skips spaces CR, and LF.
//

static Boolean SkipSpaces(void)
{
	while(gTheOBJData[gFileIndex] == ' ')
	{
		if (++gFileIndex >= gFileSize)
			return(true);
	}
	
	return(false);
}

/********************** SKIP UNWANTED ***********************/
//
// skips spaces CR, and LF.
//

static Boolean SkipUnwanted(void)
{
	while((gTheOBJData[gFileIndex] == ' ') ||
		(gTheOBJData[gFileIndex] == CHAR_RETURN) || 	// also skip CR / LF's
		(gTheOBJData[gFileIndex] == CHAR_LF))
	{
		if (++gFileIndex >= gFileSize)
			return(true);
	}
	
	return(false);
}





